#include <iostream>
#include <sstream>
#include "rank_recursive.h"
using namespace std;

void usage() {
  cerr << "unwind <rank list>" << endl;
  cerr << "  Unwinds a recursive rank list, showing all ranks in it." << endl;
  cerr << endl;
  cerr << "Example:" << endl;
  cerr << "  0   1   2   3   4      For the node topology  " << endl;
  cerr << "    +----------+         at left, the rank list " << endl;
  cerr << "  5 | 6   7   8|  9      for the ranks in the   " << endl;
  cerr << "    |          |         box would be:          " << endl;
  cerr << " 10 |11  12  13| 14                             " << endl;
  cerr << "    |          |           1 -2 6 3 5 3 1       " << endl;
  cerr << " 15 |16  17  18| 19                             " << endl;
  cerr << "    +----------+                                " << endl;
  cerr << " 20  21  22  23  24                             " << endl;
  cerr << "                                                " << endl;
  cerr << " 25  26  27  28  29                             " << endl;
  cerr << endl;
  cerr << "To unwind this and get the rank set, use:" << endl;
  cerr << "  > unwind 1 -2 6 3 5 3 1" << endl;
  cerr << "  6 7 8 11 12 13 16 17 18" << endl;
  exit(1);
}


void output_unrolled(int *unrolled) {
  if (*unrolled != RANK_END) {
    cout << *unrolled;
    for (int *i = &unrolled[1]; *i != RANK_END; i++) {    
      cout << " " << *i;
    }
    cout << endl;
  }
}


int main(int argc, char **argv) {
  if (argc < 2) usage();
  
  ostringstream line;
  line << argv[1];
  for (int i=2; i < argc; i++) {
    line << " " << atoi(argv[i]);
  }

  int *rlist = parse_recursive_list(line.str().c_str());
  int *unrolled = new int[524288];  // 4x size of bgl
  merge_unroll(rlist, unrolled);
  output_unrolled(unrolled);

  cout << member_count(rlist) << " members." << endl;

  delete [] unrolled;
  exit(0);
}
