#ifndef __PARAM_H__
#define __PARAM_H__

#ifdef __cplusplus
#include <string>
#include <iostream>
#include <string.h>
#include <mpi.h>
#include <mpi_util.h>
#include <stdio.h>
#include <stdlib.h>

enum Type {
  INT,
  FLOAT,
  DOUBLE,
  STRING
};

enum ParamType {
  SCALAR,
  VECTOR
};

union UParam {
  int ivalue;
  float fvalue;
  double dvalue;
  char* svalue;
};

class MPIParam {

private:
  UParam uparam;
  enum Type identifier;
  long counter;

public:
  MPIParam() { }

  ~MPIParam() {
    if(identifier == STRING)
      free(uparam.svalue);
  }
 
  explicit MPIParam(int value) {
    uparam.ivalue = value;
    identifier = INT;
    counter = 0;
  }
  explicit MPIParam(float value) {
    uparam.fvalue = value;
    identifier = FLOAT;
    counter = 0;
  }
  explicit MPIParam(double value) {
    uparam.dvalue = value;
    identifier = DOUBLE;
    counter = 0;
  }
  explicit MPIParam(char* value) {
    uparam.svalue = value;
    identifier = STRING;
    counter = 0;
  }

  void add(int value) {
    uparam.ivalue = value;
    identifier = INT;
  }
  void add(float value) {
    uparam.fvalue = value;
    identifier = FLOAT;
  }
  void add(double value) {
    uparam.dvalue = value;
    identifier = DOUBLE;
  }
  void add(char* value) {
    uparam.svalue = value;
    identifier = STRING;
  }

  int getIdentifier() {
    return identifier;
  }

  void getParamValue(void *val) {
    if(identifier == INT){
      int *temp = (int *)val;
      *temp = uparam.ivalue;
    }
    else if(identifier == FLOAT) {
      float *temp = (float *)val;
      *temp = uparam.fvalue;
    }
    else if(identifier == DOUBLE) {
      double *temp = (double *)val;
      *temp = uparam.dvalue;
    }
    else if(identifier == STRING) {
      char *temp = (char *)val;
      temp = uparam.svalue;
    }
  }

  int getValue(int val) {
    return uparam.ivalue;
  }

  float getValue(float val) {
    return uparam.fvalue;
  }

  double getValue(double val) {
    return uparam.dvalue;
  }

  char* getValue(char *val) {
    return uparam.svalue;
  }
  
  long getCounter() {
    return counter;
  }

  void inc_counter() {
    counter++;
  }

  void print() {
    //std::cout << "MPIParam val " << uparam.ivalue << std::endl;
  }

  int soutput_value(char *buf) {
    char *pos = buf;
    if(identifier == INT)
	pos += sprintf(pos, "%d", uparam.ivalue);
    else if(identifier == FLOAT)
	pos += sprintf(pos, "%f", uparam.fvalue);
    else if(identifier == DOUBLE)
	pos += sprintf(pos, "%f", uparam.dvalue);
    else if(identifier == STRING)
	pos += sprintf(pos, "%s", uparam.svalue);
    return (pos-buf);
  }

  int get_packed_size(MPI_Comm comm);    

  void pack(void *buf, int bufsize, int *position, MPI_Comm comm);

  void unpack(void *buf, int bufsize, int *position, MPI_Comm comm);

};

int compareMPIParam (MPIParam * m1, MPIParam * m2, int * diff_percent);/* {

  int identifier = m1->getIdentifier();
  if(identifier != m2->getIdentifier()) {
    return 0;
  }

  if(identifier == INT) {
    int val1, val2;
    val1 = m1->getValue(val1);
    val2 = m2->getValue(val2);
    if(val1 == val2)
	return 1;
  }
  else if(identifier == FLOAT) {
    float val1, val2;
    val1 = m1->getValue(val1);
    val2 = m2->getValue(val2);
    if(val1 == val2)
	return 1;
  }
  else if(identifier == DOUBLE) {
    double val1, val2;
    val1 = m1->getValue(val1);
    val2 = m2->getValue(val2);
    if(val1 == val2)
	return 1;
  }
  else if(identifier == STRING) {
    char *val1;
    char *val2;
    val1 = m1->getValue(val1);
    val2 = m2->getValue(val2);
    if(!strcmp(val1, val2))
	return 1;
  }

  return 0;
}*/

template <class T>
class GenericParam {

public:
  virtual ~GenericParam(){
   //std::cout << "Inside destructor of GenericParam" << std::endl;
  }

  virtual void add(T value) = 0;

  virtual void add_histo (T value) = 0;

  virtual int compare(GenericParam&) = 0;

  virtual void printValue() = 0;

  virtual MPIParam getMPIParam() = 0;

  virtual int soutput_param(char *pos) = 0;

  virtual void** getVector() = 0;

  virtual int** get_end_point_list() = 0;

  virtual int param_packed_size(MPI_Comm comm) = 0;

  virtual void unpack(void *buf, int bufsize, int *position, MPI_Comm comm) = 0;

  virtual void pack(void *buf, int bufsize, int *position, MPI_Comm comm) = 0;

  virtual int get_class_type() = 0;

  virtual void* getValueList() = 0;

  virtual void param_merge_value_list(void *p, int *m_rlist, int *s_rlist) = 0;

  virtual void* getHisto() = 0;
};

extern "C" {
#endif /*end of __cplusplus*/
typedef struct MPIParam MPIParam;

void init_mpiparam_int(MPIParam **param, int val);

void init_mpiparam_float(MPIParam **param, float val);

void init_mpiparam_double (MPIParam **param, double val);

void init_mpiparam_string (MPIParam **param, char* val);

#ifdef __cplusplus
}
#endif /*__cplusplus*/

#endif /*__PARAM_H__*/
