#ifndef __SCALAR_PARAM_H__
#define __SCALAR_PARAM_H__

#ifdef __cplusplus

#include <iostream>
#include <vector>
#include "Param.h"
#include "Histogram.h"
#include "ScalarParamList.h"
#include "cvector.h"

#define USER_DEFINED_ERROR 80

//template <class MPIParam>
class GenericScalarParam : public GenericParam<MPIParam> {

private:
  MPIParam value;
  int track_rlist;
  Histogram * histo;
  std::vector<ScalarParamList *> value_list;
  cvector * vec;
  int *end_point_list;

public:

  GenericScalarParam():histo(0),track_rlist(0),vec(0),end_point_list(0){ }

  GenericScalarParam (MPIParam value, int track_rlist);

  GenericScalarParam& operator= (const GenericScalarParam&);

  GenericScalarParam( const GenericScalarParam& );

  ~GenericScalarParam();

  void add (MPIParam value);

  void add_histo (MPIParam value);

  void ** getVector() {return (void**)&vec;}

  int compare(GenericParam<MPIParam>&);

  MPIParam getMPIParam() {
    return value;
  }

  int ** get_end_point_list() { return &end_point_list; }

  void add_to_rlist(MPIParam mp);

  void printValue();

  int soutput_param(char *);

  int param_packed_size(MPI_Comm comm);

  void unpack(void *buf, int bufsize, int *position, MPI_Comm comm);

  void pack(void *buf, int bufsize, int *position, MPI_Comm comm);

  int get_class_type() { return SCALAR; }

  void param_merge_value_list(void *p, int *m_rlist, int *s_rlist);

  void* getValueList(){ return (void*)&value_list; }

  void add_to_value_list( ScalarParamList* s);

  void merge_value_list(vector<ScalarParamList*>* slave);

  void* getHisto() { return (void*)histo; }
};
//typedef struct GenericScalarParam<MPIParam> ScalarP;
int param_merge_vectors(GenericParam<MPIParam>* merge, GenericParam<MPIParam>* target);

void param_vector_reset_undo(GenericParam<MPIParam>* p);

void param_vector_undo_addition(GenericParam<MPIParam>* p);


extern "C" {
#endif /*end of __cplusplus*/
typedef struct GenericScalarParam Scalar;

void init_scalar_param_int (Scalar **scalar, int val);

void init_scalar_param_float (Scalar **scalar, float val);

void init_scalar_param_double (Scalar **scalar, double val);

void init_scalar_param_string (Scalar **scalar, char* val);

#ifdef __cplusplus
}
#endif /*__cplusplus*/


#endif /*__SCALAR_PARAM_H__*/
