/*
 * Filename: TimeList.h
 * Author:   ratn1
 * Created:  Wed Sep 19 11:16:08 2007
 * $Id: TimeList.h,v 1.3 2008/03/20 06:50:42 pratn Exp $
 */
#ifndef __TIMELIST_H__
#define __TIMELIST_H__

#include "stack_sig.h"

#include "Histogram.h"
#include "Intseq.h"

#ifdef FEATURE_TIMING_BITS
  typedef struct Intseq TimeRecord;
#else
  typedef struct Histogram TimeRecord;
#endif 

#ifdef __cplusplus
#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <cassert>
#include <cstring>
#include <sstream>
#include <list>
#include <stdint.h>
#include <mpi.h>

class TimeList {
private:
  // what this class holds depends on what kind of timing we're doing.
  typedef std::list<TimeRecord*> TRList;
  typedef TRList::iterator tr_iter;
  typedef TRList::const_iterator trc_iter;

  TRList tList;

  struct list_deleter {
    void operator()(TimeRecord *elt) { 
      delete elt;
    }
  };
  
public:
  TimeList ();
  ~TimeList();
  int output (char * buf, int index) const;
  int get_packed_size (MPI_Comm comm) const;
  void pack (void *buf, int bufsize, int *position, MPI_Comm comm) const;
  static TimeList * unpack (void * buf, int bufsize, int * position, MPI_Comm comm);

  TimeRecord * begin();
  void insert (TimeRecord *t);

  int size () { return tList.size(); }

  TimeRecord *find_match (const stack_sig* sig) const;

  void merge (long int dest_time, const stack_sig * dest_sig, int dest_rank,
	      TimeList * psrc,
	      long int src_time, const stack_sig* src_sig, int src_rank,  int B);
};

extern "C" {
#else //__cplusplus

  typedef struct TimeList TimeList;

#endif //__cplusplus

  void timing_list_free (TimeList ** list);
  int timing_list_output(char *buf, TimeList *histo_list, int index);
  TimeRecord * timing_find_match (TimeList * rec_list, const stack_sig *sig);
  TimeRecord * timing_list_begin (TimeList * histo_list);
  int timing_list_size (TimeList * list);
  
  void timing_list_merge  (TimeList ** psrc, 
    long int src_time, const stack_sig* src_sig, int dest_rank,
    TimeList ** pdest,
    long int dest_time, const stack_sig* dest_sig, int src_rank, int B
  );

  int timing_list_packed_size (TimeList* histo_list, MPI_Comm comm);

  void timing_list_pack (TimeList * histo_list, void *buf, int bufsize, 
			 int *position, MPI_Comm comm);

 void timing_list_unpack (TimeList ** histo_list, void * buf, int bufsize, 
			  int *position,MPI_Comm comm);

 void timing_list_insert (TimeList ** histo_list, TimeRecord *record);


#ifdef __cplusplus
}
#endif  // __cplusplus

#endif // __TIMELIST_H__
/* vim: set tw=80 sw=2 expandtab: */
