/* 
 * filename : list.h
 * pratn    Prasun Ratn
 *
 * $Id: list.h,v 1.4 2008/04/16 21:17:56 pratn Exp $
 */
#ifndef __LIST_H__
#define __LIST_H__

#ifdef __cplusplus
extern "C" {
#endif
#include <stdint.h>
#include <mpi.h>
#include "mpi_util.h"

struct list_node {
    void * v;
    struct list_node * next;
    struct list_node * prev;
};
typedef struct list_node * list_t;
typedef void (*list_f) (void *);

/* Insert an item in the list */
list_t list_insert (list_t N, void * v);

/* size of list */
int list_size (list_t N);

/* Delete an item from the list */
list_t list_delete (list_t N, void * v, list_f func);

/* Next item in the list */
list_t list_next   (list_t N);

/* Previous item in the list */
list_t list_prev   (list_t N);

/* Value of an item in the list */
void * list_value (list_t N);

/* concat two lists */
list_t list_concat (list_t l1, list_t l2);

/* empty a list */
void list_free (list_t * N, list_f func);

/* print a symbol table */
typedef void (list_print_fn) (void * v);
void list_print (list_t N, list_print_fn * fn);

int list_packed_size(list_t list, MPI_Comm comm);

void list_pack(list_t list, void * buf, int bufsize, int * position, MPI_Comm comm);

list_t list_unpack (void *buf,  int bufsize, int * position, MPI_Comm comm);

#ifdef __cplusplus
}
#endif
#endif
/* EOF */
