#ifndef OPSTRUCT_MACRO_H
#define OPSTRUCT_MACRO_H

#include "TimeList.h"

#ifdef __cplusplus
extern "C" {
#endif

#define OP_ANY_SOURCE (INT_MAX)
#define ENCODE_SOURCE(s) (((s) == MPI_ANY_SOURCE) ?  OP_ANY_SOURCE : ((s) - my_rank + my_size) % my_size)
#define DECODE_SOURCE(s) (((s) ==  OP_ANY_SOURCE) ? MPI_ANY_SOURCE : ((s) + my_rank)%my_size)
  
#define ENCODE_DEST(d) ((d - my_rank + my_size) % my_size)
#define DECODE_DEST(d) ((d + my_rank) % my_size)

#define ENCODE_KEY(key)   (key)
#define DECODE_KEY(key)   (key)

#define   set_rank_offset(op, field)  ((op).rank_offsets |= (field))
#define   has_rank_offset(op, field)  ((op).rank_offsets &  (field) ?  1  :  0 )
#define  show_rank_offset(op, field)  ((op).rank_offsets &  (field) ? 'o' : 'c')

// send/recv tags may not be relevant for replay purposes
// So we don't record this information for better compression
//#define FEATURE_RETAIN_TAGS
#ifdef FEATURE_RETAIN_TAGS
#define ENCODE_TAG(a) (a)
#else
#define ENCODE_TAG(a) (0)
#endif

#define SYNC                       0x526526

/* list of possible arrays associated */
#define ARRAY_MASK                 0xFFll  /* First 16 bits denote arrays */
#define NO_ARRAYS                  0ll
#define RECVCOUNTS                 1ll << 1
#define DISPLS                     1ll << 2
#define COUNTS                     1ll << 3
#define RDISPLS                    1ll << 4
#define ARRAY_OF_DIMS              1ll << 5
#define ARRAY_OF_PERIODS           1ll << 6
#define DIMSP                      1ll << 7
#define ARRAY_OF_DEGREES           1ll << 8
#define ARRAY_OF_EDGES             1ll << 9
#define ARRAY_OF_RANKS             1ll << 10
#define RANGES                     1ll << 11
#define ARRAY_OF_REQUESTS          1ll << 12
#define ARRAY_OF_BLOCKLENS         1ll << 13
#define ARRAY_OF_DISPLACEMENTS     1ll << 14
#define ARRAY_OF_TYPES             1ll << 15

/* list of op types for ascii parsing */
/* TODO: should we put a prefix on these? */
#define NO_FIELDS                  0ll
#define COMM                       1ll << 16
#define COUNT                      1ll << 18
#define DATATYPE                   1ll << 20
#define RECVCOUNT                  1ll << 21
#define RECVTYPE                   1ll << 22
#define MPI_OP                     1ll << 24
#define ROOT                       1ll << 25
#define DEST                       1ll << 26
#define TAG                        1ll << 27
#define REQUEST                    1ll << 28
#define REORDER                    1ll << 29
//#define COMM_OUT                   1ll << 30
#define NEWRANK                    1ll << 32
#define GROUP                      1ll << 33
//#define GROUP_OUT                  1ll << 34
#define COLOR                      1ll << 35
#define KEY                        1ll << 36
#define GROUP1                     1ll << 37
#define GROUP2                     1ll << 38
#define LOCAL_LEADER               1ll << 39
#define PEER_COMM                  1ll << 40
#define REMOTE_LEADER              1ll << 41
#define HIGH                       1ll << 42
#define SOURCE                     1ll << 43
#define SENDTAG                    1ll << 44
#define RECVTAG                    1ll << 45
#define NEWTYPE                    1ll << 46
#define OLDTYPE                    1ll << 48
#define BLOCKLEN                   1ll << 49
#define STRIDE                     1ll << 50
#define GLOBAL_WORLD               1ll << 51
#define NULL_REQ                   1ll << 52
#define OUTCOUNT                   1ll << 53

/*Added by Karthik - Start*/
#ifdef FEATURE_FILE_IO   
/*Commented out the values present previously as the parameters are extended now*/
/*#define FILENAME	                 1ll << 55
#define AMODE		                   1ll << 56
#define FH		                     1ll << 57
#define OFFSET		                 1ll << 58
#define DISP 		                   1ll << 59
#define DATAREP 	                 1ll << 60
#define ETYPE 	 	                 1ll << 61
#define FILETYPE 	                 1ll << 62
#define WHENCE		                 1ll << 63*/

/*Used to differentiate whether 1st set or 2nd set of fields to be used */
#define FIELD_SEPARATOR_MASK       1ll << 63

#define ARRAY_OF_GSIZES            1ll << 1
#define ARRAY_OF_DISTRIBS          1ll << 2
#define ARRAY_OF_DARGS             1ll << 3
#define ARRAY_OF_PSIZES            1ll << 4

#define FILENAME                   1ll << 16
#define AMODE                      1ll << 17
#define FH                         1ll << 18
#define OFFSET                     1ll << 19
#define DISP                       1ll << 20
#define DATAREP                    1ll << 21
#define ETYPE                      1ll << 22
#define FILETYPE                   1ll << 23
#define WHENCE                     1ll << 24
#define COMM_SIZE                  1ll << 25
#define RANK                       1ll << 26
#define NDIMS                      1ll << 27
#define ORDER                      1ll << 28
#define POSIX_MODE                 1ll << 29
#define POSIX_FLAGS                1ll << 30
#endif
/*Added by Karthik - End*/

struct time_data
{
  TimeList * time_list;
  long long int time;
};

enum
{
  TIMING_INVALID,
  TIMING_MIN_MAX_AVG,
  TIMING_HISTOGRAM
};

#if defined (FEATURE_TIMING_HISTO) || defined (FEATURE_TIMING_BITS)
#define NUM_BUCKETS 5
  
enum
{
  TIMING_PRE,  /* pre-wrapper histogram  */
  TIMING_POST, /* post-wrapper histogram */
  TIMING_MAX
};

#endif

/* Macro for allocating arrays in opstructs. */
#define OP_ARRAY_ALLOC(size) ((array_elt_t*) malloc (sizeof(array_elt_t) * size))

#define OP_AINT_ARRAY_ALLOC(size) ((MPI_Aint*) malloc (sizeof(MPI_Aint) * size))

/* Macro for copying values from mpi params to arrays in opstructs.
   This uses assignment and not memcpy so that ints are converted to 
   long where needed. 
   TODO: optimize so that this uses memcpy when sizes of contents match.
*/
#define __OP_ARRAY_COPY(dest_arr, src_arr, copy_size, type) do {\
    int i; \
    for (i=0; i < copy_size; i++) dest_arr[i] = (type)src_arr[i]; \
} while (0)

#define OP_ARRAY_COPY(dest_arr, src_arr, copy_size) __OP_ARRAY_COPY(dest_arr, \
    src_arr, copy_size, array_elt_t)

#define OP_AINT_ARRAY_COPY(dest_arr, src_arr, copy_size) \
  __OP_ARRAY_COPY(dest_arr, src_arr, copy_size, MPI_Aint)

#define OP_ARRAY_COPY_TYPES(dest_arr, src_arr, copy_size, t) do {\
    int i; \
    for (i=0; i < copy_size; i++) \
      dest_arr[i] = (array_elt_t) type_to_index (t, src_arr[i]); \
} while (0)

#define OP_ARRAY_GET_TYPES(dest_arr, src_arr, copy_size, t) do {\
    int i; \
    for (i=0; i < copy_size; i++) \
      dest_arr[i] = (MPI_Datatype) index_to_type(t, src_arr[i]); \
} while (0)

/*Macros copied from params.h*/
#define MATCH      (0)
#define MFAIL_GEN  (1)
#define MFAIL_ARR  (1<<1)
#define MFAIL_SRC  (1<<2)
#define MFAIL_DST  (1<<3)
#define MFAIL_REQ  (1<<4)
#define MFAIL_CNT  (1<<5)
#define MFAIL_CLR  (1<<6)
#define MFAIL_KEY  (1<<7)

//Added by Karthik - Start
#ifdef FEATURE_FILE_IO

#define MFAIL_FLN  (1<<8)
#define MFAIL_OFF  (1<<9) 
#define MFAIL_DISP (1<<10)
#define MFAIL_ROOT (1<<11)
#endif
//Added by Karthik - End
/*End - Macros copied from params.h*/

/*New constants for parameters*/
#define TYPE_RECVCOUNTS                 1
#define TYPE_DISPLS                     2
#define TYPE_COUNTS                     3
#define TYPE_RDISPLS                    4
#define TYPE_ARRAY_OF_DIMS              5
#define TYPE_ARRAY_OF_PERIODS           6
#define TYPE_DIMSP                      7
#define TYPE_ARRAY_OF_DEGREES           8
#define TYPE_ARRAY_OF_EDGES             9
#define TYPE_ARRAY_OF_RANKS             10
#define TYPE_RANGES                     11
#define TYPE_ARRAY_OF_REQUESTS          12
#define TYPE_ARRAY_OF_BLOCKLENS         13
#define TYPE_ARRAY_OF_DISPLACEMENTS     14
#define TYPE_ARRAY_OF_TYPES             15

/* list of op types for ascii parsing */
/* TODO: should we put a prefix on these? */
#define TYPE_NO_FIELDS                  0
#define TYPE_COMM                       16
#define TYPE_COUNT                      18
#define TYPE_DATATYPE                   20
#define TYPE_RECVCOUNT                  21
#define TYPE_RECVTYPE                   22
#define TYPE_MPI_OP                     24
#define TYPE_ROOT                       25
#define TYPE_DEST                       26
#define TYPE_TAG                        27
#define TYPE_REQUEST                    28
#define TYPE_REORDER                    29
//#define TYPE_COMM_OUT                   30
#define TYPE_NEWRANK                    32
#define TYPE_GROUP                      33
//#define TYPE_GROUP_OUT                  34
#define TYPE_COLOR                      35
#define TYPE_KEY                        36
#define TYPE_GROUP1                     37
#define TYPE_GROUP2                     38
#define TYPE_LOCAL_LEADER               39
#define TYPE_PEER_COMM                  40
#define TYPE_REMOTE_LEADER              41
#define TYPE_HIGH                       42
#define TYPE_SOURCE                     43
#define TYPE_SENDTAG                    44
#define TYPE_RECVTAG                    45
#define TYPE_NEWTYPE                    46
#define TYPE_OLDTYPE                    48
#define TYPE_BLOCKLEN                   49
#define TYPE_STRIDE                     50
#define TYPE_GLOBAL_WORLD               51
#define TYPE_NULL_REQ                   52
#define TYPE_OUTCOUNT                   53

/*Added by Karthik - Start*/
#ifdef FEATURE_FILE_IO   
/*Used to differentiate whether 1st set or 2nd set of fields to be used */

#define TYPE_ARRAY_OF_GSIZES           54 
#define TYPE_ARRAY_OF_DISTRIBS         55 
#define TYPE_ARRAY_OF_DARGS            56 
#define TYPE_ARRAY_OF_PSIZES           57 

#define TYPE_FILENAME                  58 
#define TYPE_AMODE                     59 
#define TYPE_FH                        60
#define TYPE_OFFSET                    61 
#define TYPE_DISP                      62
#define TYPE_DATAREP                   63 
#define TYPE_ETYPE                     64 
#define TYPE_FILETYPE                  65 
#define TYPE_WHENCE                    66 
#define TYPE_COMM_SIZE                 67 
#define TYPE_RANK                      68 
#define TYPE_NDIMS                     69 
#define TYPE_ORDER                     70
#define TYPE_POSIX_MODE                71 
#define TYPE_POSIX_FLAGS               72 
#endif

#ifdef __cplusplus
}
#endif

#endif /*End: OPSTRUCT_MACRO_H*/
