#ifndef _PRSD_STACK_H
#define _PRSD_STACK_H

#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include "stack_sig.h"
#include "Histogram.h"

#define BUFSIZE 256

#ifdef __cplusplus
extern "C" {
#endif

struct rsd_node_t;

typedef struct prsd_node_t
{
  int loop_iters;
  int loop_mems;

  /*Karthik - new counter to track member count after merging of
    unmatching loops
  */
  int new_loop_mems;

  // callpath of tail of the prsd
  stack_sig tail_sig;

  //Karthik - Added histogram to collect statistical info 
  //to add user induced error.
  Histogram *prsd_stat;

  // memo-ized prsd tail pointer used by replay
  // after tail signature is initially found.
  struct rsd_node_t *tail_ptr;
  struct rsd_node_t *parent_ptr;

  struct prsd_node_t *next;
  struct prsd_node_t *prev;
} prsd_node;

typedef struct
{
  prsd_node *head;
  prsd_node *tail;
} prsd_stack;

void init_prsd_stack(prsd_stack *stack);

void init_prsd_node(prsd_node *node);
void free_prsd_node(prsd_node *node);

void prsd_push_node(prsd_stack *stack, prsd_node *node);
void prsd_push_node_end(prsd_stack *stack, prsd_node *node);

//TODO: make these consistent
void prsd_push(prsd_stack *stack, int members, stack_sig *signature);
void prsd_push_end(prsd_stack *stack, int members);

prsd_node *prsd_pop(prsd_stack *stack);
int get_distance(prsd_node *node);
int get_depth(prsd_node *node);
int soutput_prsd_stack(char *buf, int len, prsd_stack *stack);
void free_prsd_stack(prsd_stack *stack);

#ifdef __cplusplus
}
#endif

#endif
