/**
 * Functions for traversing prsd trees.  Inclused in-order
 * and replay traversal.
 * 
 * Added by Todd Gamblin, reworked from prsd_utils.
 */
#ifndef _PRSD_TRAVERSE_H
#define _PRSD_TRAVERSE_H

#include <mpi.h>
#include <stdio.h>
#include <string.h>
#include "rsd_queue.h"
#include "opstruct.h"
#include "req_handler.h"
#include "rank_recursive.h"

#ifdef __cplusplus
extern "C" {
#endif

#ifdef FEATURE_CODE_GENERATOR
/**
 * Type for visitor function to generate MPI code.  Return value should be 0 if
 * the algorithm should continue traversing, or non-0 to halt traversal.
 * <p>
 * Depth gives the nesting level in PRSDs for the node.
 * iter  gives the iteration in a given prsd
 * <p>
 * For more details, see functions below.
 */
typedef int (rsd_queue_visitor_fn_t)(
  rsd_node *node, int rank, int depth, int iter, int last, FILE * skel_fh
);

#else
/**
 * Type for visitor function.  Return value should be 0 if
 * the algorithm should continue traversing, or non-0 to halt traversal.
 * <p>
 * Depth gives the nesting level in PRSDs for the node.
 * iter  gives the iteration in a given prsd
 * <p>
 * For more details, see functions below.
 */
typedef int (rsd_queue_visitor_fn_t)(
  rsd_node *node, int rank, int depth, int iter, int last
);
#endif

/**
 * Performs a linear traversal of a prsd queue.  PRSD info is
 * ignored, and each rsd_node in the queue is visited once.  Note that nesting
 * depth is still passed to the visitor function, even though we only iterate
 * over things once.  Returns non-zero if the user-defined callback function
 * 'visit' terminated execution before the traversal was complete.
 */
int traverse_queue(rsd_node *head, rsd_queue_visitor_fn_t visit);


/**
 * Performs a replay traversal of an rsd queue.  That is,
 * rsd_nodes are visited multiple times for prsd loops, just as 
 * they were initially recorded.  Returns non-zero if the user-defined
 * callback function 'visit' terminated execution before the replay was
 * complete.
 */
int replay_queue(rsd_node *head, int rank, rsd_queue_visitor_fn_t visit);

int traverse_queue_for_generation(rsd_node *head, int rank, rsd_queue_visitor_fn_t visit, FILE * fh);
#ifdef __cplusplus
}
#endif

#endif /* PRSD_TRAVERSE_H */
