#ifndef _RSD_QUEUE_H
#define _RSD_QUEUE_H

#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include "opstruct.h"
#include "prsd_stack.h"
#include "stack_sig.h"
#include "params_queue.h"
#include "params.h"
#include "list.h"
#include "Event.h"

#ifdef __cplusplus
extern "C" {
#endif

#define LINKLIST_DEBUG_MAX 1000000

/* Bits for rsd_node.rsd_info */
typedef enum {
  SINGLE_RSD   = 1,
  MEMBER_RSD   = 1<<1,
  LEADER_RSD   = 1<<2
} rsd_info_type;

enum {
  Master_Queue,
  Slave_Queue
};

#define BUFSIZE      256
//FIXME: this is getting extreme now. 
#define PBUF         (8*(1<<20))

extern int my_rank;

enum {
  White,
  Gray,
  Black
};

typedef struct {
  int color;
  int index;
  list_t  out_edges;      /* list of dependence edges */
  list_t   in_edges;
  int * rank_list;
} dep_graph_info;

typedef struct rsd_node_t
{
  int node;               /* rank where this originated */
  int *rlist;             /* List of ranks */
  stack_sig signature;    
  rsd_info_type rsd_info; /* if this is part of RSD */
  prsd_stack prsd_info;   /* linked list of RSD info such as iters, members,
                             tail signature, tail pointer, parent pointer */
  int match;              /* stores the result of a compare during inter-node merge */

  replay_op data;         /* MPI information */

  Event *event; /*Karthik - Event to store MPI info*/

  Histogram *hist_param;
  field_mask_t hist_op;

  dep_graph_info graph;
  int location;

  struct rsd_node_t *next;
  struct rsd_node_t *prev;
} rsd_node;

typedef struct
{
  rsd_node *head;
  rsd_node *tail;
  int last_op_added;
} rsd_queue;

void init_rsd_node(rsd_node *node);

/**
 * Frees node and all data structures hanging off of it.
 */
void free_rsd_node(rsd_node *node);

/** Frees all nodes in the queue. */
void free_rsd_nodes(rsd_queue *queue);

/** 
 * Compares two prsd stacks, returns true if their nodes are 
 * the same from the tail down to <depth>.
 */ 
int prsd_match(prsd_stack *ptr1, prsd_stack *ptr2, int depth);
void output_rsd_node(FILE *fp, rsd_node *node);

void init_rsd_queue(rsd_queue *queue);
void rsd_push(rsd_queue *queue, replay_op *operation, Event *event);
void rsd_push_head(rsd_queue *queue, rsd_node *new_head);
void rsd_push_tail(rsd_queue *queue, rsd_node *new_tail);
void rsd_insert_pre(rsd_queue *queue, rsd_node *target, rsd_node *seq_start, rsd_node *seq_end);
void rsd_insert_post(rsd_queue *queue, rsd_node *target, rsd_node *seq_start, rsd_node *seq_end);
rsd_node *rsd_pop(rsd_queue *queue);
rsd_node *rsd_pop_tail(rsd_queue *queue);
int rsd_get_length(rsd_queue *queue);
int rsd_checkqueue(rsd_queue *queue, char *msg);

/**
 * Removes the specified sequence of nodes from the rsd_queue provided.
 */
void unsplice(rsd_queue *queue, rsd_node *start, rsd_node *end);


/**
 * Dumps the entire contents of the rsd queue to the specified file.
 */
void output_rsd_queue(FILE *dest, rsd_queue *queue);

/**
 * upper bound on queue size
 */
int rsd_queue_packed_size(rsd_queue *queue, MPI_Comm comm);

int  dep_graph_packed_size(dep_graph_info * g, MPI_Comm comm);
void dep_graph_pack       (dep_graph_info * g, void *buf, int bufsize, int * pos, MPI_Comm comm);
void dep_graph_unpack     (dep_graph_info * g, void *buf, int bufsize, int * pos, MPI_Comm comm);

#ifdef __cplusplus
}
#endif

#endif  /* _RSD_QUEUE_H */
