#ifndef _PRSD_READER_H
#define _PRSD_READER_H

#include "rsd_queue.h"
#include "list.h"

#ifdef __cplusplus
extern "C" {
#endif

typedef struct {
  int *rank_arr;
  int curr_ptr;
  int dest;
} rank_array_t;


/**
 * Reads an rsd file and builds an rsd queue containing only entries generated
 * by the process with specified rank.
 * <p>
 * Specifying negative rank includes all processes.
 */
rsd_queue *read_rsd_file_rank(const char *filename, int rank);

/**
 * Reads an entire rsd trace into a queue.  This can be either a per-node
 * or merged trace, and all data is included in the queue, regardless of 
 * rank.
 * <p>
 * Note: Same as read_rsd_file_rank(filename, -1);
 */
rsd_queue *read_rsd_file(const char *filename);


void parse_rank(char *buf, rsd_node *target);
void parse_rsd_header(char *buf, rsd_node *target, list_t *tail_list, int rec);
void parse_prsd_header(char *buf, rsd_node *target, list_t *tail_list);
void parse_params_header(char *buf, rsd_queue *target);
void parse_prsd_content(char *buf, rsd_queue *target);
void patch_prsd_tail (list_t * tail_list, rsd_node *target);
void parse_prsd_content_extn(char *buf, rsd_queue *target);
#ifdef __cplusplus
}
#endif

#endif /* _PRSD_READER_H */
