#ifndef STACK_SIG_H
#define STACK_SIG_H

#include <mpi.h>

#include <stdio.h>
#include <setjmp.h>
#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

#define SIG_MATCH(sig1, sig2) \
  ((sig1).sig == (sig2).sig)

#define SIG_LT(sig1, sig2) \
  ((sig1).sig < (sig2).sig)

#define WRAP_DEPTH 4

  /**
   * Structure for tail signatures providing xor of the signature for
   * quick comparison, along with the full tail as a '_'-delineated
   * string of addresses in ascii hex.
   */
  typedef struct stack_sig {
    const intptr_t *sig;
  } stack_sig;

  void sig_init(stack_sig *signature);
  void sig_assign(stack_sig *dest, const stack_sig *src);
  void sig_destroy(stack_sig *signature);

  int sig_packed_size(const stack_sig *signature, MPI_Comm comm);
  void sig_pack(const stack_sig *signature, void *buf, int bufsize, int *position, MPI_Comm comm);
  void sig_unpack(stack_sig *signature, void *buf, int bufsize, int *position, MPI_Comm comm);

  int sig_soutput(char *buf, const stack_sig *signature);
  void sig_input(stack_sig *signature, char *buf);

  /** This does the work of unwinding the stack. */
  void sig_unwind(stack_sig *signature);

  /** Gets the depth of a signature's callpath. */
  int sig_len(const intptr_t *sig);

#ifdef __cplusplus
}
#endif

#endif /* STACK_SIG_H */
