
#include <iostream>
#include "Event.h"
#include "mpi_util.h"

using std::cout; using std::endl;
using std::map;

Event::~Event() {

  //cout << "Event: Inside Event destructor" << endl;
  //Delete all GenericParam objects pointed by map
   for(map<int64_t, GenericParam<MPIParam>* >::const_iterator it = params.begin(); 
    it != params.end(); ++it) {
    //cout << "Deleting param " << it->second << endl;
    delete it->second;
  }
  params.clear();

  for(int i=0; i<TIMING_MAX; i++){
    if(time_info[i].time_list) {
      delete time_info[i].time_list;
      time_info[i].time_list = NULL;
    }
  }
 
  if(user_def_histo_1)
    delete user_def_histo_1;
  if(user_def_histo_2)
    delete user_def_histo_2;
 
}

Event& Event::operator=(const Event& e) {
  if(&e != this) {
    params = e.params;
    //cout << "Event:copy operator: param size: " << params.size() << endl;
    event_id = e.event_id;
    param_list = e.param_list;
    seq_num = e.seq_num;
    for(int i=0; i<TIMING_MAX; i++){
      time_info[i].time_list = e.time_info[i].time_list;
      time_info[i].time = e.time_info[i].time;
    }

    user_def_histo_1 = e.user_def_histo_1;
    user_def_histo_2 = e.user_def_histo_2;
  }
}

void Event::init() {
  params.clear();
  event_id = 0;
  param_list = 0;
  seq_num = 0;
  user_def_histo_1 = 0;
  user_def_histo_2 = 0;
  for(int i=0; i<TIMING_MAX; i++){
    time_info[i].time_list = 0;
    time_info[i].time = 0;
  }

}

void Event::setParamList (int64_t val) {
  param_list = val;
}

void Event::setEventId (int val) {
  event_id = val;
}

void Event::setSeqNum (int val) {
  seq_num = val;
}

void Event::addParam(int64_t param_id, int val) {
  if(param_id == TYPE_DEST || param_id ==TYPE_SOURCE || param_id == TYPE_ROOT) {
    MPIParam mparam(val);
    GenericParam<MPIParam>* scalarparam = new GenericScalarParam(mparam, 1);
    params[param_id] = scalarparam;
  } else {
    GenericParam<MPIParam>* scalarparam = new GenericScalarParam;
    //GenericScalarParam* scalarparam = new GenericScalarParam;
    MPIParam mparam(val);
    scalarparam->add(mparam);
    params[param_id] = scalarparam;
  }
}

void Event::addParam(int64_t param_id, float val) {
  GenericParam<MPIParam>* scalarparam = new GenericScalarParam;
  //GenericScalarParam* scalarparam = new GenericScalarParam;
  MPIParam mparam(val);
  scalarparam->add(mparam);
  params[param_id] = scalarparam;
}

void Event::addParam(int64_t param_id, double val) {
  GenericParam<MPIParam>* scalarparam = new GenericScalarParam;
  //GenericScalarParam* scalarparam = new GenericScalarParam;
  MPIParam mparam(val);
  scalarparam->add(mparam);
  params[param_id] = scalarparam;
}

void Event::addParam(int64_t param_id, char* val) {
  GenericParam<MPIParam>* scalarparam = new GenericScalarParam;
  //GenericScalarParam* scalarparam = new GenericScalarParam;
  MPIParam mparam(val);
  scalarparam->add(mparam);
  params[param_id] = scalarparam;
}

void Event::addVectorParam(int64_t param_id, int *arr_param, int size) {
  GenericParam<MPIParam>* vectorparam = new VectorParam(arr_param, size);
  params[param_id] = vectorparam;
}

GenericParam<MPIParam>* Event::getParam(int64_t param_id) {
  if( (params.find(param_id)) != params.end() )
    return params[param_id];
  else
    return NULL;
}

void Event::printParams() {
 /*if(seq_num == 205) {
  for(map<int64_t, GenericParam<MPIParam>* >::const_iterator it = params.begin(); 
    it != params.end(); ++it) {
    cout << it->first << " ";  it->second->printValue(); 

  }
  cout << endl;
 }*/
}

int Event::compareParams(Event *e) {
  //cout << "Event: compareParams" << endl;
  int retval = MATCH;
  if(event_id != e->getEventId() ){
    return MFAIL_GEN;
  }
  typedef map<int64_t, GenericParam<MPIParam>* > pmap;
  typedef pmap::const_iterator map_iter;
  pmap param2 = e->getParamMap();
  //map_iter it2 = param2.begin();

  map_iter it1;
  
  for(it1 = params.begin(); it1 != params.end(); ++it1) {
    //cout << "Event compareParams: p1 " << it1->second << " p2 " << param2[it1->first] << endl;
    if(!(it1->second->compare( *param2[it1->first])) ) {
      //cout << "Event " << event_id << " seqnum " << seq_num << " param " << it1->first << " does not match" << endl;
      retval |= checkCompareException(it1->first);
    } 
  }
  
  return retval;
}

int Event::checkCompareException(int64_t param_id) {

//TODO: Handle ARRAY and FILE IO special cases.
  int64_t field = 1ll << param_id;

  if(!(param_special_case & field))
    return MFAIL_GEN;

  if((param_special_case & field) == (1ll << TYPE_COUNT))
    return MFAIL_CNT;

  if((param_special_case & field) == (1ll << TYPE_DEST))
    return MFAIL_DST;
  
  if((param_special_case & field) == (1ll << TYPE_REQUEST))
    return MFAIL_REQ;

  if((param_special_case & field) == (1ll << TYPE_COLOR))
    return MFAIL_CLR;

  if((param_special_case & field) == (1ll << TYPE_KEY))
    return MFAIL_KEY;

  if((param_special_case & field) == (1ll << TYPE_SOURCE))
    return MFAIL_SRC;

  if((param_special_case & field) == (1ll << TYPE_ROOT))
    return MFAIL_ROOT;

  return MFAIL_GEN;
}

int Event::soutput_params(char *buf) {
  char *pos = buf;
   for(map<int64_t, GenericParam<MPIParam>* >::const_iterator it = params.begin(); 
    it != params.end(); ++it) {
    //Print the param name
    pos += sprintf(pos, "%hhx ", FORMAT_OUT(it->first));
    pos += it->second->soutput_param(pos);
    pos += sprintf(pos, "\n");
  }
 
  return (pos-buf);
}

int Event::soutput_timing(char *buf) {
  char *pos = buf;
  for (int i = 0; i < TIMING_MAX; i++)
  {
    if (time_info[i].time_list){
      pos += timing_list_output (pos, time_info[i].time_list, i);
      //pos += sprintf(pos, "\n");
    }
    else
      pos += sprintf (pos, "TIME%d-0: %ld\n", i, time_info[i].time);
  }
  return (pos-buf);
}

void Event::setTime(long long int time, int index) {
  time_info[index].time = time;
}

long long int Event::getTime(int index) {
  return time_info[index].time;
}

TimeList ** Event::getTimeList(int index) {
  return &(time_info[index].time_list);
} 

int Event::event_packed_size(MPI_Comm comm) {
  //cout << "Event:event_packed_size " << event_id << " ";
  //printParams();
  //cout << endl;
  int size = 0;
  size += packed_size(2, MPI_INT, comm); //event_id, seq_num
  size += packed_size(sizeof(int64_t), MPI_BYTE, comm);//param_list
  for (int i = 0; i < TIMING_MAX; i++) {
    size += packed_size(sizeof(long long int), MPI_BYTE, comm);//time_info.time
    size += packed_size(1, MPI_INT, comm);//Flag to chk timelist presence.
    if (time_info[i].time_list) {
      size += timing_list_packed_size
        (time_info[i].time_list, comm);
    }
  }

  size += packed_size(2, MPI_INT, comm);//Flags for user_def_histo1,2
  if(user_def_histo_1)
    size += histogram_packed_size(user_def_histo_1, comm);
  if(user_def_histo_2)
    size += histogram_packed_size(user_def_histo_2, comm);

  size += packed_size(1, MPI_INT, comm);//To hold size of map
  size += packed_size(params.size(), MPI_INT, comm);//flag to represent Scalar or Vector
  for(map<int64_t, GenericParam<MPIParam>* >::const_iterator it = params.begin(); 
    it != params.end(); ++it) {
      size += packed_size(sizeof(int64_t), MPI_BYTE, comm);
      //cout << "Event " << event_id << " Param " << it->first << " address is " << it->second << endl;
      size += it->second->param_packed_size(comm);
  }
  return size;
}

void Event::unpack(void *buf, int bufsize, int *position, MPI_Comm comm) {
  /*event_id*/
  PMPI_Unpack(buf, bufsize, position, &event_id, 1, MPI_INT, comm);
  /*int64_t param_list*/
  param_list = 0;
  PMPI_Unpack(buf, bufsize, position, &param_list, sizeof(int64_t), MPI_BYTE, comm);
  /*int seq_num*/
  PMPI_Unpack(buf, bufsize, position, &seq_num, 1, MPI_INT, comm);
  /*struct time_data time_info[TIMING_MAX]*/
  for(int i=0; i < TIMING_MAX; i++) {
    time_info[i].time = 0;
    PMPI_Unpack(buf, bufsize, position, &time_info[i].time, 
                 sizeof(long long int), MPI_BYTE, comm);
    int is_list = 0;
    PMPI_Unpack(buf, bufsize, position, &is_list, 1, MPI_INT, comm);
    if(is_list)
      timing_list_unpack(&time_info[i].time_list, buf, bufsize, position, comm);
  }
  /*Histogram * user_def_histo_1;*/
  int is_user_histo1 = 0;
  user_def_histo_1 = NULL;
  PMPI_Unpack(buf, bufsize, position, &is_user_histo1, 1, MPI_INT, comm);
  if(is_user_histo1)
    user_def_histo_1 = histogram_unpack(buf, bufsize, position, comm);

  /*Histogram * user_def_histo_2;*/
  int is_user_histo2 = 0;
  user_def_histo_2 = NULL;
  PMPI_Unpack(buf, bufsize, position, &is_user_histo2, 1, MPI_INT, comm);
  if(is_user_histo2)
    user_def_histo_2 = histogram_unpack(buf, bufsize, position, comm);

  /*std::map<int64_t,GenericParam<MPIParam>* > params;*/
  int map_size = 0;
  PMPI_Unpack(buf, bufsize, position, &map_size, 1, MPI_INT, comm);
  
  int class_type = 0;
  for(int i=0; i < map_size; i++) {
    PMPI_Unpack(buf, bufsize, position, &class_type, 1, MPI_INT, comm);
    int64_t p = 0;
    PMPI_Unpack(buf, bufsize, position, &p, sizeof(int64_t), MPI_BYTE, comm);
    GenericParam<MPIParam> * mpiparam = NULL;
    if(class_type == SCALAR)
      mpiparam = new GenericScalarParam;
    else
      mpiparam = new VectorParam;
  
    mpiparam->unpack(buf, bufsize, position, comm);
    params[p] = mpiparam; //Add a new map entry
  }

}

void Event::pack(void *buf, int bufsize, int *position, MPI_Comm comm) {
  /*int event_id;*/
  PMPI_Pack(&event_id, 1, MPI_INT, buf, bufsize, position, comm);
  /*int64_t param_list;*/
  PMPI_Pack(&param_list, sizeof(int64_t), MPI_BYTE, buf, bufsize, position, comm);
  /*int seq_num;*/
  PMPI_Pack(&seq_num, 1, MPI_INT, buf, bufsize, position, comm);

  /*struct time_data time_info[TIMING_MAX];*/
  for (int i = 0; i < TIMING_MAX; i++) {
    PMPI_Pack(&time_info[i].time, sizeof(long long int), MPI_BYTE, buf, bufsize, position, comm);
    int is_list = 0;
    if(time_info[i].time_list)
      is_list = 1;

    PMPI_Pack(&is_list, 1, MPI_INT, buf, bufsize, position, comm);
    if (time_info[i].time_list) {
      timing_list_pack(time_info[i].time_list, buf, bufsize, position, comm);
    }
  }

  /*Histogram * user_def_histo_1;*/
  int is_user_histo1 = 0;
  if(user_def_histo_1)
    is_user_histo1 = 1;

  PMPI_Pack(&is_user_histo1, 1, MPI_INT, buf, bufsize, position, comm);
  if(is_user_histo1)
    histogram_pack(user_def_histo_1, buf, bufsize, position, comm);

  /*Histogram * user_def_histo_2;*/
  int is_user_histo2 = 0;
  if(user_def_histo_2)
    is_user_histo2 = 1;

  PMPI_Pack(&is_user_histo2, 1, MPI_INT, buf, bufsize, position, comm);
  if(is_user_histo2)
    histogram_pack(user_def_histo_2, buf, bufsize, position, comm);

  /*std::map<int64_t,GenericParam<MPIParam>* > params;*/
  int map_size = params.size();
  PMPI_Pack(&map_size, 1, MPI_INT, buf, bufsize, position, comm);
  
  for(map<int64_t, GenericParam<MPIParam>* >::iterator it = params.begin(); 
    it != params.end(); ++it) {
      int class_type = it->second->get_class_type();
      PMPI_Pack(&class_type, 1, MPI_INT, buf, bufsize, position, comm);
      int64_t p = it->first;
      PMPI_Pack(&p, sizeof(int64_t), MPI_BYTE, buf, bufsize, position, comm);
      it->second->pack(buf, bufsize, position, comm);
  }

}

void init_event(Event *event) {
  event->init();
}

void create_event(Event **event) {
  (*event) = new Event;
}

void add_scalar_param_int(Event *event, int64_t param_id, int val) {
  event->addParam(param_id, val);
}
  
void add_scalar_param_float(Event *event, int64_t param_id, float val) {
  event->addParam(param_id, val);
}

void add_scalar_param_double(Event *event, int64_t param_id, double val) {
  event->addParam(param_id, val);
}

void add_scalar_param_string(Event *event, int64_t param_id, char *val) {
  event->addParam(param_id, val);
}

void add_vector_param(Event *event, int64_t param_id, int *arr_param, int size) {
  event->addVectorParam(param_id, arr_param, size);
}

void event_set_param_list(Event *event, int op, int seqnum, int64_t value) {
  //cout << "Event: Inside event_set_param_list method" << endl;
  event->setEventId(op);
  event->setSeqNum(seqnum);
  event->setParamList(value);
}

void event_print(Event *event) {
  event->printParams();
}

void event_copy (Event **dest, Event **src) {
  if((*dest) == NULL) {
    (*dest) = new Event;
    if((*dest) == NULL)
	cout << "Event_copy:Error allocating memory for Event" << endl;
  }
  
  **dest = **src;/*Call assignment operator of event*/

  //cout << "Event: Event copy successful.. Printing values" << endl;
  (*dest)->printParams();
}

void free_event(Event **e) {
  delete (*e);
  *e = NULL;
}

int compare_event_params(Event *e1, Event *e2) {
  if(!e1 || !e2) return MFAIL_GEN;

  return e1->compareParams(e2);  
}

int soutput_event_params(char *buf, Event *e) {
  return e->soutput_params(buf);
}

int soutput_event_timing(char *buf, Event *e) {
  return e->soutput_timing(buf);
}

void set_event_time(Event *e, long long int time, int index) {
  e->setTime(time, index);
}

void event_time_merge(int index, Event *dest, const stack_sig *dest_sig, 
  int dest_rank, Event *src, const stack_sig* src_sig, 
    int src_rank, int B) {
  if(dest && src) {
    timing_list_merge(dest->getTimeList(index),dest->getTime(index),dest_sig,
      dest_rank, src->getTimeList(index), src->getTime(index), src_sig, src_rank, B);  
  }
}
int event_merge_vectors(Event *merge, Event *target, int64_t param){
  return param_merge_vectors(merge->getParam(param), target->getParam(param));
}

void event_vector_reset_undo(Event *e, int64_t param){
  if(e)
    param_vector_reset_undo(e->getParam(param));
}

  void event_vector_undo_addition(Event *e, int64_t param){
  if(e)
    param_vector_undo_addition(e->getParam(param));
}

int event_packed_size(Event *e, MPI_Comm comm) {
  int size = 0;
  size += packed_size(1, MPI_INT, comm);
  if(e)
    size += e->event_packed_size(comm);
  return size;
}

void event_pack(Event *e, void *buf, int bufsize, int *position, MPI_Comm comm){
  int is_event = 0;
  if(e)
    is_event = 1;
  PMPI_Pack(&is_event, 1, MPI_INT, buf, bufsize, position, comm);
  if(is_event)
    e->pack(buf, bufsize, position, comm);
}

void event_unpack(Event **e, void *buf, int bufsize, int *position, MPI_Comm comm) {
  int is_event = 0;
  *e = NULL;
  PMPI_Unpack(buf, bufsize, position, &is_event, 1, MPI_INT, comm);
  if(is_event) {
    *e = new Event;
    (*e)->unpack(buf, bufsize, position, comm);
  }
}

void event_merge_param_value_list(Event *master, Event *slave, int64_t param,
                                  int *m_list, int *s_list) {
  if(master && slave) {
    master->getParam(param)->param_merge_value_list( 
      (void *)slave->getParam(param), m_list, s_list);
  }
}

void event_get_param (Event *event, int64_t param_id, void *param) {
  GenericParam<MPIParam> * p = event->getParam(param_id);
  if(p->get_class_type() == SCALAR) {
    MPIParam mp = p->getMPIParam();
    mp.getParamValue (param);
  }
  else {
    vector<MPIParam> * vec = (vector<MPIParam>*)p->getValueList();
    int * arr = (int *)param;
    int i = 0;
    for(vector<MPIParam>::iterator it = vec->begin();
        it != vec->end(); ++it) {
      int val;
      it->getParamValue((void *)&val);
      arr[i] = val;
      i++;
    }
  }
}
