/*
 * Filename: Intseq.C
 * Author:   ratn1
 * Created:  Fri Jul 13 15:28:28 2007
 * $Id: Intseq.C,v 1.2 2008/01/30 01:51:41 pratn Exp $
 */
#include <iostream>
#include <assert.h>
#include <mpi.h>
#include <inttypes.h>
#include "Intseq.h"
#include "TimeList.h"
#include "mpi_util.h"

#ifndef  ARRAY_SIZE
# define ARRAY_SIZE(a)  sizeof (a)/sizeof(a[0])
#endif

# define NUM_IGNORE_BITS (4)

# define set_intptr_datatype(datatype)  do {\
  if (sizeof (int) == sizeof(intptr_t)) { \
    datatype = MPI_INT; \
  } else if (sizeof (long int) == sizeof (intptr_t)) {  \
    datatype = MPI_LONG; /* not MPI_LONG_INT !*/ \
  } else if (sizeof (long long int) == sizeof (intptr_t)) { \
    datatype = MPI_LONG_LONG_INT /* same as MPI_LONG_LONG */; \
  } \
}while (0)


Intseq::Intseq ()  
  : k(NUM_IGNORE_BITS), vect(NULL)
{
  init ();
}

Intseq::Intseq (long int dest_time, int ignore,
		const stack_sig *sig, int my_rank, int min_rank, int max_rank)

  : k(NUM_IGNORE_BITS), vect(NULL)
{
  init ();
  add (dest_time, my_rank);
  sig_assign(&signature, sig);
}

void Intseq::init ()
{
  k_mask = ~((1 << (k)) - 1);
  sig_init(&signature);
}

Intseq::~Intseq () {
  sig_destroy(&signature);
}

short most_significant (int a)
{
  int i =0;

  if (a == 0)
    return 0;

  for (i = sizeof (int) * 8 - 1; i >= 0; i--)
  {
    if (( 1 << i ) & a)
      return i;
  } 
  assert (0);
  return 0;
}

value_t bin2gray (value_t b)
{
  return (b ^ (b>>1));
}

int Intseq::add (value_t value, int rank)
{
  value_t v = (value & k_mask);
  //v = bin2gray (v);
  int rem = 1<<k;

  // round off
  if ((value % rem) >= rem/2) {
    value += rem;
  } 
  
  if (vect) {
    int size = vect->size ();
    value_t cii = vect->back ();
    value_t c2 = (*vect)[size-1];

    bool match =  ((*vect)[size-1] == v) || \
      (size > 2 && (*vect)[size-2] == v && (*vect)[size-1] == (*vect)[size-3])\
    ||(size > 4 && (*vect)[size-3] == v && (*vect)[size-1] == (*vect)[size-4] && (*vect)[size-2] == (*vect)[size-5]);
    
    if (match) {
      return -1;
    } else {
      vect->push_back (v);
    }
  } else {
    vect = new vector<value_t>();
    vect->push_back (v);
  }
  return 0;
}

int Intseq::output (char * buf)
{
  vector<value_t>::const_iterator cii;

  char * pos = buf;
  pos += sprintf(pos, "(");
  pos += sig_soutput(pos, &signature);
  pos += sprintf(pos, ")");

  pos += sprintf (pos, "[%ld] ", vect->size());
  for (cii = vect->begin(); cii != vect->end(); cii++)
  {
    pos += sprintf (pos, "%"PRIvalINT" ", *cii);
  }
  pos += sprintf (pos, "\n");

  return (pos - buf);
}

int Intseq::get_packed_size (MPI_Comm comm) const
{
  int size = 0;
  size += packed_size   (1,        MPI_INT,         comm);  // k
  size += packed_size   (1,        VALUE_DATATYPE,  comm);  // k_mask
  size += packed_size   (1,        MPI_INT,         comm);  // vector size
  size += packed_size   (vect->size(), VALUE_DATATYPE,  comm);  // vector
  size += sig_packed_size(&signature, comm);
  return size;
}

void Intseq::pack (void* buf, int bufsize, int *position, MPI_Comm comm)
{
  int size = vect->size();
  vector<value_t>::const_iterator vi;
  value_t v;

  PMPI_Pack   (&k,       1,     MPI_INT,        buf, bufsize, position, comm);
  PMPI_Pack   (&k_mask,  1,     VALUE_DATATYPE, buf, bufsize, position, comm);
  PMPI_Pack   (&size,    1,     MPI_INT,        buf, bufsize, position, comm);

  int i = 0;
  for (vi =  vect->begin(); vi !=vect->end(); vi++) {
    v = *vi;
    PMPI_Pack (&v,       1,     VALUE_DATATYPE, buf, bufsize, position, comm); 
    i++;
  }

  sig_pack(&signature, buf, bufsize, position, comm);
}

Intseq * Intseq::unpack (void * buf, int bufsize, int* position, MPI_Comm
    comm)
{
  Intseq * seq = new Intseq ();
  int size = 0;
  value_t v;

  PMPI_Unpack   (buf, bufsize, position, &seq->k,       1,   MPI_INT,        comm);
  PMPI_Unpack   (buf, bufsize, position, &seq->k_mask,  1,   VALUE_DATATYPE, comm);
  PMPI_Unpack   (buf, bufsize, position, &size,         1,   MPI_INT,        comm);

  for (int i = 0; i < size; i++) {
    PMPI_Unpack (buf, bufsize, position, &v,            1,   VALUE_DATATYPE, comm);
    seq->add (v, -1); 
  }

  sig_unpack(&seq->signature, buf, bufsize, position, comm);
  return seq;
}

void Intseq::merge (const Intseq *other)
{
  vector<value_t>::const_iterator cii;

  for (cii = other->vect->begin(); cii != other->vect->end(); cii++)
  {
    //FIXME: get rank here
    add (*cii, 0);
  }
}


extern "C" void * intseq_input (char * buf) {
  return (void*)NULL;
}
/* vim: set tw=80 ts=2 sw=2 expandtab: */
