#include "Param.h"

int MPIParam::get_packed_size(MPI_Comm comm) {
  int size = 0;
  size += packed_size(1, MPI_INT, comm); //identifier
  if(identifier == INT)
    size += packed_size(1, MPI_INT, comm);
  else if(identifier == FLOAT)
    size += packed_size(1, MPI_FLOAT, comm);
  else if(identifier == DOUBLE)
    size += packed_size(1, MPI_DOUBLE, comm);
  else {
    size += packed_size(1, MPI_INT, comm);//Length of string
    if(uparam.svalue) {
      size += packed_size(strlen(uparam.svalue), MPI_CHAR, comm);
    }
  }
  //std::cout << "MPIParam:identifier: " << identifier << " get_packed_size: " << size << std::endl;
  return size;
}

void MPIParam::pack(void *buf, int bufsize, int *position, MPI_Comm comm) {
  PMPI_Pack(&identifier, 1, MPI_INT, buf, bufsize, position, comm);
  if(identifier == INT)
    PMPI_Pack(&uparam.ivalue, 1, MPI_INT, buf, bufsize, position, comm);
  else if(identifier == FLOAT)
    PMPI_Pack(&uparam.fvalue, 1, MPI_FLOAT, buf, bufsize, position, comm);
  else if(identifier == DOUBLE)
    PMPI_Pack(&uparam.dvalue, 1, MPI_DOUBLE, buf, bufsize, position, comm);
  else {
    int slen = 0;
    if(uparam.svalue)
      slen = strlen(uparam.svalue)+1;

    PMPI_Pack(&slen, 1, MPI_INT, buf, bufsize, position, comm);
    if(uparam.svalue) {
      PMPI_Pack(uparam.svalue, slen, MPI_CHAR, buf, bufsize, position, comm);
    }
  }
}

void MPIParam::unpack(void *buf, int bufsize, int *position, MPI_Comm comm) {
  int iden = 0;
  PMPI_Unpack(buf, bufsize, position, &iden, 1, MPI_INT, comm);
  if(iden == INT) {
    identifier = INT;
    PMPI_Unpack(buf, bufsize, position, &uparam.ivalue, 1, MPI_INT, comm);
  }
  else if(iden == FLOAT) {
    identifier = FLOAT;
    PMPI_Unpack(buf, bufsize, position, &(uparam.fvalue), 1, MPI_FLOAT, comm);
  }
  else if(iden == DOUBLE) {
    identifier = DOUBLE;
    PMPI_Unpack(buf, bufsize, position, &(uparam.dvalue), 1, MPI_DOUBLE, comm);
  }
  else {
    int slen = 0;
    identifier = STRING;
    PMPI_Unpack(buf, bufsize, position, &slen, 1, MPI_INT, comm);
    if(slen > 0) {
      uparam.svalue = (char*)malloc(slen*sizeof(char));
      PMPI_Unpack(buf, bufsize, position, &uparam.svalue, slen, MPI_CHAR, comm);
    }
  }
}

void init_mpiparam_int(MPIParam **param, int val){
  (*param) = new MPIParam(val);
}

void init_mpiparam_float(MPIParam **param, float val){
  
  (*param) = new MPIParam(val);
}

void init_mpiparam_double (MPIParam **param, double val){

  (*param) = new MPIParam(val);
}
void init_mpiparam_string (MPIParam **param, char* val){

  (*param) = new MPIParam(val);
}

int compareMPIParam (MPIParam * m1, MPIParam * m2, int * diff_percent) {

  int identifier = m1->getIdentifier();
  if(identifier != m2->getIdentifier()) {
    return 0;
  }
  *diff_percent = 100;
  if(identifier == INT) {
    int val1, val2;
    val1 = m1->getValue(val1);
    val2 = m2->getValue(val2);
    if(val1 == val2) {
      *diff_percent = 0;
      m1->inc_counter();
      return 1;
    }
    else if(val2 && (val1 > val2)) {
      *diff_percent = (val1 - val2)*100/val2;
    }
    else if(val1){
      *diff_percent = (val2 - val1)*100/val1;
    }
  }
  else if(identifier == FLOAT) {
    float val1, val2;
    val1 = m1->getValue(val1);
    val2 = m2->getValue(val2); 
    if(val1 == val2) {
      *diff_percent = 0;
      m1->inc_counter();
      return 1;
    }
    else if(val2 && (val1 > val2)) {
      *diff_percent = (val1 - val2)*100/val2;
    }
    else if(val1) {
      *diff_percent = (val2 - val1)*100/val1;
    }
  }
  else if(identifier == DOUBLE) {
    double val1, val2;
    val1 = m1->getValue(val1);
    val2 = m2->getValue(val2); 
    if(val1 == val2) {
      *diff_percent = 0;
      m1->inc_counter();
      return 1;
    }
    else if(val2 && (val1 > val2)) {
      *diff_percent = (val1 - val2)*100/val2;
    }
    else if(val1){
      *diff_percent = (val2 - val1)*100/val1;
    }
  }
  else if(identifier == STRING) {
    char *val1;
    char *val2;
    val1 = m1->getValue(val1);
    val2 = m2->getValue(val2);
    if(!strcmp(val1, val2))
	return 1;
  }

  return 0;
}
