#include "RList.h"
#include <string.h>
#include <iostream>
#include <memory>
using namespace std;

RList::RList() : rlist(NULL) { }

RList::RList(int rank) { 
  create_recursive_list(rank, &rlist);
}

RList::RList(int *rl) : rlist(rl) { }

extern "C" int packed_size(int count, MPI_Datatype type, MPI_Comm comm);


void RList::merge(const RankSet& olist) {
  const RList& other = dynamic_cast<const RList&>(olist);


  char me[4096];
  char him[4096];
  if (rlist && other.rlist) {

    soutput(me);
    other.soutput(him);
    cerr << "merge " << me << " + " << him << endl;

    merge_recursive_lists(&rlist, other.rlist);

    soutput(me);
    other.soutput(him);
    cerr << " ==> " << me << " + " << him << endl;

  } else if (other.rlist) {
    soutput(me);
    other.soutput(him);
    cerr << "dup " <<  me << " + " << him << endl;

    dup_recursive_list(&rlist, other.rlist);

    soutput(me);
    other.soutput(him);
    cerr << " ==> " << me << " + " << him << endl;
  }
}

void RList::merge (const int *rl) {
  assert (rlist && rl);
  merge_recursive_lists(&rlist, (int*)rl);
}

int RList::packed_size(MPI_Comm comm) const {
  int size = ::packed_size(1, MPI_INT, comm);
  int rlist_size = size_recursive_list(rlist);
  if (rlist_size) {
    size += ::packed_size(rlist_size, MPI_INT, comm);
  }
  return size;
}


void RList::pack(void *buf, int bufsize, int *position, MPI_Comm comm) const {
  int rlist_size = size_recursive_list(rlist);
  PMPI_Pack(&rlist_size, 1, MPI_INT, buf, bufsize, position, comm);
  if (rlist_size) {
    PMPI_Pack(rlist,  rlist_size, MPI_INT, buf, bufsize, position, comm);
  }
}


/** Unpacks an RList from a buffer. */
RList *RList::unpack(void *buf, int bufsize, int *position, MPI_Comm comm) {
  int rlist_size;
  auto_ptr<RList> recvd(new RList());
  
  PMPI_Unpack(buf, bufsize, position, &rlist_size, 1, MPI_INT, comm);
  if (rlist_size) {
    recvd->rlist = alloc_rlist(rlist_size);
    PMPI_Unpack(buf, bufsize, position, recvd->rlist, rlist_size, MPI_INT, comm);
  } else {
    recvd->rlist = NULL;
  }
  return recvd.release();
}


RList *RList::input(const char *buf) {
  const char *list_start = buf;
  assert(list_start);

  // skip whitespace
  while (*list_start == ' ') list_start++;  
  assert(*list_start == '[');

  const char *list_end = strchr(list_start + 1, ']');
  assert(list_end);
  
  int *list;
  if (list_end == list_start + 1) {
    list = NULL; // nothing in the list.
  } else {
    list = parse_recursive_list(list_start + 1);
  }

  return new RList(list);
}


int RList::soutput(char *buf) const {
  char *pos = buf;
  pos += sprintf(pos, "[");
  pos += soutput_range_list(rlist, pos);
  pos += sprintf(pos, "]");
  return (pos - buf);
}


RList *RList::clone() const {
  RList *retval = new RList();
  dup_recursive_list(&retval->rlist, rlist);
  return retval;
}

RList *RList::clone (int *rl) {
  RList *retval = new RList();
  dup_recursive_list(&retval->rlist, rl);
  return retval;
}
