#include "ScalarParam.h"
#include <iostream>

using std::cout; using std::endl;
using std::vector;

void GenericScalarParam::add_to_rlist(MPIParam val) {
  int end_point;
  end_point = val.getValue(end_point);
  //end_point = (end_point+my_rank)%my_size;
  if(!end_point_list) {
    create_recursive_list(end_point, &end_point_list);
  }
  else {
    if(!member_rank(end_point, end_point_list)) {
      int * new_list = NULL;
      create_recursive_list(end_point, &new_list);
      merge_recursive_lists(&end_point_list, new_list);
      destroy_recursive_list(new_list);
    }
  }
}

//template <class MPIParam>
GenericScalarParam::GenericScalarParam(MPIParam val, int node = 0):histo(0),vec(0),end_point_list(0) {
  value = val;
  track_rlist = node;
  if(track_rlist == 1) {
    add_to_rlist(value);
  }
}

//template <class MPIParam>
GenericScalarParam& GenericScalarParam::operator= (const GenericScalarParam&) {
  //cout << "GenericScalarParam: Inside =operator implementation" << endl;
}

//template <class MPIParam>
GenericScalarParam::GenericScalarParam(const GenericScalarParam&) {
  //cout << "GenericScalarParam: Inside copy constructor" << endl;
}

//template <class MPIParam>
GenericScalarParam::~GenericScalarParam() {
  if (histo)
    delete histo;
  if(vec)
    vector_free(&vec);

  if(end_point_list)
    destroy_recursive_list(end_point_list);

  for(vector<ScalarParamList*>::iterator it = value_list.begin(); 
           it != value_list.end(); ++it) {
    delete (*it);
  }
  value_list.clear();
}

//template <class MPIParam>
void GenericScalarParam::add (MPIParam val) {
  //cout << "GenericScalarParam: Inside add function" << endl;
  value = val;
  if(track_rlist == 1)
    add_to_rlist(value);
  //value.print();
}

//template <class MPIParam>
void GenericScalarParam::add_histo (MPIParam value) {
  //cout << "GenericScalarParam: Inside add_histo function" << endl;
  double val = 0;//value;
  if(!histo) {
    histo = new Histogram(val);
    return;
  }

  histo->add(val);
}

int GenericScalarParam::compare (GenericParam<MPIParam>& m1) {
  //cout << "GenericScalarParam: Inside compare function" << endl;
  MPIParam slave_mp = m1.getMPIParam();
  MPIParam master_mp = value;
  
  if(histo) {

    Histogram * slave = (Histogram*)m1.getHisto();
    if(slave) {
      histo->skyline_merge(slave);
      int ** slave_end_list = m1.get_end_point_list();
      if(track_rlist)
        merge_recursive_lists(&end_point_list, *slave_end_list);
    }
    else {
      int val= 0;
      slave_mp.getParamValue((void *)&val);
      histogram_add(histo, val );
      if(track_rlist)
        add_to_rlist(slave_mp);
    }
    return 1;
  }

  /*Check if value_list is populated,
    If yes, then return false,
        caller should take care of value_list*/
  vector<ScalarParamList*> *val_list = (vector<ScalarParamList*>*)m1.getValueList();
  if(value_list.size() > 0 || val_list->size() > 0)
    return 0;

  /*If any of the vector is populated, 
    then call vector_match for comparison.*/
  cvector ** m1_vec = (cvector**)m1.getVector();
  if(vec || *m1_vec) {
    return vector_match( vec, *m1_vec );
  }
  int diff_percent = 0;
 #ifndef FEATURE_COMPRESS_HISTO
  return compareMPIParam(&value, &slave_mp, &diff_percent);
 #else
  /*Else compare just the params.*/
  if( !compareMPIParam(&value, &slave_mp, &diff_percent) ){
    if(diff_percent <= USER_DEFINED_ERROR) {
      int master_val = 0, slave_val = 0;
      //master_mp.getParamValue(&master_val);
      value.getParamValue(&master_val);
      slave_mp.getParamValue(&slave_val);
      if(!histo){
        //histogram_init(&histo, 5);
        histo = new Histogram(5, 1, 10, 30);
        histogram_add(histo, master_val);
        if(track_rlist)
          add_to_rlist(value);
      }
      histogram_add(histo, slave_val);
      if(track_rlist)
        add_to_rlist(slave_mp);
      return 1;
    }
  }
  else {
    return 1;
  }
  return 0;
 #endif
}

void GenericScalarParam::printValue() {
 /* cout << "GenericScalarParam Val " << endl;
  value.print();
  if(vec) {
    int size = vector_packed_size(vec, MPI_COMM_WORLD);
    char * buf = (char *)malloc(size);
    vector_output(buf, vec);
    printf("CVector %s\n", buf);
    free(buf);
  }
  if(value_list.size() > 0)
    cout << "Value list is not null\n" << endl;*/
}

int GenericScalarParam::soutput_param(char *buf) {
  char * pos = buf;
  if(histo) {
    if(end_point_list) {
      pos += sprintf(pos, " [");
      pos += soutput_range_list(end_point_list, pos);
      pos += sprintf(pos, "] ");
    }
    pos += histo->output(pos);
    return (pos-buf);
  }

  if(value_list.size() > 0){
   // cout << "Value List is not null" << endl;
    for(vector<ScalarParamList *>::iterator it = value_list.begin();
       it != value_list.end(); ++it) {
      pos += (*it)->output(pos);
      pos += sprintf(pos, " ");
    }
    return (pos - buf);
  }
  else if(vec)
    return vector_output(pos, vec);
  else
    return value.soutput_value(pos);
}

int GenericScalarParam::param_packed_size(MPI_Comm comm) {
  int size=0;
  size += value.get_packed_size(comm);
  size += packed_size(1, MPI_INT, comm); //track_rlist
  size += recursive_list_packed_size(end_point_list, comm);
  //Histo
  size += packed_size(1, MPI_INT, comm); //Flag to check if histo is present/not
  if(histo)
    size += histo->get_packed_size(comm);

  //Vector of ScalarParamList
  size += packed_size(1, MPI_INT, comm); //Size of ScalarParamList vector
  if(value_list.size() > 0){
    for(vector<ScalarParamList *>::iterator it = value_list.begin();
       it != value_list.end(); ++it) {
      size += (*it)->get_packed_size(comm);
    }
  }

  //cvector vec  
  size += vector_packed_size(vec, comm);//This internally checks for null

  //cout << "GenericScalarParam:param_packed_size " << size << endl;
  return size;
}

void GenericScalarParam::pack(void *buf, int bufsize, int *position, MPI_Comm comm){
  /*MPIParam value;*/
  value.pack(buf, bufsize, position, comm);
  /*int track_rlist;*/
  PMPI_Pack(&track_rlist, 1, MPI_INT, buf, bufsize, position, comm);
  if(track_rlist)
    recursive_list_pack(end_point_list, buf, bufsize, position, comm);

  /*Histogram * histo;*/
  int is_histo = 0;
  if(histo)
    is_histo = 1;
  PMPI_Pack(&is_histo, 1, MPI_INT, buf, bufsize, position, comm);
  if(histo)
    histo->pack(buf, bufsize, position, comm);

  /*std::vector<ScalarParamList> value_list;*/
  int size = value_list.size();
  PMPI_Pack(&size, 1, MPI_INT, buf, bufsize, position, comm);
  if(size > 0){
    for(vector<ScalarParamList *>::iterator it = value_list.begin();
       it != value_list.end(); ++it) {
      (*it)->pack(buf, bufsize, position, comm);
    }
  }

  /*cvector * vec;*/
  vector_pack(vec, buf, bufsize, position, comm);
}

void GenericScalarParam::unpack(void *buf, int bufsize, int *position, MPI_Comm comm){
  /*MPIParam value*/
  value.unpack(buf, bufsize, position, comm);

  PMPI_Unpack(buf, bufsize, position, &track_rlist, 1, MPI_INT, comm);//track_rlist
  if(track_rlist)
    end_point_list = recursive_list_unpack(buf, bufsize, position, comm);

  /*Histogram *histo */
  int is_histo=0;
  histo = NULL;
  PMPI_Unpack(buf, bufsize, position, &is_histo, 1, MPI_INT, comm);
  if(is_histo)
    histo = histogram_unpack(buf, bufsize, position, comm);

  /*std::vector<ScalarParamList> value_list;*/
  int size = 0;
  PMPI_Unpack(buf, bufsize, position, &size, 1, MPI_INT, comm);
  if(size > 0){
    for(int i=0; i<size; i++) {
      ScalarParamList * slist = new ScalarParamList; 
      slist->unpack(buf, bufsize, position, comm);
      value_list.push_back(slist);
    }
  }
  
  /*cvector *vec*/
  vec = vector_unpack(buf, bufsize, position, comm);
}

void init_scalar_param_int (Scalar **scalar, int val){

}

void init_scalar_param_float (Scalar **scalar, float val){

}

void init_scalar_param_double (Scalar **scalar, double val){

}

void init_scalar_param_string (Scalar **scalar, char* val){

}


int param_merge_vectors(GenericParam<MPIParam>* merge, GenericParam<MPIParam>* target){
  cvector** merge_vector = (cvector**)merge->getVector();
  cvector** target_vector = (cvector**)target->getVector();

  /*If any of vector is null, then 
  incompatible comparison like Scalar vs VectorParam*/
  if(!merge_vector || !target_vector)
	return 0;

  int merge_val, target_val;

  merge_val = (merge->getMPIParam()).getValue(merge_val);
  target_val = (target->getMPIParam()).getValue(target_val);
  long merge_counter = (merge->getMPIParam()).getCounter();

  if(NULL == *merge_vector && NULL == *target_vector) {
    vector_init (merge_vector, merge_val);
    for(int i = 0; i < merge_counter; i++)
      vector_push_back(*merge_vector, merge_val);
    vector_undo_info (*merge_vector, Undo_pos);
    vector_push_back (*merge_vector, target_val);
  } else if (*merge_vector && NULL == *target_vector) {
    if (vector_match_first (*merge_vector, target_val)) {
      return 0;
    } else {
      vector_undo_info (*merge_vector, Undo_pos);
      vector_push_back (*merge_vector, target_val);
    }
  } else if (merge_vector && target_vector) {
    if (vector_match (*merge_vector, *target_vector)) {
      vector_undo_info (*merge_vector, Undo_iters);
      vector_increment_count (*merge_vector);
    } else {
      vector_undo_info (*merge_vector, Undo_pos);
      vector_concat (*merge_vector, *target_vector);
    }
  } else {
    assert (0);
  }
  return 1;
}

void param_vector_reset_undo(GenericParam<MPIParam>* p){
  if(!p) return;

  cvector** vec = (cvector**)p->getVector();
  if(!(*vec))
	return;

  vector_reset_undo_info(*vec);
}

void param_vector_undo_addition(GenericParam<MPIParam>* p){
  if(!p) return;
  cvector** vec = (cvector**)p->getVector();
  if(!(*vec))
	return;

  int val, backup;
  val = (p->getMPIParam()).getValue(val); 
  backup = val;
  vector_undo_addition(vec, &val);
  if(val != backup){
    MPIParam mparam(val);
    p->add(mparam);
  }
}

void GenericScalarParam::add_to_value_list( ScalarParamList* s) {
  /*1.Iterate thro the vector
    2.If any vector matches, then update rlist and free s
    3.Else add s into the vector*/
  typedef vector<ScalarParamList*> vec_scalar;
  for(vec_scalar::iterator it = value_list.begin(); it != value_list.end(); ++it) {
    if((*it)->match(s)) {
      //merge rlist
      (*it)->merge_rlist( s->get_rlist() );
      return;
    }
  }
  value_list.push_back(s);
}

void GenericScalarParam::merge_value_list(vector<ScalarParamList*>* s) {
  vector<ScalarParamList*>::const_iterator vii;
  for(vii = s->begin(); vii != s->end(); ++vii) {
    add_to_value_list( *vii );
  }
}

void GenericScalarParam::param_merge_value_list(void *slaveParam, int *m_rlist, int *s_rlist) {
  int *master_arr=NULL;
  int *slave_arr = NULL;
  int master_size, slave_size;
  vector<MPIParam> master_vec, slave_vec;
  GenericParam<MPIParam> *p = (GenericParam<MPIParam> *)slaveParam;

  /*Move params to local master_vec and slave_vec.
    These vectors will be pushed to value_list.
   */
  if(vec) {
    master_arr = vector_to_array(vec, &master_size);
    for(int i=1; i < master_size; i++) {
      master_vec.push_back( MPIParam(master_arr[i]) );
    }
  }
  else {
    master_vec.push_back(value);
  }

  cvector ** slave_cvector = (cvector**) p->getVector();
  if(*slave_cvector) {
    slave_arr = vector_to_array(*slave_cvector, &slave_size);
    for(int i=1; i< slave_size; i++)
      slave_vec.push_back( MPIParam(slave_arr[i]) );
  }
  else {
    slave_vec.push_back( p->getMPIParam() );
  }

  /*Get value_list from master and slave*/
  vector<ScalarParamList*> * slave_val_list = 
                    (vector<ScalarParamList*>*)p->getValueList();
  vector<ScalarParamList*> * master_val_list = &value_list;

  /*Check both value_list and merge into master */  
  if(master_val_list->size() == 0 && slave_val_list->size() == 0) {
    ScalarParamList * s1 = new ScalarParamList(master_vec, m_rlist);
    ScalarParamList * s2 = new ScalarParamList(slave_vec, s_rlist);
    add_to_value_list(s1);
    add_to_value_list(s2);
  }
  else if(master_val_list->size()>0 && slave_val_list->size() == 0) {
    ScalarParamList *s = new ScalarParamList(slave_vec, s_rlist);
    add_to_value_list(s);
  }
  else if(master_val_list->size()==0 && slave_val_list->size() > 0) {
    ScalarParamList *s = new ScalarParamList(master_vec, m_rlist);
    add_to_value_list(s);
    merge_value_list(slave_val_list);
  }
  else { //both exist
    merge_value_list(slave_val_list);
  }

  //Clear the slave list
  slave_val_list->clear();

  //Free integer arrays created for the purpose of transfer.
  if(master_arr)
    free(master_arr);
  if(slave_arr)
    free(slave_arr);

}
