
#include "ScalarParamList.h"
using std::vector;

ScalarParamList::ScalarParamList(vector<MPIParam> v, int *rlist) {
  values = v;
  rank_list = RList::clone(rlist);
}

void ScalarParamList::merge_rlist(RList *r) {
  rank_list->merge( r->get_rlist() );
}

int ScalarParamList::match(ScalarParamList *s) {
  if(values.size() != s->values.size())
    return 0;

  typedef vector<MPIParam>::const_iterator vec_it;
  vec_it it, it1;
  int diff = 0;
  for(it = values.begin(), it1 = s->values.begin();
      it!=values.end(), it1 != s->values.end();
      ++it, ++it1 ) {
    MPIParam m = *it;
    MPIParam m1 = *it1;
    if( !compareMPIParam(&m, &m1, &diff) ) {
      return 0;
    }
  }
  return 1;
}

ScalarParamList::~ScalarParamList() {
  delete rank_list;
}

int ScalarParamList::get_packed_size(MPI_Comm comm){
  int size = 0;
  size += packed_size(2, MPI_INT, comm);//vector size and rank list flag.
  if(values.size() > 0) {
    //Iterate MPIParam vector and collect size
    for (vector<MPIParam>::iterator it = values.begin();
         it != values.end(); ++it){
      size += it->get_packed_size(comm); 
    }
    //Collect size from RList
    if(rank_list)
      size += rank_list->packed_size(comm);
  }
  return size;
}

void ScalarParamList::pack(void *buf, int bufsize, int *position, MPI_Comm comm) {
  //Pack vector<MPIParam> values
  int size = values.size();
  PMPI_Pack(&size, 1, MPI_INT, buf, bufsize, position, comm);
  for(vector<MPIParam>::iterator it = values.begin();
      it != values.end(); ++it) {
    it->pack(buf, bufsize, position, comm);
  }

  //Pack RList rank_list
  int is_rlist = 0;
  if(rank_list)
    is_rlist = 1;

  PMPI_Pack(&is_rlist, 1, MPI_INT, buf, bufsize, position, comm);
  if(is_rlist)
    rank_list->pack(buf, bufsize, position, comm);

}

void ScalarParamList::unpack(void *buf, int bufsize, int *position, MPI_Comm comm) {

  int size = 0;
  PMPI_Unpack(buf, bufsize, position, &size, 1, MPI_INT, comm);

  if(size > 0) {
    for(int i=0; i < size; i++){
      MPIParam param;
      param.unpack(buf, bufsize, position, comm);
      values.push_back(param);
    }
  }

  int is_rlist = 0;
  PMPI_Unpack(buf, bufsize, position, &is_rlist, 1, MPI_INT, comm);
  if(is_rlist)
    rank_list = RList::unpack(buf, bufsize, position, comm);
}

int ScalarParamList::output(char * buf) {
  char * pos = buf;
  //pos += sprintf(pos, "( %d ", values.size());
  pos += sprintf(pos, "( 1 ");
  
  for (vector<MPIParam>::iterator it = values.begin();
       it != values.end(); ++it){
    pos += it->soutput_value(pos);
    pos += sprintf(pos, " ");
  }
  pos += sprintf(pos, ")");
  pos += rank_list->soutput(pos);
  return (pos - buf);
}
