#include "Timing.h"

#ifdef HAVE_RTS_GETTIMEBASE
// -------------------------------------------------------- //
// Timing code for BlueGene/L
// -------------------------------------------------------- //
#include <rts.h>

// this will return number of nanoseconds in a single BGL cycle
// use for converting from cycle units returned by rts_gettimebase
// to nanoseconds.
static double get_ns_per_cycle() {
  BGLPersonality personality;
  if (rts_get_personality(&personality, sizeof(personality)) != 0)
    return 0;
  return 1.0e9/((double) personality.clockHz);
}

// returns time in nanoseconds.
long long get_cur_time_ns () {
  static double ns_per_cycle = get_ns_per_cycle();
  return (ns_per_cycle * rts_get_timebase());
}



#elif defined(USE_LINUX_TIMERS)
// -------------------------------------------------------- //
// Timing code using Linux hires timers.
// -------------------------------------------------------- //

#include <ctime>
#include <sys/time.h>

unsigned long long get_cur_time_ns() {
  struct timespec ts;
  //long long int sec,nsec;
  clock_gettime(CLOCK_MONOTONIC, &ts);
  return ((unsigned long long)ts.tv_sec * 1000000000 + ts.tv_nsec);
}
/*
static inline void rdtsc(unsigned int *upper, unsigned int *lower) {
  asm volatile("rdtsc\n" : "=a"(*lower), "=d"(*upper));
}

unsigned long long get_cur_time_ns() {
  unsigned int low, high;
  unsigned long long time;
  rdtsc(&high, &low);

  time = 0xFFFFFFFFFFFFFFFF & high;
  time = time << 32;
  time = time+low;

  //printf("%u %u\n", high, low);
  //printf("%llu \n", time);

  return time;
}*/

#else
// -------------------------------------------------------- //
// Generic timing code using gettimeofday.
// -------------------------------------------------------- //

#include <ctime>
#include <sys/time.h>

long long get_cur_time_ns() {
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return tv.tv_sec * 1000000000 + tv.tv_usec * 1000;
}

#endif // types of timers
