/*
 * Filename: VParam.C
 * Author:   ratn1
 * Created:  Wed Oct 31 14:46:44 2007
 * $Id: VParam.C,v 1.6 2009/01/13 23:26:16 gamlin Exp $
 */
#include <iostream>
#include <utility>
#include <assert.h>
#include "VParam.h"
#include "opstruct.h"
#include "mpi_util.h"


#define VPTRACE()
//#define VPTRACE() printf ("[%d]"__FILE__" %d\n", get_my_rank(),__LINE__); 

RLParam::RLParam () : param (NO_FIELDS) , vlist () {}
RLParam::RLParam (int64_t p) : param (p) , vlist () {}

int RLParam::add (int64_t val, int * rlist)
{
  value_list * v = 0;
  list<value_list*>::const_iterator vii;
  VPTRACE ();

  for (vii = vlist.begin(); vii != vlist.end(); vii++) {
    if (((ScalarParam*)(*vii)->value)->getValue() == val) {
      v = (*vii);
      break;
    }
  }
  
  if (!v) {
    v = new value_list;
    v->type  = Param_scalar; 
    v->value = new ScalarParam(val);
    v->rlist = RList::clone (rlist);
    vlist.push_back (v);
  } else {
    v->rlist->merge (rlist);
  }

  return 0;
}

int RLParam::add (const pair<int, int*> & arr, int * rlist)
{
  value_list * v = 0;
  list<value_list*>::const_iterator vii;
  ArrayParam *val = new ArrayParam (arr.first, arr.second);

  VPTRACE ();

  for (vii = vlist.begin(); vii != vlist.end(); vii++) {
    value_list * tmp = (*vii);
    if (tmp->type == Param_array && *((ArrayParam*)(tmp->value)) == val) {
      v = (*vii);
      break;
    }
  }
  
  if (!v) {
    v = new value_list;
    v->type  = Param_array; 
    v->value = val;
    v->rlist = RList::clone (rlist);
    vlist.push_back (v);
  } else {
    v->rlist->merge (rlist);
    delete val;
  }

  return 0;
}

RLParam::~RLParam () {
  while (!vlist.empty()) {
    value_list * v = (vlist.front ());
    delete v->rlist;
    delete v->value;
    delete v;
    vlist.pop_front();
  VPTRACE ();
  }
}

param_type_t RLParam::param_type ()
{
  list<value_list*>::const_iterator vii;
  param_type_t p = Param_undef;
  for (vii = vlist.begin(); vii != vlist.end(); vii++) {
    if (vii != vlist.begin()) {
      //assert ( (*vii)->type == p);
    }
    p = (*vii)->type;
  }
  return p;
}

int RLParam::output (char * buf)
{
  char * pos = buf;
  list<value_list*>::const_iterator vii;
  char *rlistbuf = new char[32768];
  
  pos += sprintf (pos, "%hhx ", log2pow2(param));
  VPTRACE ();

  /* find the list which occupies most number of bytes */
  RList * maxlist = NULL;
  param_type_t p = Param_undef;
  int max = 0;
  for (vii = vlist.begin(); vii != vlist.end(); vii++) {
    int length = (*vii)->rlist->list_length ();
    if (length > max) {
      max = length;
      maxlist = (*vii)->rlist;
    }
    p = (*vii)->type;
  }

  for (vii = vlist.begin(); vii != vlist.end(); vii++) {
    pos += sprintf (pos, "( ");
    pos += ((*vii)->value)->output (pos);
    pos += sprintf (pos, " )");

    if ((*vii)->type == Param_array || (*vii)->rlist != maxlist) {
      (void) (*vii)->rlist->soutput (rlistbuf);
      pos += sprintf (pos, "%s", rlistbuf);
    }
    if ((*vii)->type == Param_array && *vii != vlist.back()) pos += sprintf (pos, "\n%hhx ", log2pow2(param));
  }
  delete [] rlistbuf;

  pos += sprintf (pos, "\n"); 

  return (pos - buf);
}
int RLParam::packed_size (MPI_Comm comm)
{
  list<value_list*>::const_iterator vii;
  int size = 0;
  
  size += ::packed_size (sizeof(int64_t), MPI_BYTE, comm); //param
  size += ::packed_size (1,               MPI_INT , comm); //list length

  for (vii = vlist.begin(); vii != vlist.end(); vii++) {
    size += ::packed_size (sizeof(param_type_t), MPI_BYTE, comm);
    if ((*vii)->type == Param_scalar)
      size += ((ScalarParam*)(*vii)->value)->packed_size (comm);
    else
      size += ((ArrayParam*)(*vii)->value)->packed_size (comm);
    size += (*vii)->rlist->packed_size (comm);
  }

  return size;
}

void RLParam::pack   (void *buf, int bufsize, int *position, MPI_Comm comm)
{
  list<value_list*>::const_iterator vii;
  int len = vlist.size ();
  
  PMPI_Pack(&param, sizeof(int64_t), MPI_BYTE, buf, bufsize, position, comm);
  PMPI_Pack(&len,   1,               MPI_INT , buf, bufsize, position, comm);

  for (vii = vlist.begin(); vii != vlist.end(); vii++) {
    PMPI_Pack(&((*vii)->type), sizeof (param_type_t), MPI_BYTE, buf, bufsize, position, comm);
    if ((*vii)->type == Param_scalar)
      ((ScalarParam*)(*vii)->value)->pack(buf, bufsize, position, comm);
    else
      ((ArrayParam*)(*vii)->value)->pack(buf, bufsize, position, comm);
    (*vii)->rlist->pack(buf, bufsize, position, comm);
  }

  return ;
}

RLParam * RLParam::unpack (void *buf, int bufsize, int *position, MPI_Comm
    comm)
{
  RLParam * V = new RLParam ();
  int len = 0;

  PMPI_Unpack(buf, bufsize, position, &V->param, sizeof (int64_t), MPI_BYTE, comm);
  PMPI_Unpack(buf, bufsize, position, &len     , 1               , MPI_INT , comm);

  for (int i = 0; i < len; i++) {
    value_list * v = new value_list;

    PMPI_Unpack(buf, bufsize, position, &v->type, sizeof (param_type_t), MPI_BYTE, comm);
    if (v->type == Param_scalar) {
      v->value = ScalarParam::unpack(buf, bufsize, position, comm);
    } else if (v->type == Param_array) {
      v->value = ArrayParam::unpack(buf, bufsize, position, comm);
    } else {
      assert (0);
    }
    v->rlist = RList::unpack(buf, bufsize, position, comm);
    V->vlist.push_back (v);
  }

  return V;
}

void RLParam::merge (RLParam * other) {
  list<value_list*>::const_iterator vii;

  for (vii = other->vlist.begin(); vii != other->vlist.end(); vii++) {
    param_type_t type = (*vii)->type;
    if (type == Param_scalar) {
      add (((ScalarParam*)(*vii)->value)->getValue(), (*vii)->rlist->get_rlist());
    } else if (type == Param_array) {
      ArrayParam * a = (ArrayParam*)(*vii)->value;
      add (make_pair (a->getArraySize(), a->getArray()), (*vii)->rlist->get_rlist());
    } else {
      assert (0);
    }
  }
}

RLParam * find_in_list (list<RLParam*> * l, int64_t param)
{
  list<RLParam*>::const_iterator lii;

  if (l == NULL ||l->empty ()) return NULL;

  for (lii = l->begin(); lii != l->end(); lii++) {
    if ( (*lii) && (*lii)->get_param() == param) {
      return (*lii);
    }
  }
  return NULL;
}

void merge_lists (list<RLParam*> *v1, list<RLParam*> *v2) {
  list<RLParam*>::const_iterator lii;

  for (lii = v1->begin(); lii != v1->end(); lii++) {
    RLParam * V = find_in_list (v2, (*lii)->get_param());
    if (V) {
      (*lii)->merge (V);
      v2->remove (V);
    }
  }
  v1->splice (v1->end(), *v2);

}

#if 0
void find_and_insert (list<RLParam*> *l, int64_t param,
    int64_t val1, int * rlist1,
    int64_t val2, int * rlist2)
{
  RLParam * found = find_in_list (l, param);
  if (found) {
    if (0 == get_my_rank ())
    found->add (val2, rlist2);
  } else {
    if (0 == get_my_rank ())
    RLParam * V = new RLParam (param, val1, rlist1, val2, rlist2);
    l->push_back (V);
  }
}
#endif

// --- ScalarParam
int ScalarParam::output  (char * buf) {
  return sprintf (buf, "%" PRId64 " ", value);
}
// ArrayParam
int ArrayParam::output  (char * buf) {
  char * pos = buf;

  VPTRACE ();

  pos += sprintf (pos, "%d ", size);
  for (int i = 0; i < size; i++)
    pos += sprintf (pos, "%d ", entries[i]);

  return (pos - buf);
}

//!---- C interface
param_type_t rlparam_type (int64_t param, RLParamList rlparam)
{
  assert (rlparam);
  list<RLParam*> * l = (list<RLParam*>*)rlparam;

  RLParam * v = find_in_list (l, param);

  if (!v) return Param_undef;
  else return v->param_type ();
}

template <class param_a, class param_b>
void rlparam_merge_internal (int64_t param,
    RLParamList * rlparam1, 
    RLParamList * rlparam2, 
    param_a val1, int * rlist1,
    param_b val2, int * rlist2)
{
  assert (rlparam1 && rlparam2);
  list<RLParam*> * l1 = (list<RLParam*>*)(*rlparam1);
  list<RLParam*> * l2 = (list<RLParam*>*)(*rlparam2);
  RLParam * v1 = find_in_list (l1, param);
  RLParam * v2 = find_in_list (l2, param);

  VPTRACE ();

  if (v1 == NULL && v2 == NULL) {
  VPTRACE ();
    v1 = new RLParam (param);
    v1->add (val1, rlist1);
    v1->add (val2, rlist2);
    l1->push_back (v1);
  } else if (v1 && v2 == NULL){
  VPTRACE ();
    v1->add (val2, rlist2);
  } else if (v1 == NULL && v2){
  VPTRACE ();
    l2->remove (v2);
    v2->add (val1, rlist1);
    l1->push_back (v2);
  } else { // both exist 
  VPTRACE ();
    (v1)->merge (v2);
  }
  return;
}

void rlparam_merge (int64_t param,
    RLParamList * rlparam1, 
    RLParamList * rlparam2, 
    int64_t val1, int * rlist1,
    int64_t val2, int * rlist2)
{
  VPTRACE();
  rlparam_merge_internal (param, rlparam1, rlparam2, val1, rlist1, val2, rlist2);
  return;
}

void rlparam_merge_arr (int64_t param,
    RLParamList* rlparam1,
    RLParamList* rlparam2,
    int arr1size, int * arr1, int * rlist1,
    int arr2size, int * arr2, int * rlist2)
{
  assert (arr1);
  assert (arr2);
  VPTRACE();
  rlparam_merge_internal (param, rlparam1, rlparam2,
      make_pair(arr1size, arr1), rlist1, 
      make_pair(arr2size, arr2), rlist2);
}

void rlparam_free (RLParamList* v) {
  list<RLParam*> * l = (list<RLParam*>*)(*v);
  list<RLParam*>::const_iterator lii;
  if(l->empty()){
    //printf("RLParam is empty\n");  
  }
  else {
    //printf("RLParam not empty\n");
  for (lii = l->begin(); lii != l->end(); lii++) {
    if(*lii!=NULL)
    delete (*lii);
    
  }
  }
  delete l;
  *v = NULL;
}

int rlparam_packed_size (RLParamList rlparam, MPI_Comm comm) {
  list<RLParam*> * l = (list<RLParam*>*)rlparam;
  list<RLParam*>::const_iterator lii;

  int size = ::packed_size (1, MPI_INT , comm); //list length
  for (lii = l->begin(); lii != l->end(); lii++)
    size += (*lii)->packed_size (comm); 

  return size;
}

void rlparam_pack (RLParamList rlparam, void * buf, int bufsize, int *position, MPI_Comm comm)
{
  list<RLParam*> * l = (list<RLParam*>*)rlparam;
  list<RLParam*>::const_iterator lii;

  int len = 0;
  if (l) len = l->size ();
  
  PMPI_Pack(&len, 1, MPI_INT , buf, bufsize, position, comm);
  for (lii = l->begin(); lii != l->end(); lii++) {
    (*lii)->pack (buf, bufsize, position, comm);
  }
  return;
}

RLParamList rlparam_unpack (void * buf, int bufsize, int *position, MPI_Comm comm)
{
  int len = 0;
  list<RLParam*> * l = new list<RLParam*>();

  PMPI_Unpack(buf, bufsize, position, &len, 1, MPI_INT, comm);

  for (int i = 0; i < len; i++) {
    RLParam * V = RLParam::unpack (buf, bufsize, position, comm);
    l->push_back (V);
  }

  return (RLParamList)l;
}

int rlparam_output (RLParamList rlparam, int64_t param, char *buf)
{
  list<RLParam*> * l = (list<RLParam*>*)rlparam;

  VPTRACE ();
  if (l->empty()) return 0;

  VPTRACE ();

  RLParam * V = find_in_list (l, param);
  if (V) {
    VPTRACE ();
    return V->output (buf);
  } else {
    VPTRACE ();
    return 0;
  }
}

void rlparam_init (RLParamList * v) {
  list<RLParam*> *l = new list<RLParam*>( );
  *v = (RLParamList)l;
}

template <class T>
void string_to_int_array (char * buf, char * end, size_t * size, T**arr)
{
  T * array = NULL;
  int array_max = 512;
  int width = sizeof (T);
  stringstream ss (stringstream::in | stringstream::out);
  array = (T*)malloc (width * array_max );

  int len = 0;
  if (end)
    len = end - buf + 1;
  else
    len = strlen (buf);
  if (len <= 0) {
    cout << len << endl;
    cout << buf << endl;
    exit (0);
  }
  string str (buf, len);

  ss << str;

  int sz = 0;
  while (ss >> array[sz]) {
    sz++;
    if (sz >= array_max) {
      array_max += (array_max>>1);
      array = (T*)realloc (array, width * array_max);
    }
  }
  if (sz == 0)
    cout << "buf = " << str << endl;
  if (size)
    *size = sz;
  *arr= array;
  return;
}

int64_t * rlparam_get_int64 (char * buf, int rank, int *size) {
  int64_t * value = 0, * default_value = 0;
  size_t sz = 0;
  char * pos = strchr (buf, '(');

  do {
    if(*pos != '('){
      pos = strchr (pos, '(');
      if(!pos) 
        break;
    }
    char * end = strchr (pos, ')');
    if (value) free (value);
    string_to_int_array (pos+1, end-1, &sz, &value);

    //look for the parenthesis after the digit ...
    pos = end+1;

    if (*pos == '[') { //read the rlist
      char * iter = strchr (pos, ']');
      int * rlist ;
      string_to_int_array (pos+1, iter-1, NULL, &rlist);
      if (member_rank (rank, rlist)) {
        free (rlist);
        if (size)
          *size = (int)sz;
        return value;
      }
      free (rlist);
      pos = iter+1;
    } else {
      assert (!default_value);
      default_value = value;
    }
  }
  while (*pos != '\0' && *pos != '\n');

  if (size)
    *size = (int)sz;
  return default_value;
}

extern "C" void str2array (char * buf, char * end, size_t * size, int64_t**arr) {
  string_to_int_array (buf, end, size, arr);
}

extern "C" void str2intarray (char * buf, char * end, size_t * size, int **arr) {
  string_to_int_array (buf, end, size, arr);
}
#if 0
int main ()
{
  int size;
  char *buf="262144 ( 1  )[2 -1 10 2 4 -1 27 1 0]( 1024  )( 4096  )[1 -1 26 2 4]";
  int * a = rlparam_get_int64 (buf, 262144, 26, &size);
  for ( i = 0; i < size; i++) cout << a[i] << " ";
  cout << endl;
  return 0;
}
#endif
/* vim: set tw=80 ts=2 sw=2 expandtab: */
