/*
 * Filename: busy_wait.c
 * Author:   pratn
 * Created:  Tue Apr 10 16:42:57 2007
 * $Id: busy_wait.c,v 1.2 2008/01/30 01:51:41 pratn Exp $
 */
#include <stdio.h>
#include <assert.h>
#include <time.h>
#include <sys/time.h>
#include "utils.h"
#include "busy_wait.h"

#ifdef HAVE_RTS_GETTIMEBASE
/*----------------------------------------------------------*/
/* elapsed-time timing functions */
/*----------------------------------------------------------*/

#include <rts.h>

#define WTIME(TB) TB = rts_get_timebase()
#define TCONV(TB1,TB2) seconds_per_cycle*((double) (TB1 - TB2))

double __bw_cur_time (int start)
{
  static int first = 1;
  static unsigned long long tb0;
  unsigned long long       tb;
  static double       seconds_per_cycle = 1.4285714285714285714e-9;
  /* 700 MHz default */
  if (first) {

    BGLPersonality personality;

    if (rts_get_personality(&personality, sizeof(personality)) != 0)
      return 0;
    seconds_per_cycle = 1.0e9/((double) personality.clockHz);

    first = 0;
  }
  if (start)
    WTIME(tb0);
  return (TCONV(WTIME(tb), tb0));
}

int busy_wait (const struct timespec *req, struct timespec *r)
{
  double twait = 0.0, tfirst = 0.0, tnow = 0.0;

  if (!req)
    return -1;

  twait = (double)req->tv_sec*1.0e9 + (double)req->tv_nsec;
  tfirst = __bw_cur_time (1) + twait;

  do {
    tnow = __bw_cur_time (0);
  } while (tnow < twait);

  return 0;
}

#else
int busy_wait (const struct timespec *req, struct timespec *r)
{
  struct timeval param, begin, now, rem;

  param.tv_sec = req->tv_sec;
  param.tv_usec = req->tv_nsec/1000;

  gettimeofday (&begin, NULL);

  timeradd (&begin, &param, &param);

  rem.tv_sec = 0;
  rem.tv_usec = 0;
  while (rem.tv_sec >= 0)
  {
    gettimeofday (&now, NULL);
    timersub (&param, &now, &rem);
  }

  return 0;
}
#endif

/* vim: set tw=80 ts=2 sw=2 expandtab: */
