#include <stdio.h>
#include <iostream>
#include <vector>
#include <assert.h>
#include <cstdlib>
#include <mpi.h>
#include "mpi_util.h"
#include "cvector.h"
#include "utils.h"

using namespace std;

cvector::cvector (int a) :
  v (1, a), 
  iterations (1),
  current (0),
  t (Undo_none),
  undo_iter (0),
  undo_pos (0)
{
}

cvector::cvector () :
  v (),
  iterations (1) ,
  current (0),
  t (Undo_none),
  undo_iter (0),
  undo_pos (0)
{}

void vector_init (cvector ** c, int a)
{
  cvector * tmp = new cvector (a);
  *c = tmp;
}

void vector_free (cvector ** c)
{
  delete (cvector*)(*c);
  *c = NULL;
}

void vector_push_back (cvector * c, int a)
{
  c->v.push_back (a);
}

int  vector_get_next (cvector * c)
{
  return c->get_next ();
}

int * vector_to_array (cvector * c, int * size)
{
  int * a = NULL;
  int sz = 0;

  if (!c || c->v.size() == 0) return NULL;

  //a = &c->v[0];
  sz = c->v.size () + 1;
  if (size) *size = sz;

  a = (int*) malloc (sz * sizeof (int));
  a[0] = c->iterations;
  for (int i = 1; i < sz; i++)
    a[i] = c->v[i-1];

  return a;
}

cvector * array_to_vector (int64_t * arr, int size)
{
  cvector * c = new cvector ();
  c->iterations = arr [0];

  for (int i = 1; i <= size; i++)
    c->v.push_back (arr [i]);

  return c;
}

int vector_output (char *buf, cvector *c)
{
  char * pos = buf;
  int size = c->v.size ();

  pos += sprintf (pos, "%d ", c->iterations);
  for (int i = 0; i < size; i++)
    pos += sprintf (pos, "%d ", c->v[i]);
  //pos += sprintf (pos, "\n");

  return pos - buf;
}

int vector_match_first (cvector * c, int a)
{
  if (!c) return FALSE;
  if (c->v[0] == a) {
    return TRUE;
  } else {
    return FALSE;
  }
}

int vector_match (cvector *c, cvector *d)
{
  if (!c && !d) return TRUE;
  else if ( (!c && d) || (c && !d) ) return FALSE;
  if( c->v == d->v) {
    return TRUE;
  } else {
    return FALSE;
  }
}

int vector_increment_count (cvector * c)
{
  if (!c) return FALSE;
  else c->iterations++;
  return TRUE;
}

int vector_concat (cvector *c, cvector *d)
{
  if (!c) return FALSE;
  c->v.insert (c->v.end(), d->v.begin(), d->v.end());
  return TRUE;
}
    
void vector_undo_info (cvector *c, vector_info_type t)
{
  if (!c) return;

  c->t = t;
  if (t == Undo_pos) {
    vector<int>::iterator ii = c->v.end();
    ii--;
    c->undo_iter = c->v.size();
  } else if (t == Undo_iters) {
    c->undo_iter = c->iterations;
  } 

  return;
}

void vector_undo_addition (cvector ** c1, int * a)
{
  cvector * c = *c1;
  if (!c) return;
  

  if (c->t == Undo_iters) {
    c->iterations = c->undo_iter;
    assert (c->iterations);
  } else if (c->t == Undo_pos) {
    //vector<int>::iterator ii = c->v.end();
    //ii--;
    //c->v.erase (c->undo_pos, ii);
    while (c->v.size() > c->undo_iter)
      c->v.pop_back();
  }

  c->t = Undo_none;

  if (c->v.size () <= 1) {
    *a = c->v[0];
    vector_free (c1);
  }
}

void vector_reset_undo_info (cvector * c)
{
  if (!c) return;
  c->t = Undo_none;
}

int vector_packed_size(cvector *c, MPI_Comm comm) {
  int size = 0;

  if (!c) return packed_size(1, MPI_INT, comm);
  
  size = c->v.size() + 2; // size + #iters

  return packed_size(size, MPI_INT, comm);
}

void vector_pack(cvector *c, void *buf, int bufsize, int *position, MPI_Comm comm) {
  int size = 0;

  if (!c) {
    PMPI_Pack(&size, 1, MPI_INT, buf, bufsize, position, comm);
    return;
  }
  
  size = c->v.size ();
  PMPI_Pack(&size, 1, MPI_INT, buf, bufsize, position, comm);
  PMPI_Pack(&c->iterations, 1, MPI_INT, buf, bufsize, position, comm);
  for (int i = 0; i < size; i++)
    PMPI_Pack(&c->v[i], 1, MPI_INT, buf, bufsize, position, comm);
}

cvector* vector_unpack(void *buf, int bufsize, int *position, MPI_Comm comm) {
  int size, iterations, num;
  
  PMPI_Unpack(buf, bufsize, position, &size, 1, MPI_INT, comm);
  if (!size) return NULL;
  PMPI_Unpack(buf, bufsize, position, &iterations, 1, MPI_INT, comm);

  cvector * c = new cvector ();
  c->iterations = iterations;
  assert (c->iterations);
  for (int i = 0; i < size; i++) {
    PMPI_Unpack(buf, bufsize, position, &num, 1, MPI_INT, comm);
    c->v.push_back (num);
  }

  return c;
}

#ifdef DEBUG
int main () {
  vector <int> v;

  vector<int>::const_iterator vi;

  v.push_back (10);
  v.push_back (2);
  v.push_back (30);
  v.push_back (40);
  v.push_back (21);
  v.push_back (32);

  for (vi = v.begin (); vi != v.end (); vi++)
    cout << *vi << endl;

  return 0;
}
#endif
