/*
 * filename list.c
 * pratn    Prasun Ratn
 *
 * this file implements a double linked list
 * method names are self explanatory
 */
 
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include "list.h"

#ifndef NULL
#define NULL 0
#endif

list_t list_create (void * v)
{
    list_t n = NULL;  

    n = (list_t) calloc (sizeof(struct list_node), 1);
    n->v = v;
    n->prev = n->next = NULL;

    return n;
}

list_t list_prev   (list_t N)
{
    if (!N)
        return NULL;
    else
        return N->prev;
}

list_t list_next   (list_t N)
{
    if (!N)
        return NULL;
    else
        return N->next;
}

void * list_value (list_t N)
{
    if (N)
        return N->v;
    else
        return NULL;
}

int list_size (list_t N)
{
  list_t n = NULL;
  int count = 0;

  n = N;
  while (n)
  {
    count++;
    n = n->next;
  }

  return count;
}

list_t list_insert (list_t N, void * v)
{
    list_t n = NULL;
    
    if (N == NULL)
    {
        return list_create (v);
    }
    n = N;
    while (n->next)
        n = n->next;
    n->next = (list_t) calloc (sizeof(struct list_node), 1);
    n->next->v = v;
    n->next->prev = n;

    return N;
}

list_t list_concat (list_t l1, list_t l2)
{
    list_t tmp = NULL;

    if (!l1) return l2;
    if (!l2) return l1;

    tmp = l1;
    while (tmp->next)
        tmp = tmp->next;
    tmp->next = l2;
    l2->prev = l1;

    return l1;
}

list_t list_delete (list_t N, void * v, list_f func)
{
    list_t tmp = N;

    if (N == NULL)
        return NULL;

    while (v != tmp->v)
    {
        tmp = tmp->next;
        if (tmp == N)
            break;
    }
    if (tmp->prev)
        tmp->prev->next = tmp->next;
    if (tmp->next)
        tmp->next->prev = tmp->prev;

    if (tmp == N)
    {
        if (N->next == N && N->prev == N)
            N = NULL;
        else
            N = tmp->next;
    } 

    if (func)
      (*func) (tmp->v);
    free (tmp);
    return N;
}

/* Free up allocated memory */
void list_free (list_t * N, list_f func)
{
    list_t tmp = * N, tmp2 = NULL;

    if (N)
    {
        while (tmp)
        {
            tmp2 = tmp;
            tmp = tmp->next;
            if (func)
              (*func) (tmp2->v);
            free (tmp2);
        }
        *N = NULL;
    }
}

/* print the list for debugging purposes */
void list_print (list_t N, list_print_fn * fn)
{
    list_t tmp = N;
    if (N)
    {
       while (tmp)
       {
           printf ("%p {%p,          ", tmp, tmp->v);
           if (fn) (*fn)(tmp->v);
           printf ("        %p, %p}\t",              tmp->next, tmp->prev);
           tmp = tmp->next;
           if (tmp == N)
               break;
       } 
       printf ("\n");
    }
    else
        printf ("Empty list\n");
}

int list_packed_size(list_t list, MPI_Comm comm)
{
  int size = 0;

  if (list) {
    size += packed_size(1, MPI_INT, comm);
    size += packed_size (list_size (list), MPI_INTPTR_TYPE, comm);
  }
  return size;
}

void list_pack(list_t list, void * buf, int bufsize, int * position, MPI_Comm comm)
{
  list_t ptr = NULL;
  if (list) {
    int len = list_size (list);
    PMPI_Pack (&len, 1, MPI_INT, buf, bufsize, position, comm);
    for (ptr = list; ptr; ptr = ptr->next) {
      PMPI_Pack (&ptr->v, 1, MPI_INTPTR_TYPE, buf, bufsize, position, comm);
    }
  }
}

list_t list_unpack (void * buf, int bufsize, int * position, MPI_Comm comm)
{
  int len = 0;
  int i = 0;
  list_t l = NULL;

  PMPI_Unpack (buf, bufsize, position, &len, 1, MPI_INT, comm);
  for (i = 0; i < len; i++) {
    void * v;
    PMPI_Unpack (buf, bufsize, position, &v, 1, MPI_INTPTR_TYPE, comm);
    l = list_insert (l, v);
  }
  return l;
}

#if 0
void list_test ()
{
    list_t x = NULL;

    x = list_insert (x, 1);
    x = list_insert (x, 2);

    list_print_test (x);
    return;
}
#endif
