#include "opstruct.h"
#include <sys/time.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>
//#define NEED_TRACE
#include "trace.h"
#include "utils.h"
#include "mpi_util.h"
#include "Timing.h"
#include "Event.h"



#define round(x) ((x)>=0?(long)((x)+0.5):(long)((x)-0.5))

op_field_desc_t op_field_table[] = {
  {GLOBAL_WORLD,    "%d",   "global_world" },
  {NULL_REQ,        "%d",   "null_req"     },
  {COMM,            "%d",   "comm"         },
  {COUNT,           "%d",   "count"        },
  {DATATYPE,        "%d",   "datatype"     },
  {RECVCOUNT,       "%d",   "recvcount"    }, 
  {RECVTYPE,        "%d",   "recvtype"     },
  {MPI_OP,          "%d",   "mpi_op"       }, 
  {ROOT,            "%d",   "root"         }, 
  {TAG,             "%d",   "tag"          }, 
  {DEST,            "%d",   "dest"         }, 
  {REQUEST,         "%d",   "request"      }, 
  {GROUP,           "%d",   "group"        }, 
  {COLOR,           "%d",   "color"        }, 
  {KEY,             "%d",   "key"          }, 
  {GROUP1,          "%d",   "group1"       }, 
  {GROUP2,          "%d",   "group2"       }, 
  {LOCAL_LEADER,    "%d",   "local_leader" }, 
  {PEER_COMM,       "%d",   "peer_comm"    }, 
  {REMOTE_LEADER,   "%d",   "remote_leader"}, 
  {HIGH,            "%d",   "high"         }, 
  {SOURCE,          "%d",   "source"       }, 
  {RECVTAG,         "%d",   "recvtag"      }, 
  {NEWTYPE,         "%d",   "newtype"      }, 
  {OLDTYPE,         "%d",   "oldtype"      }, 
  {BLOCKLEN,        "%d",   "blocklen"     }, 
  {STRIDE,          "%d",   "stride"       }, 
  {REORDER,         "%d",   "reorder"      },
  {NEWRANK,         "%d",   "newrank"      },
  {OUTCOUNT,        "%d",   "outcount"     }, 
  {NO_FIELDS,       NULL,   NULL           }
};

#include "ptr_handler.h"

int soutput_field_value(char *buf, replay_op *op, op_field_desc_t *fdesc) {
  if (fdesc->field & GLOBAL_WORLD)         return sprintf(buf, fdesc->format, op->global_world);
  else if (fdesc->field & NULL_REQ)        return sprintf(buf, fdesc->format, op->null_req);
  else if (fdesc->field & COMM)            return sprintf(buf, fdesc->format, op->comm);
  else if (fdesc->field & COUNT)           return sprintf(buf, fdesc->format, op->count);
  else if (fdesc->field & DATATYPE)        return sprintf(buf, fdesc->format, op->datatype);
  else if (fdesc->field & RECVCOUNT)       return sprintf(buf, fdesc->format, op->recvcount);
  else if (fdesc->field & RECVTYPE)        return sprintf(buf, fdesc->format, op->recvtype);
  else if (fdesc->field & MPI_OP)          return sprintf(buf, fdesc->format, op->mpi_op);
  else if (fdesc->field & ROOT)            return sprintf(buf, fdesc->format, op->root);
  else if (fdesc->field & TAG)             return sprintf(buf, fdesc->format, op->tag);
  else if (fdesc->field & DEST)            return sprintf(buf, fdesc->format, op->dest);
  else if (fdesc->field & REQUEST)         return sprintf(buf, fdesc->format, op->request);
  else if (fdesc->field & GROUP)           return sprintf(buf, fdesc->format, op->group);
  else if (fdesc->field & COLOR)           return sprintf(buf, fdesc->format, op->color);
  else if (fdesc->field & KEY)             return sprintf(buf, fdesc->format, op->key);
  else if (fdesc->field & GROUP1)          return sprintf(buf, fdesc->format, op->group1);
  else if (fdesc->field & GROUP2)          return sprintf(buf, fdesc->format, op->group2);
  else if (fdesc->field & LOCAL_LEADER)    return sprintf(buf, fdesc->format, op->local_leader);
  else if (fdesc->field & PEER_COMM)       return sprintf(buf, fdesc->format, op->peer_comm);
  else if (fdesc->field & REMOTE_LEADER)   return sprintf(buf, fdesc->format, op->remote_leader);
  else if (fdesc->field & HIGH)            return sprintf(buf, fdesc->format, op->high);
  else if (fdesc->field & SOURCE)          return sprintf(buf, fdesc->format, op->source);
  else if (fdesc->field & RECVTAG)         return sprintf(buf, fdesc->format, op->recvtag);
  else if (fdesc->field & NEWTYPE)         return sprintf(buf, fdesc->format, op->newtype);
  else if (fdesc->field & OLDTYPE)         return sprintf(buf, fdesc->format, op->oldtype);
  else if (fdesc->field & BLOCKLEN)        return sprintf(buf, fdesc->format, op->blocklen);
  else if (fdesc->field & STRIDE)          return sprintf(buf, fdesc->format, op->stride);
  else if (fdesc->field & REORDER)         return sprintf(buf, fdesc->format, op->reorder);
  else if (fdesc->field & NEWRANK)         return sprintf(buf, fdesc->format, op->newrank);
  else if (fdesc->field & OUTCOUNT)        return sprintf(buf, fdesc->format, op->outcount);
  else {
    fprintf(stderr, "Error: unhandled op value in soutput_field_value: %" PRIfieldINT "\n", fdesc->field);
    return 0;
  }
}


void map_over_fields(replay_op *op, field_handler_t func) {
  if (!op->fields) return;  // small optimization for fieldless ops

  op_field_desc_t *field = &op_field_table[0];
  while (field->field) {
    if (op->fields & field->field) {
      func(op, field);
    }
    field++;
  }
}

op_field_desc_t *get_field_descriptor(field_mask_t value) {
  op_field_desc_t *field = &op_field_table[0];
  while (field->field && field->field != value) field++;
  return field;  
}


int soutput_array(char *buf, field_mask_t arr_type, int size, array_elt_t *array) {
  char *pos = buf;
  int i = 0;
  pos += sprintf(pos, "%hhx %d", log2pow2(arr_type), size);
  for(i = 0; i < size; i++) {
    pos += sprintf(pos, " %" PRImpiARR, array[i]);
  }
  pos += sprintf(pos, "\n");

  return (pos - buf);
}

/*Added by Karthik - Start
#ifdef FEATURE_FILE_IO
int soutput_array_extn(char *buf, field_mask_t arr_type, int size, array_elt_t *array) {
  char *pos = buf;

  pos += sprintf(pos, "%hhx %d", (log2pow2(arr_type)^0xff)+1, size);
  for(int i = 0; i < size; i++) {
    pos += sprintf(pos, " %" PRImpiARR, array[i]);
  }
  pos += sprintf(pos, "\n");

  return (pos - buf);
}
#endif
Added by Karthik - End*/

replay_op operation;
Event* event;
int is_mpi_inited;

void init_op(replay_op *operation)
{
  is_mpi_inited = 1;
  /* meta info to parse the op */
  operation->sync_bytes = SYNC;
  operation->op = 0; 
  operation->seq_num = -1;

#ifdef FEATURE_TIMING_HISTO
  for (int i = 0; i < TIMING_MAX; i++)
  {
    operation->time_info[i].time_list = (void*) 0;
    operation->time_info[i].time = 0;
  }
#endif

  /* No arrays initially modified, and arrays initially NULL. */
  operation->arrays = NO_ARRAYS;
  operation->array_size1 = 0;
  operation->array_size2 = 0;
  operation->array_size3 = 0;
  operation->array_size4 = 0;


  /* Set this to no fields initially. */
  operation->fields = NO_FIELDS;

  /* empty initially */
  operation->pvectors = NO_FIELDS;
  rlparam_init (&operation->rlparam);

  /* array ptrs */
  operation->array1 = NULL;
  operation->array2 = NULL;
  operation->array3 = NULL;
  operation->array4 = NULL;

#ifdef FEATURE_FILE_IO
  /* MPI_FILE_xxx */
  operation->filename = NULL;
  operation->datarep = NULL;
  operation->amode = -1;
  operation->fh = -1;
  operation->offset = -1;
  operation->offsetinfo.startpos = 0;
  operation->offsetinfo.stride = 0;
  operation->offsetinfo.totalcount = 0;
  operation->disp = 0;
  operation->etype = -1;
  operation->filetype = -1;
  operation->whence = -1;
  //operation->info = NULL;
  operation->fields_extn = NO_FIELDS;
  operation->arrays_extn = NO_ARRAYS;
  operation->filepath = NULL;
  operation->is_mpi_io=0;
  operation->rank = -1;
  operation->comm_size = 0;
  operation->order = 0;
  operation->disp_vector = NULL;
  operation->posixargs.flags=-1;
  operation->posixargs.mode=-1;
#endif
}

#if defined (FEATURE_TIMING_HISTO) || defined (FEATURE_TIMING_BITS)
static unsigned long long startTime; // last timestamp in us

void initTime () {
  startTime = get_cur_time_ns() / 1000;
}

void recordTime(replay_op *operation, int index) {
  /*long long int timeStamp = get_cur_time_ns() / 1000;
  if (operation) {
    if((timeStamp-startTime) < 0) {
      operation->time_info[index].time = 0;
      printf("DEFECT:Recorded time is less than 0\n");fflush(stdout);
    } else {
      operation->time_info[index].time = timeStamp - startTime;
    }
  }

  startTime = timeStamp;*/
}

void recordEventTime(Event *operation, int index) {
  unsigned long long int timeStamp = get_cur_time_ns() / 1000;
  if (operation) {
    if((timeStamp-startTime) < 0) {
      set_event_time(operation, 0, index);
      printf("DEFECT:Recorded Event time is less than 0\n");fflush(stdout);
    } else {
      set_event_time(operation, timeStamp-startTime, index);
    }
  }

  startTime = timeStamp;
}

#endif // Timing features.

/* vim: set tw=80 ts=2 sw=2 expandtab: */
