#include <stdlib.h>
#include "params_queue.h"

void init_params_node(params_node *node)
{
  node->offset = 0;
  node->next = NULL;
  node->prev = NULL;
}

void init_params_queue(params_queue *queue)
{
  queue->update_flag = 0;
  queue->head = NULL;
  queue->tail = NULL;
  queue->cur = NULL;
}

void params_push(params_queue *queue, int offset)
{
  params_node *new_tail = (params_node *) malloc(sizeof(params_node));

  if(new_tail == NULL) {
    perror("malloc");
    exit(0);
  }

  init_params_node(new_tail);
  new_tail->offset = offset;
  new_tail->prev = queue->tail;

  if(queue->head == NULL) {
    queue->head = new_tail;
    queue->tail = new_tail;
    queue->cur = new_tail;
  }
  else {
    queue->tail->next = new_tail;
    queue->tail = new_tail;
  }

  queue->update_flag = 1;
}

void params_push_replay(params_queue *queue, params_node *new_tail)
{
  new_tail->next = NULL;
  new_tail->prev = queue->tail;

  if(queue->head == NULL) {
    queue->head = new_tail;
    queue->tail = new_tail;
  }
  else {
    queue->tail->next = new_tail;
    queue->tail = new_tail;
  }
}

params_node *params_pop(params_queue *queue)
{
  params_node *ptr = queue->head;

  if(!queue->head)
    return NULL;

  /* single element to pop */
  if(!queue->head->next) {
    queue->head = NULL;
    queue->tail = NULL;
    queue->cur = NULL;
  }
  else {
    queue->head = queue->head->next;
    queue->cur = queue->head;
    queue->head->prev = NULL;
  }

  return ptr;
}

int soutput_params(char *buf, int len, params_queue *queue)
{
  char *pos = buf;
  params_node *ptr = queue->head;
  int i = 0;

  /* print the length first */
  while(ptr) {
    i++;
    ptr = ptr->next;
  }
  pos += snprintf(pos, len - strlen(buf), "PARAMS: %d", i);

  /* print data */
  ptr = queue->head;
  while(ptr) {
    pos += snprintf(pos, len - strlen(buf), " %d", ptr->offset);
    ptr = ptr->next;
  }
  pos += snprintf(pos, len - strlen(buf), "\n");

  return (pos - buf);
}

void free_params_queue(params_queue *queue)
{
  params_node *ptr = queue->head;

  while(queue->head) {
    queue->head = queue->head->next;
    free(ptr);
    ptr = queue->head;
  }

  queue->head = NULL;
  queue->tail = NULL;
  queue->cur = NULL;
}

int params_queue_compare(params_queue *queue1, params_queue *queue2)
{
  params_node *ptr1 = queue1->head;
  params_node *ptr2 = queue2->head;

  while(ptr1 && ptr2) {
    if(ptr1->offset != ptr2->offset)
      return 0;

    ptr1 = ptr1->next;
    ptr2 = ptr2->next;
  }

  if(ptr1 || ptr2)
    return 0;
  
  return 1;
}

void reset_flag(params_queue *queue)
{
  queue->update_flag = 0;
}

void undo_addition(params_queue *queue)
{
  params_node *ptr = queue->tail;

  queue->tail = queue->tail->prev;
  if(queue->tail == NULL)
    queue->head = NULL;
  else
    queue->tail->next = NULL;

  queue->update_flag = 0;

  free(ptr);
}
