#include <stdlib.h>
#ifndef __STDC_FORMAT_MACROS
#define __STDC_FORMAT_MACROS
#endif
#include <stdint.h>
#include <inttypes.h>
#include "prsd_stack.h"
#include "Histogram.h"

void init_prsd_node(prsd_node *node)
{
  node->loop_iters = 1;
  node->loop_mems = -1;

  sig_init(&node->tail_sig);
  node->prsd_stat = NULL;
  node->next = NULL;
  node->prev = NULL;
}

void free_prsd_node(prsd_node *node)
{
  if(node) {
    /* free string signature */
    sig_destroy(&node->tail_sig);
    free(node);
  }
}

void init_prsd_stack(prsd_stack *stack)
{
  stack->head = NULL;
  stack->tail = NULL;
}


void prsd_push_node(prsd_stack *stack, prsd_node *new_head) {
  new_head->next = stack->head;
  new_head->prev = NULL;

  if(stack->head) {
    stack->head->prev = new_head;
    stack->head = new_head;
  } else {
    stack->head = stack->tail = new_head;
  }
}


void prsd_push(prsd_stack *stack, int members, stack_sig *signature) {
  prsd_node *new_head = (prsd_node *) malloc(sizeof(prsd_node));
  if(new_head == NULL) {
    perror("malloc");
    exit(0);
  }
  
  init_prsd_node(new_head);
  new_head->loop_mems = members;
  sig_assign(&new_head->tail_sig, signature);
  prsd_push_node(stack, new_head);
}


void prsd_push_node_end(prsd_stack *stack, prsd_node *node) {
  node->prev = stack->tail;
  node->next = NULL;

  if (stack->head) {
    stack->tail->next = node;
    stack->tail = node;
  } else {
    stack->head = stack->tail = node;
  }
}


void prsd_push_end(prsd_stack *stack, int members)
{
  prsd_node *new_tail = (prsd_node *) malloc(sizeof(prsd_node));
  if(new_tail == NULL) {
    perror("malloc");
    exit(0);
  }

  init_prsd_node(new_tail);
  new_tail->loop_mems = members;
  prsd_push_node_end(stack, new_tail);
}

prsd_node *prsd_pop(prsd_stack *stack)
{
  prsd_node *ptr = stack->head;

  if(!stack->head)
    return NULL;

  /* single element to pop */
  if(!stack->head->next) {
    stack->head = NULL;
    stack->tail = NULL;
  }
  else {
    stack->head = stack->head->next;
    stack->head->prev = NULL;
  }

  return ptr;
}

int get_depth(prsd_node *node)
{
  int depth = 0;

  while(node) {
    node = node->next;
    depth++;
  }

  return depth;
}

int get_distance(prsd_node *node)
{
  if(!node)
    return 1;
  
  if(node->next)
    return ((get_distance(node->next) + (node->loop_mems - 1)) * (node->loop_iters + 1)) - 1;

  return node->loop_mems * (node->loop_iters + 1);
}

void free_prsd_stack(prsd_stack *stack)
{
  prsd_node *ptr1, *ptr2 = stack->head;

  while(ptr2 != NULL) {
    ptr1 = ptr2->next;
    free_prsd_node(ptr2);
    ptr2 = ptr1;
  }

  stack->head = NULL;
  stack->tail = NULL;
}

/* Returns num chars printed. */
int soutput_prsd_stack(char *buf, int len, prsd_stack *stack) {
  char *pos = buf;
  prsd_node* ptr = NULL; 
  for (ptr = stack->head; ptr != NULL; ptr = ptr->next) {
    pos += snprintf(pos, len - (pos - buf), "PRSD: %d %d ", ptr->loop_iters, ptr->loop_mems);
    pos += sig_soutput(pos, &ptr->tail_sig);
    if(ptr->prsd_stat){
      pos += sprintf(pos, " ");
      pos += histogram_output(pos, ptr->prsd_stat);
    }
    pos += sprintf(pos, "\n");
  }

  return (pos - buf);
}
