#include "prsd_traverse.h"

#include <stdlib.h>
#include <assert.h>

#include "opstruct.h"
#include "ptr_handler.h"

#ifdef FEATURE_CODE_GENERATOR
#include "op_generator.h"
extern var_dump_t * var;
extern int comm_total;
void add_condition(void* data, int rank) {
  int* dump = (int*)data;
  int index = 0;
  if(dump[0]>0) {
    index = dump[0];
    dump[index] = rank;
    dump[0]--;
  }
}

/*void print_rank_condition(rsd_node* cur_op, FILE* skel_fh) {
  int op_tot_nodes = member_count(cur_op->rlist);
  if(comm_total>0 && comm_total > op_tot_nodes) {
    //list_size depth start [length stride] [length stride]
    int list_size = cur_op->rlist[0];
    int depth = -cur_op->rlist[1];
    int start = cur_op->rlist[2];
    //int length = cur_op->rlist[3];
    //int stride = cur_op->rlist[4];
    for(int i=0;i<depth;i++) {
      
    }
    if(stride > 1 && length > 1) {
      fprintf(skel_fh, "\tif(myrank >= %d && myrank <= %d && myrank % %d == %d) {\n",
                      start, start+((length-1)*stride), stride, start%stride);
    }
    else if(stride == 1 && length > 1) {
      fprintf(skel_fh, "\tif(myrank >= %d && myrank <=%d) {\n", start, start+((length-1)*stride));
    }
    else if(stride == 0 && length == 1) {
      fprintf(skel_fh, "\tif(myrank == %d) {\n", start);
    }
  }
}*/

#endif

/**
 * Recursive replay function.  Does the work for traverse_queue and replay_queue.
 * @param head Head of the queue to replay
 * @param info Prsd_node describing tail of current iteration sequence.  This includes # of 
 *             iterations and signature of tail of iteration sequence.
 * @param iterate Flag determines whether we iterate over PRSDs or go over them only once.
 *                Pass TRUE to iterate, FALSE for only once.
 * @param depth Recursion depth.
 * @param visit This function is called on each node traversed, multiple times if iteration is on.
 */
static int replay_queue_recursive(
  rsd_node *head, int rank, prsd_node *info, int iterate, int depth, rsd_queue_visitor_fn_t visit, FILE * skel_fh
  ) {
  // check parameters; ignore empty lists
  if (head == NULL) return 0;
  if (visit == NULL) {
    fprintf(stderr, "Error: replay_queue_recursive called with NULL visitor.\n");
    exit(1);
  }

  int iters = 0;
  int i = 0;
  rsd_node *last_op;
  int gen_skeleton = 0;
  #ifdef FEATURE_CODE_GENERATOR
  int* ranklist = NULL;
  #endif

  if(info == NULL) {
    // skip to tail
    last_op = head; 
    while (last_op && last_op->next) last_op = last_op->next;

  } else {
    // PRSD info contains the tail ptr.
    if (iterate) iters = info->loop_iters;
    last_op = info->tail_ptr;
    if(skel_fh != NULL) {
      #ifdef FEATURE_CODE_GENERATOR
      if(add_var_dump(var, "int i = 0;",0)) {
        fprintf(stderr,"Write buffer to hold variables is possibly full.");
        exit(0);
      }
      fprintf(skel_fh,"\tfor(i=0;i<%d;i++) {\n",iters);
      #endif
    }
  }


  assert (last_op);
  for (i = 0; i <= iters; i++) {
    rsd_node *cur_op = head;
    prsd_node *cur_info = (info == NULL) ? NULL : info->next;

    int done = 0;
    int first_item = 1;

    if(i==0 && skel_fh != NULL) {
      gen_skeleton = 1;
    }
    else {
      gen_skeleton = 0;
    }
    if(i>0 && i==iters && skel_fh) {
      fprintf(skel_fh, "\t}\n");
    }
    /* continue to play ops until we reach the last element */
    while (!done) {
      if (!cur_op) {
        fprintf(stderr, "error: RSD operation pointer NULL\n");
        exit(0);
      }

      /* leader of loop has multiple PRSDs */
      if (first_item && cur_info) {
        if (replay_queue_recursive(cur_op, rank, cur_info, iterate, depth + 1, visit, skel_fh)) {
          return 1; // Visitor aborted looping.
        }

        /* advance to end of PRSD - already played within recursive calls */
        cur_op = cur_info->tail_ptr;
        done = (cur_op == last_op);
        cur_op = cur_op->next;
        first_item = 0;
      }

      if (done) {
        break;
      }

      /* found a loop leader w/ PRSD info - recursively process the loop */
      if (!first_item && cur_op && (cur_op->rsd_info & LEADER_RSD) && cur_op->prsd_info.head) {
        cur_info = cur_op->prsd_info.head;

        if (replay_queue_recursive(cur_op, rank, cur_info, iterate, depth + 1, visit, skel_fh)) {
          return 1; // Visitor aborted looping.
        }

        /* advance to end of PRSD - already played within recursive calls */
        cur_op = cur_info->tail_ptr;
        done = (cur_op == last_op);
        cur_op = cur_op->next;
        first_item = 0;
      }

      /* replay a single terminal */
      else if (cur_op) {
        int j = (i == 0) ? 1 : i;

        done = (cur_op == last_op);
        // Visit the node and figure out whether we should keep going.
        #ifdef FEATURE_CODE_GENERATOR
        if(gen_skeleton) {
          int op_tot_nodes = member_count(cur_op->rlist);
          if(comm_total>0 && comm_total > op_tot_nodes) {
            /*ranklist = (int*)malloc((op_tot_nodes+1)*sizeof(int));
            ranklist[0] = op_tot_nodes;
            traverse_rlist(cur_op->rlist, (void*)ranklist, add_condition);
            
            for(int a = 1; a<=op_tot_nodes; a++) {
              if(a==1)
                fprintf(skel_fh, "if(myrank==%d", ranklist[a]);
              else
                fprintf(skel_fh, " || myrank==%d", ranklist[a]);
            }
            fprintf(skel_fh, ") {\n");*/
            int start = cur_op->rlist[2];
            int length = cur_op->rlist[3];
            int stride = cur_op->rlist[4];
            if(stride > 1 && length > 1) {
              fprintf(skel_fh, "\tif(myrank >= %d && myrank <= %d && myrank % %d == %d) {\n", 
                      start, start+((length-1)*stride), stride, start%stride);
            }
            else if(stride == 1 && length > 1) {
              fprintf(skel_fh, "\tif(myrank >= %d && myrank <=%d) {\n", start, start+((length-1)*stride));
            }
            else if(stride == 0 && length == 1) {
              fprintf(skel_fh, "\tif(myrank == %d) {\n", start);
            }
          }
          if (visit(cur_op, rank, depth, j, done, skel_fh)) {
            return 1; // Visitor aborted looping.
          }
          if(comm_total>0 && comm_total > op_tot_nodes) {
            fprintf(skel_fh, "\t}\n");//Close if(myrank==...) statement
            //free(ranklist);
            //ranklist = NULL;
          }
        }
        #else
        if (visit(cur_op, rank, depth, j, done)) {
          return 1; // Visitor aborted looping.
        }
        #endif
        cur_op = cur_op->next;
        first_item = 0;
      }
    }
  }
  return 0;
}


int traverse_queue(rsd_node *head, rsd_queue_visitor_fn_t visit) {
  rsd_node * ptr = NULL;
  for (ptr = head; ptr; ptr = ptr->next) {
    #ifdef FEATURE_CODE_GENERATOR
    visit(ptr, -1, 0, 0, !(ptr->next), NULL);
    #else
    visit(ptr, -1, 0, 0, !(ptr->next));
    #endif
  }
  return 0;
}


int replay_queue(rsd_node *head, int rank, rsd_queue_visitor_fn_t visit) {
  return replay_queue_recursive(head, rank, NULL, 1, 0, visit, NULL);
}

int traverse_queue_for_generation(rsd_node *head, int rank, rsd_queue_visitor_fn_t visit, FILE * fh) {
  #ifdef FEATURE_CODE_GENERATOR
  comm_total = 0;
  #endif
  return replay_queue_recursive(head, rank, NULL, 1, 0, visit, fh);
  /*rsd_node * ptr = NULL;
  for (ptr = head; ptr; ptr = ptr->next) {
    #ifdef FEATURE_CODE_GENERATOR
    visit(ptr, -1, 0, 0, !(ptr->next),fh);
    #else
    visit(ptr, -1, 0, 0, !(ptr->next));
    #endif
  }
  return 0;*/
}
