#include "rank_recursive.h"
#include <string.h>
#include <stdlib.h>
#include "rank_recursive.h"
#include <assert.h>

int size_recursive_list(int *list) {
  if(!list) return 0;

  int cur = 1;
  int i = 0;
  int num_lists = list[0];
  for(i = 0; i < num_lists; i++) {
    int num_rsds = -(list[cur]);
    cur += ((num_rsds+1)<<1);
  }

  return cur;
}

int member_rank_helper(int rank_target, int *list, int start, int depth)
{
  int i, rank, iters, stride;

  if(depth > -1)
    return 0;
  
  iters = list[0];
  stride = list[1];
  
  for(i = 0, rank = start; i < iters; i++, rank += stride) {
    if(rank == rank_target)
      return 1;
    if(depth != -1) {
      if(member_rank_helper(rank_target, list + 2, rank, depth + 1))
        return 1;
    }
  }
  //VT_AK: member_rank_helper should return a zero if its not a member
  return 0;
}

int member_rank(int rank, int *list) {
  int pos, cur_prsd, depth, start;

  if(!list)
    return 0;

  pos = 1;
  cur_prsd = 0;

  while(cur_prsd < list[0]) {
    depth = list[pos];
    start = list[pos + 1];
    pos += 2;

    if(member_rank_helper(rank, list + pos, start, depth))
      return 1;

    /* find next prsd */
    if(cur_prsd + 1 == list[0]) {
      break;
    }

    pos++;
    while(list[pos] > 0) {
      pos++;
    }
    
    cur_prsd++;
  }
  //VT_AK: Must return to verify that the node is not a memeber
  return 0;
}


int member_count(int *list) {
  if (!list) return 0; // treat NULL as an empty list

  int num_lists = list[0];
  int total_members = 0;
  int i = 0, l=0;
  int list_start = 1;  // start at first list entry
  for (i=0; i < num_lists; i++) {
    int depth = -list[list_start];

    // calculate product of all the iter counts in list to get member count
    if (depth != 0) {
      int cur_members = 1;
      for (l=1; l <= depth; l++) {
	cur_members *= list[list_start + (2*l)];
      }
      total_members += cur_members;
    }
    list_start += 2 * (depth + 1); // skip to next list
  }
  return total_members;
}


static void traverse_rlist_helper(int *list, int start, int depth, void *data, rank_visitor_fn_t visit) {
  int i, rank, iters, stride;

  if(depth > -1)
    return;
  
  iters = list[0];
  stride = list[1];
  
  for(i = 0, rank = start; i < iters; i++, rank += stride) {
    if(depth != -1) {
      traverse_rlist_helper(list + 2, rank, depth + 1, data, visit);
      continue;
    }
    visit(data, rank);
  }
}


void traverse_rlist(int *list, void *data, rank_visitor_fn_t visit) {
  int pos, cur_prsd, depth, start;

  if(!list)
    return;

  pos = 1;
  cur_prsd = 0;

  while(cur_prsd < list[0]) {
    depth = list[pos];
    start = list[pos + 1];
    pos += 2;

    traverse_rlist_helper(list + pos, start, depth, data, visit);
    if(cur_prsd + 1 == list[0]) {     /* find next prsd */
      break;
    }

    pos++;
    while(list[pos] > 0) {
      pos++;
    }
    
    cur_prsd++;
  }
}


int soutput_range_list(int *list, char *buf) {
  int i, j, num_lists, num_rsds, start_rank, cur = 1;
  char *pos = buf;

  if(list) {
    /* print the number of sublists */
    num_lists = list[0];
    pos += sprintf(pos, "%d", num_lists);

    for(i = 0; i < num_lists; i++) {
      /* print the number of RSDs and start rank */
      num_rsds = list[cur];
      start_rank = list[cur + 1];
      pos += sprintf(pos, " %d %d", num_rsds, start_rank);
      cur += 2;

      /* print each RSD */
      for(j = 0; j < -(num_rsds); j ++, cur += 2) {
        pos += sprintf(pos, " %d %d", list[cur], list[cur + 1]);
      }
    }
  }
  return (pos - buf);
}

int *parse_recursive_list(const char *buf) {
  // skip any initial whitespace
  const char *iter = NULL;
  int i = 0;
  const char *start_pos = buf;
  while (*start_pos == ' ') start_pos++;
  int len = 1; // found first int

  /* count subsequent numbers in list. */
  for (iter = start_pos + 1; ; iter++) {
    if (*iter == ' ') {
      len++;
      continue;
    }

    // end of sequence is neither int nor space.
    if ((*iter < '0' || *iter > '9') && *iter != '-') {
      break;
    }
  }

  int *rlist = (int*)malloc(sizeof(int) * len);
  if(!rlist) {
    perror("malloc");
    exit(1);
  }

  /* parse list */
  int num_prsds;
  sscanf(start_pos, "%d", &num_prsds);
  rlist[0] = num_prsds;

  start_pos = strchr(start_pos, ' ') + 1;
  for(i = 1; i < len; i++) {
    sscanf(start_pos, "%d", &rlist[i]);
    start_pos = strchr(start_pos, ' ') + 1;
  }

  return rlist;
}

void dup_recursive_list(int **dest, int *src) {
  if (*dest) {
    free(*dest);
  }

  if (src) {
    int len = size_recursive_list(src);
    *dest = (int*)malloc(sizeof(int) * len);
    memcpy(*dest, src, sizeof(int) * len);
  } else {
    *dest = NULL;
  }
}

void foutput_recursive_list(FILE *out, int *list)
{
  int i, j, num_lists, num_rsds, start_rank, cur = 1;

  if(list) {
    /* print the number of sublists */
    num_lists = list[0];
    fprintf(out, "%d", num_lists);
    
    for(i = 0; i < num_lists; i++) {
      /* print the number of RSDs and start rank */
      num_rsds = list[cur];
      start_rank = list[cur + 1];
      fprintf(out, " %d %d", num_rsds, start_rank);
      cur += 2;

      /* print each RSD */
      for(j = 0; j < -(num_rsds); j ++, cur += 2)
        fprintf(out, " %d %d", list[cur], list[cur + 1]);
    }

    fprintf(out, "\n");
  }
}

void output_helper(FILE *out, int *list, int start, int depth)
{
  int i, rank, iters, stride;

  iters = list[0];
  stride = list[1];

  for(i = 0, rank = start; i < iters; i++, rank += stride) {
    if(depth == -1)
      fprintf(out, "%d ", rank);
    else
      output_helper(out, list + 2, rank, depth + 1);
  }
}

void foutput_expanded_list(FILE *out, int *list)
{
  int pos, cur_prsd, cur, depth, start;

  pos = 1;
  cur_prsd = 0;
  cur = 0;

  while(cur_prsd < list[0]) {
    depth = list[pos];
    start = list[pos + 1];
    pos += 2;

    output_helper(out, list + pos, start, depth);

    /* find next prsd */
    if(cur_prsd + 1 == list[0])
      break;
    else {
      pos++;
      while(list[pos] > 0)
        pos++;
    }

    cur_prsd++;
  }
  fprintf(out, "\n");
}

int create_empty_rlist(int **list) {
  int *tlist = (int *) malloc(sizeof(int) * 5);
  if(!tlist) {
    perror("malloc tlist - create_empty_rlist");
    exit(1);
  }

  tlist[0] = 1;
  tlist[1] = -1;
  tlist[2] = 0;
  tlist[3] = 0;
  tlist[4] = 0;

  *list = tlist;

  return 5; // size of created list 
}


int *alloc_rlist(int len) {
  return (int *) malloc(sizeof(int) * len);
}

int create_recursive_list(int rank, int **list)
{
  int *tlist = NULL;

  tlist = (int *) malloc(sizeof(int) * 5);
  if(!tlist) {
    perror("malloc tlist - create_recursive_list");
    exit(0);
  }

  tlist[0] = 1;
  tlist[1] = -1;
  tlist[2] = rank;
  tlist[3] = 1;
  tlist[4] = 0;

  *list = tlist;

  return 5; // size of created list 
}

void destroy_recursive_list(int *list) {
  free(list);
}

void add_single(int rank, int **list) {
  int *tlist, num_prsds, next_rank, i, j, num_rsds, cur,
    last, last_depth, last_start, last_1iter, last_1stride, last_adjusted;

  /* inits */
  tlist = *list;
  if(!tlist) {
    fprintf(stderr, "error: add_single pass NULL list\n");
    exit(0);
  }

  num_prsds = tlist[0];
  last_adjusted = 0;
  last = 1;
  cur = 1;

  /* determine last prsd in the list (also use cur for size of list) */
  for(i = 0; i < num_prsds; i++) {
    num_rsds = -(tlist[cur]);

    if(i == num_prsds - 1)
      last = cur;
    cur += ((num_rsds+1)<<1);
  }

  /* setup some useful vals for easier code reading */
  last_depth = tlist[last];
  last_start = tlist[last + 1];
  last_1iter = tlist[last + 2];
  last_1stride = tlist[last + 3];

  /* insert into existing prsd */
  if(last_depth == -1) {
    next_rank = last_start + last_1iter*last_1stride;

    if (next_rank == rank || next_rank == last_start) {
      tlist[last + 2]++;  // iter is now 2
      tlist[last + 3] += rank - next_rank; // sets the stride if this is the second element
        last_adjusted = 1;
      }
    }

  /* append a new prsd to the end of the list */
  if(!last_adjusted) {
    /* create new prsd */
    *list = tlist = (int *) realloc(tlist, sizeof(int) * (cur + 4));
    if(!tlist) {
      perror("realloc tlist - add_single");
      exit(0);
    }

    tlist[0]++;
    tlist[cur] = -1;
    tlist[cur + 1] = rank;
    tlist[cur + 2] = 1;
    tlist[cur + 3] = 0;
  }
  return;
}

void merge_single(int rank, int **list) {
  add_single (rank, list);
  if ((*list)[0] > 1) {
    while(merge_compress(list))  {    }
  }
  return;
}

/*
 * find last and penultimate PRSDs
 * if depth matches and the (iter, stride) pairs match
 *    increment depth and add a (iter, stride) pair
 * else if depth differs by 1
 *    check if iters of the 'lowest depth' can be incremented
 */
int merge_compress(int **list) {
  int *tlist, num_prsds, i, j, num_rsds, cur, last, prev;

  /* inits */
  tlist = *list;
  if(!tlist) {
    fprintf(stderr, "error: merge_compress pass NULL list\n");
    exit(0);
  }

  num_prsds = tlist[0];
  last = -1;
  prev = -1;
  cur = 1;

  /* find last & last->prev elements */
  for(i = 0; i < num_prsds; i++) {
    num_rsds = -(tlist[cur]);

    if(i == num_prsds - 2)
      prev = cur;
    if(i == num_prsds - 1)
      last = cur;

    cur += ((num_rsds+1)<<1);
  }

  /* could not find one of the markers */
  if(last < 0 || prev < 0) 
    return 0;

  /* depths match - attempt to realloc & merge 
   * example:
   *    2 -2 448 2,16 4,2 -2 472 2,16 4,2
   *
   * compresses to
   *    2 -3 448 2,24 2,16, 4,2
   * */
  if(tlist[prev] == tlist[last]) {
    /* ensure rsds match */
    num_rsds = -(tlist[last]);
    for(i = 0; i < (num_rsds * 2); i += 2) {
      if(tlist[last + i + 2] != tlist[prev + i + 2] ||
         tlist[last + i + 3] != tlist[prev + i + 3])
        return 0;
    }

    tlist[0]--;          // decrement total prsd count
    tlist[prev]--;       // decrement the depth
    tlist[prev + 2] = 2; // set iters to 2
    tlist[prev + 3] = tlist[last + 1] - tlist[prev + 1]; // set stride

    /* copy rsds (we overwrote the first one in the above 3 lines */
    for(i = 0; i < (num_rsds * 2); i += 2) {
      tlist[prev + 4 + i] = tlist[last + 2 + i];
      tlist[prev + 5 + i] = tlist[last + 3 + i];
    }
    /* reclaim some memory! */
    *list = tlist = (int *) realloc(tlist, sizeof(int) * (last + 2));
    if(!tlist) {
      perror("malloc tlist - merge_compress");
      exit(0);
    }

    return 1;
  }

  /* depths off by 1 - attempt to simply inc iters 
   * example:
   * 2  -3  216  2,120 4.16 4,2
   *    -2  456  4,16 4,2
   *
   *    merges to:
   * 2  -4  216  3,120 4,16 4,2
   * */
  else if(tlist[prev] == tlist[last] - 1) {
    /* ensure correct start val */
    if(tlist[prev + 1] + (tlist[prev + 2] * tlist[prev + 3]) != tlist[last + 1])
      return 0;

    /* ensure rsds match */
    num_rsds = -(tlist[last]);
    for(i = 0; i < (num_rsds * 2); i += 2) {
      if(tlist[last + 2 + i] != tlist[prev + 4 + i] ||
         tlist[last + 3 + i] != tlist[prev + 5 + i])
        return 0;
    }

    tlist[0]--;        // decrement total prsd count
    tlist[prev + 2]++; // inc iters

    /* reclaim some memory! */
    *list = tlist = (int *) realloc(tlist, sizeof(int) * last);
    if(!tlist) {
      perror("malloc tlist - merge_compress");
      exit(0);
    }

    return 1;
  }
  
  /* depths do not match */
  else
    return 0;
}


void fillin_prsd(int *list, int start, int depth, int *output, int *cur) {
  int i, rank, iters, stride;

  iters = list[0];
  stride = list[1];

  for(i = 0, rank = start; i < iters; i++, rank += stride) {
    if(depth == -1) {
      output[*cur] = rank;
      (*cur)++;

      if(*cur >= UNROLL_BUFSIZE - 1) {
        fprintf(stderr, "error: exceeded allocated buffer for unrolled rank list; cur was %d\n", *cur);
        exit(0);
      }
    }
    else
      fillin_prsd(list + 2, rank, depth + 1, output, cur);
  }
}


void merge_unroll(int *list, int *unrolled) {
  int pos, cur_prsd, cur, depth, start;

  pos = 1;
  cur_prsd = 0;
  cur = 0;

  while(cur_prsd < list[0]) {
    depth = list[pos];
    start = list[pos + 1];
    pos += 2;

    fillin_prsd(list + pos, start, depth, unrolled, &cur);

    /* find next prsd */
    if(cur_prsd + 1 == list[0])
      break;
    else {
      pos++;
      while(list[pos] > 0)
        pos++;
    }

    cur_prsd++;
  }

  unrolled[cur] = RANK_END;
}



void merge_recursive_lists(int **range1, int *range2) {
  // if range2 was NULL, nothing to add, so just bail.
  if (!range2) return;

  // if range1 was NULL, we just take range2 entirely.
  if (*range1 == NULL) {
    dup_recursive_list(range1, range2);
    return;
  } 

  // otherwise, we need to do a full merge.  
  int unrolled1[UNROLL_BUFSIZE];
  int *next1 = unrolled1;
  merge_unroll(*range1, unrolled1);

  int unrolled2[UNROLL_BUFSIZE];
  int *next2 = unrolled2;
  merge_unroll(range2, unrolled2);

  int *output = NULL;

  while (*next1 != RANK_END || *next2 != RANK_END) {
    int next;

    if (*next1 == *next2) {
      next = *next1++;
      next2++;
    } 
    else if (*next1 < *next2) {
      next = *next1++;
    }
    else {
      next = *next2++;
    }

    if(!output) {
      create_recursive_list(next, &output);
    } else {
      merge_single(next, &output);
    }
  }

  free(*range1);
  *range1 = output;
}

int recursive_list_packed_size (int * rlist, MPI_Comm comm)
{
  int size = 0;

  if (rlist) {
    size += packed_size(1, MPI_INT, comm);
    size += packed_size (size_recursive_list (rlist), MPI_INT, comm);
  }
  return size;
}

void recursive_list_pack (int * rlist, void * buf, int bufsize, int * position, MPI_Comm comm)
{
  if (rlist) {
    int len = size_recursive_list(rlist);
    PMPI_Pack(&len, 1, MPI_INT, buf, bufsize, position, comm);
    PMPI_Pack(rlist, len, MPI_INT, buf, bufsize, position, comm);
  }
}

int * recursive_list_unpack (void *buf, int bufsize, int *position, MPI_Comm comm)
{
  int num_ranks;
  int * rlist = NULL;

  PMPI_Unpack(buf, bufsize, position, &num_ranks, 1, MPI_INT, comm);

  rlist = (int*)malloc(num_ranks * sizeof(int));
  PMPI_Unpack(buf, bufsize, position, rlist, num_ranks, MPI_INT, comm);

  return rlist;
}
