
#include <stdlib.h>
#include <stdio.h>
#include "req_handler.h"
#include <assert.h>
#include "mpi_util.h"
extern int is_mpi_inited;
void init_req(req_handler_t * r, int size)
{
  req_handler_t req;
  int i;

  req = (req_handler_t) malloc (sizeof(struct req_handler));

  req->req_bufcur = 0;
  req->req_bufsize = size;
  req->req_buf = (MPI_Request *) malloc(sizeof(MPI_Request) * req->req_bufsize);
  for(i = 0; i < req->req_bufsize; i++)
    req->req_buf[i] = MPI_REQUEST_NULL;

  *r = req;
}

void add_request_entry(req_handler_t req, MPI_Request request)
{
  int bufiter = 0;
  if(request != MPI_REQUEST_NULL) {
    /*//Karthik - Start
    if(req->req_bufcur>0) {
      bufiter = req->req_bufcur-1;
      while(bufiter>0) {
        if(req->req_buf[bufiter]==request) {
          return;
        }
        bufiter--;
      }
    }
    //Karthik - End*/
    req->req_buf[req->req_bufcur] = request;
    req->req_bufcur++;
    if(req->req_bufcur >= req->req_bufsize)
      req->req_bufcur = 0;
  }
  else
  {
    fprintf (stderr, "error add_request_entry: request NULL\n");
  }

}

#ifdef DEBUG
void rfinalize() {
  extern req_handler_t record_req;

  fprintf(stderr, "Freeing...\n");
  finalize_req(record_req);
  exit(1);
}
#endif

void finalize_req(req_handler_t req)
{
  free(req->req_buf);  //TODO: figure out why this segfaults!!!!!
  req->req_buf = NULL;
  req->req_bufsize = 0;
  req->req_bufcur = 0;
  free(req);
  is_mpi_inited=0;
}
//Karthik - Start

void print_req(req_handler_t req) {
  int i = 0;
  printf("print_req: Buffer prints:\n");
  while(i<req->req_bufcur) {
    printf("%p\n",req->req_buf[i]);
    i++;
  }
}
//Karthik - End
void lookup_offset (req_handler_t req, MPI_Request request, int *offset)
{
  int i;
  *offset = -1;

  // MPI_REQUEST_NULL is just -1
  if(request == MPI_REQUEST_NULL) return;

  // search for request in the buffer.
  for(i = 0; i < req->req_bufsize && request != req->req_buf[i]; i++);
    
  if (i == req->req_bufsize) { // check if found
    fprintf(stderr, "%d: %d error: unable to locate request handle "
	    "(allocate additional space for the request buffer)\n", my_rank, __LINE__);
    exit(1);
  }

  *offset = (i - req->req_bufcur + req->req_bufsize) % req->req_bufsize;
}

void reset_offset (req_handler_t req, int offset)
{
  // don't reset offset if request is MPI_REQUEST_NULL
  if (offset < 0) {
    return;
  }

  assert (req);
  assert (req->req_bufsize > 0);

  int i = (req->req_bufcur + offset)%req->req_bufsize ;

  req->req_buf[i] = MPI_REQUEST_NULL;

  return;
} 

void lookup_offsetlist(req_handler_t req, MPI_Request *request, int count, int *offsets, int reset) {
  int i, j;

  for(i = 0; i < count; i++) {
    if(request[i] != MPI_REQUEST_NULL) {
      offsets[i] = -1;
      for(j = 0; j < req->req_bufsize; j++) {
        if(request[i] == req->req_buf[j]) {
          offsets[i] = (j - req->req_bufcur + req->req_bufsize) % req->req_bufsize;
          if(reset)
          {
            req->req_buf[j] = MPI_REQUEST_NULL;
          }
          break;
        }
      }
      if(offsets[i] == -1) {
        fprintf(stderr, "%d: error: unable to locate request handle (allocate additional space for the request buffer)\n",
                my_rank);
        exit(0);
      }
    } else {
      offsets[i] = -1;
    }
  }
}

void reset_internal(req_handler_t req, int *offsets, int *indices, int count) {
  int i;

  for(i = 0; i < count; i++)
    req->req_buf[(req->req_bufcur + offsets[indices[i]]) % req->req_bufsize] = MPI_REQUEST_NULL;
}


//NOTE: Null request isn't really a request; it's an offset... just make it -1?
void lookup_request(req_handler_t req,int offset, int null_request, MPI_Request *request) {
  if(offset != null_request) {
    *request = req->req_buf[(req->req_bufcur + offset) % req->req_bufsize];

    if(*request == MPI_REQUEST_NULL) {
      fprintf(stderr, "[%d,%d]: error: unable to locate request handle offset=%d,"
          "null_request=%d \n", __LINE__,my_rank, offset, null_request);
      exit(1);
    }
  } else {
    *request = MPI_REQUEST_NULL;
  }
}

//NOTE: Null request isn't really a request; it's an offset... just make it -1?
void lookup_requestlist(req_handler_t req, int *offsets, int count, int null_request, MPI_Request *requests) {
  int i;

  for(i = 0; i < count; i++) {
    if(offsets[i] != null_request) {
      requests[i] = req->req_buf[(req->req_bufcur + offsets[i]) % req->req_bufsize];

      if(requests[i] == MPI_REQUEST_NULL) {
        fprintf(stderr, "[%d,%d]: error: unable to locate request handle offset=%d,"
            "null_request=%d \n", __LINE__, my_rank, offsets[i], null_request);
        exit(1);
      }
    } else {
      requests[i] = MPI_REQUEST_NULL;
    }
  }
}

void lookup_requestlist_waitsome(req_handler_t req,int *offsets, int count, int null_request, MPI_Request *requests)
{
  int i;

  for(i = 0; i < count; i++) {
    if(offsets[i] != null_request) {
      requests[i] = req->req_buf[(req->req_bufcur + offsets[i]) % req->req_bufsize];
    }
    else
      requests[i] = MPI_REQUEST_NULL;
  }
}
