#include <sys/types.h>
#include <stdio.h>
#include "umpi_internal.h"
#include "internal.h"
#include "umpi_pc.h"
#include "opstruct.h"

extern int is_mpi_inited;
extern "C" ssize_t __real_write( int __fd, const void* __buf, size_t __nbytes );
extern "C" ssize_t __real_pwrite( int __fd, const void* __buf, size_t __nbytes, off_t __offset );

extern "C"
int umpi_posix_write_immediate_local_pre(umpi_op_t * op)
{
  {
    {
      init_op (&operation);
      recordTime (&operation, 0);
    }
  }
  return 0;
}

extern "C"
int
umpi_posix_write_immediate_local_post (umpi_op_t * op)
{

  {
    {
      operation.fields = COUNT;
      operation.fields_extn = FH;
      operation.op = umpi_MPI_op_min;
      operation.seq_num = op->seq_num;
      operation.fh = file_handle_to_index (record_ptr, op->data.mpi.fh);
      operation.count = op->data.mpi.count;
    
      recordTime (&operation, 1);
      compress_rsd (&op_queue, &operation);
      recordTime (0, 0);
    }
  }
  return 0;
}


extern "C"
int
umpi_posix_write_pre (void* pc, int fd, const void* buf, size_t nbytes)
{
  int rc = 0;
  umpi_op_t *uop = (umpi_op_t *) calloc (sizeof (umpi_op_t), 1);

  assert (uop);
  uop->op = 10000;
  uop->rank = umpi_rank;
  uop->seq_num = umpi_task_seq_num++;
  uop->pc = pc;
  uop->data.mpi.fh = (MPI_File)fd;
  //uop->data.mpi.buf = buf;
  uop->data.mpi.count = nbytes;
  umpi_inc_ref_count (uop);
#ifdef UMPI_DEBUG_LEVEL_2
  q_append (umpi_task_op_q, uop);
#endif

  uop->order = umpi_preop;

  rc = umpi_posix_write_immediate_local_pre (uop);
  umpi_dec_ref_count (uop, UMPI_TASK_SYNCHRONOUS);
  return 0;
}

extern "C"
int
umpi_posix_write_post (int MPI_rc,
                             void *pc,int fd, const void* buf, size_t nbytes)
{
  int rc = 0;
  umpi_op_t *uop = (umpi_op_t *) calloc (sizeof (umpi_op_t), 1);

  assert (uop);
  uop->op = 10000;
  uop->rank = umpi_rank;
  uop->seq_num = umpi_task_seq_num++;
  uop->pc = pc;
  uop->data.mpi.fh = (MPI_File)fd;
  //uop->data.mpi.buf = buf;
  uop->data.mpi.count = nbytes;
 
  umpi_inc_ref_count (uop);
#ifdef UMPI_DEBUG_LEVEL_2
  q_append (umpi_task_op_q, uop);
#endif

  uop->order = umpi_postop;
  uop->res = MPI_rc;

  rc = umpi_posix_write_immediate_local_post (uop);
  umpi_dec_ref_count (uop, UMPI_TASK_SYNCHRONOUS);
  return rc;
}



inline
ssize_t
__wrap_write_aux( int fd, const void* buf, size_t nbytes )
{
	ssize_t ret = -1;
    uint64_t enter_ts;
    uint64_t leave_ts;
    void *pc = 0x0;
    GET_FP;
  if(is_mpi_inited) 
    ret = umpi_posix_write_pre (pc, fd, buf, nbytes);
   fprintf(stderr, "Before real write call\n");
        /* do the real operation */
	ret = __real_write( fd, buf, nbytes );
	
   fprintf(stderr, "After real write call\n");

  if(is_mpi_inited) 
    ret = umpi_posix_write_post (ret,
                             pc, fd, buf, nbytes);

	return ret;
}


extern "C"
ssize_t
__wrap_write( int fd, const void* buf, size_t nbytes )
{
    return __wrap_write_aux( fd, buf, nbytes );
}

extern "C"
ssize_t
__wrap___write( int fd, const void* buf, size_t nbytes )
{
    return __wrap_write_aux( fd, buf, nbytes );
}

inline
ssize_t
__wrap_pwrite_aux( int fd, const void* buf, size_t nbytes, off_t offset )
{
  ssize_t ret = -1;
  void *pc = 0x0;
  GET_FP;
  
  if(is_mpi_inited) 
    ret = umpi_posix_write_pre (pc, fd, buf, nbytes);

  fprintf(stderr, "Before real pwrite call\n");
  /* do the real operation */
  ret = __real_pwrite( fd, buf, nbytes, offset );
  fprintf(stderr, "After real pwrite call\n");

  if(is_mpi_inited) 
    ret = umpi_posix_write_post (ret,
                             pc, fd, buf, nbytes);
  return ret;
}


extern "C"
ssize_t
__wrap_pwrite( int fd, const void* buf, size_t nbytes, off_t offset )
{
    return __wrap_pwrite_aux( fd, buf, nbytes, offset );
}


