#include <sys/types.h>
#include <sys/fcntl.h>
//#include "IOTrace_p.h"
#include <stdio.h>
#include <stdarg.h>
#include <stdint.h>
#include "umpi_internal.h"
#include "internal.h"
#include "umpi_pc.h"
#include "opstruct.h"

#ifdef HAVE_MODE_T
#  define MODE_T  mode_t
#else
#  define MODE_T  int
#endif /* HAVE_MODE_T */

extern int is_mpi_inited;
extern "C" int __real_open( const char* __pathname, int __flags, ... );
extern "C" int __real_creat( const char* __pathname, MODE_T mode );

extern "C"
int umpi_posix_open_immediate_local_pre(umpi_op_t * op)
{
  {
    {
      init_op (&operation);
      recordTime (&operation, 0);
    }
  }
  return 0;
}

extern "C"
int
umpi_posix_open_immediate_local_post (umpi_op_t * op)
{

  {
    {
      operation.fields = COUNT;
      operation.fields_extn = FILENAME;// | AMODE;
      operation.op = umpi_MPI_op_min;//op->op;
      operation.seq_num = op->seq_num;
      operation.filepath = (char *) malloc ((strlen (op->data.mpi.filename) + 1) * sizeof (char));      //&(op->data.mpi.filename);
      //strcpy(operation.filepath,op->data.mpi.filename);
      strcpy (operation.filepath, "btio.full.out");
      char *sep = strrchr (operation.filepath, '/');
      if (sep != NULL)
        {
          operation.filename = (char *) malloc ((strlen (sep + 1) + 1) * sizeof (char));        //Reason for spe+1 to remove / from filename
          strcpy (operation.filename, sep + 1);
          memset (sep + 1, '\0', strlen (sep + 1));
          printf ("Rank %d:Write Wrapper: filepath %s, filename %s\n",
                  get_my_rank (), operation.filepath, operation.filename);
        }
      else
        {
          operation.filename =
            (char *) malloc ((strlen (operation.filepath) + 1) *
                             sizeof (char));
          strcpy (operation.filename, operation.filepath);
        }

      //operation.amode = op->data.mpi.amode;
      operation.count = op->data.mpi.flag;

      recordTime (&operation, 1);
      compress_rsd (&op_queue, &operation);
      recordTime (0, 0);
    }
  }
  return 0;
}


extern "C"
int
umpi_posix_open_pre (void* pc, const char* pathname, int flags, int mode)
{
  int rc = 0;
  umpi_op_t *uop = (umpi_op_t *) calloc (sizeof (umpi_op_t), 1);

  assert (uop);
  uop->op = 10000;
  uop->rank = umpi_rank;
  uop->seq_num = umpi_task_seq_num++;
  uop->pc = pc;
  uop->data.mpi.filename =
    (char *) malloc ((strlen (pathname) + 1) * sizeof (char));
  assert (uop->data.mpi.filename);
  bcopy (pathname, uop->data.mpi.filename,
         (strlen (pathname) + 1) * sizeof (char));

  uop->data.mpi.amode = mode;
  uop->data.mpi.flag = flags;
  umpi_inc_ref_count (uop);
#ifdef UMPI_DEBUG_LEVEL_2
  q_append (umpi_task_op_q, uop);
#endif

  uop->order = umpi_preop;

  rc = umpi_posix_open_immediate_local_pre (uop);
  umpi_dec_ref_count (uop, UMPI_TASK_SYNCHRONOUS);
  return 0;
}

extern "C"
int
umpi_posix_open_post (int MPI_rc,
                             void *pc, const char* pathname, int flags, int mode)
{
  int rc = 0;
  umpi_op_t *uop = (umpi_op_t *) calloc (sizeof (umpi_op_t), 1);

  assert (uop);
  uop->op = 10000;
  uop->rank = umpi_rank;
  uop->seq_num = umpi_task_seq_num++;
  uop->pc = pc;
  uop->data.mpi.filename =
    (char *) malloc ((strlen (pathname) + 1) * sizeof (char));
  assert (uop->data.mpi.filename);
  bcopy (pathname, uop->data.mpi.filename,
         (strlen (pathname) + 1) * sizeof (char));

  uop->data.mpi.amode = mode;
  uop->data.mpi.flag = flags;

  umpi_inc_ref_count (uop);
#ifdef UMPI_DEBUG_LEVEL_2
  q_append (umpi_task_op_q, uop);
#endif

  uop->order = umpi_postop;
  uop->res = MPI_rc;

  rc = umpi_posix_open_immediate_local_post (uop);
  umpi_dec_ref_count (uop, UMPI_TASK_SYNCHRONOUS);
  return rc;
}


extern "C"
int
__wrap_open( const char* pathname, int flags, ... )
{
	int ret = -1;
    MODE_T mode = 0;
    uint64_t enter_ts;
    uint64_t leave_ts;

    void *pc = 0x0;
    GET_FP;
    //DEFINE_PC;
    //GATHER_PC;

    if( flags & O_CREAT )
    {
        /* we should have been given a mode parameter */
        va_list ova;
        va_start( ova, flags );
        mode = va_arg( ova, MODE_T );
        va_end( ova );
    }

	/* do the real operation */
    fprintf(stderr, "IO TRACE:1. before open call\n");
  if(is_mpi_inited) 
    ret = umpi_posix_open_pre(pc, pathname, flags, mode);
    fprintf(stderr, "IO TRACE:1. After open_pre call pc value %p\n");
    ret = __real_open( pathname, flags, mode );
    fprintf(stderr, "IO TRACE:1. After real open\n");
  if(is_mpi_inited) 
    ret = umpi_posix_open_post(ret, pc, pathname, flags, mode);
    fprintf(stderr, "IO TRACE:1. after open call\n");

    return ret;	
}



extern "C"
int
__wrap___open( const char* pathname, int flags, ... )
{
	int ret = -1;
    MODE_T mode = 0;
    uint64_t enter_ts;
    uint64_t leave_ts;

    if( flags & O_CREAT )
    {
        /* we should have been given a mode parameter */
        va_list ova;
        va_start( ova, flags );
        mode = va_arg( ova, MODE_T );
        va_end( ova );
    }

	/* do the real operation */
    printf("IO TRACE: before open call");
    ret = __real_open( pathname, flags, mode );
    printf("IO TRACE: after open call");

    return ret;	
}



extern "C"
int
__wrap_creat( const char* pathname, MODE_T mode )
{
	int ret = -1;
	
	/* do the real operation */
        printf("IO TRACE: before create call");
	ret = __real_creat( pathname, mode );
        printf("IO TRACE: after create call");
	
	return ret;		
}


extern "C"
int
__wrap___creat( const char* pathname, MODE_T mode )
{
	int ret = -1;
	
	/* do the real operation */
        printf("IO TRACE: before create call");
	ret = __real_creat( pathname, mode );
        printf("IO TRACE: after create call");
	
	return ret;		
}

