#include "config.h"
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <stdlib.h>

#include "Histogram.h"

int no_send_recv = 0;                            // include sends/recvs in trace
int window = 10000;//500                                // default window size.
int request_buffer_size = 10000;                 // default req buffer size
char output_trace_dir[4096] = "recorded_ops";    // default location of output trace
int alltoallv_limit = 10;                        // default variance for alltoallv

/// Types for configuration data.  
/// CONFIG_BOOL can be "TRUE" or "FALSE", and translates to 0 or 1
/// CONFIG_INT can be any int-parsable string, and is eval'd w/atoi
typedef enum {
  CONFIG_NULL,
  CONFIG_BOOL,
  CONFIG_INT,
  CONFIG_LONG_LONG,
  CONFIG_STRING,
  CONFIG_DBL
} config_type;


struct config_record {
  const char *name;        // name of an env var to check for config info
  const config_type type;  // type of data we expect there
  void *dest;              // addr of var to store the data in

  config_record() 
    : name(NULL), type(CONFIG_NULL), dest(NULL) { }

  config_record(const char *n, const config_type t, void *d) 
    : name(n), type(t), dest(d) { }
};


config_record global_configuration[] = {
  config_record("HISTO_RANKS",         CONFIG_BOOL,        &Histogram::ranks_in_histograms),
  config_record("HISTO_BINS",          CONFIG_INT,         &Histogram::default_buckets),
  config_record("HISTO_FIXWIDTH",      CONFIG_BOOL,        &Histogram::default_fixwidth),
  config_record("HISTO_SMOOTH_THRESH", CONFIG_INT,         &Histogram::default_smooth_threshold),
  config_record("HISTO_MERGE_LIMIT",   CONFIG_INT,         &Histogram::default_merge_limit),
  config_record("HISTO_MERGE_RATIO",   CONFIG_INT,         &Histogram::default_merge_ratio),
  config_record("NO_SENDS",            CONFIG_BOOL,        &no_send_recv),
  config_record("WINDOW_SIZE",         CONFIG_INT,         &window),
  config_record("REQUEST_BUFSIZE",     CONFIG_INT,         &request_buffer_size),
  config_record("TRACE_OUTPUT_DIR",    CONFIG_STRING,      output_trace_dir),

  config_record() // keep this here as a terminal.
};


void setup_configuration() {
  char *err;
  
  for (config_record *record = global_configuration; record->name; record++) {
    char *env_val = getenv(record->name);
    if (env_val == NULL) {
      continue;
    }

    switch (record->type) {
    case CONFIG_BOOL:
      *((int*)record->dest) = (env_val && strcmp("TRUE", env_val) == 0);
      break;

    case CONFIG_INT:
      *((int*)record->dest) = (int)strtol(env_val, &err, 10);
      if (*err) {
        fprintf(stderr, "Error: Invalid integer value for %s.\n", record->name);
        exit(1);
      }
      break;

    case CONFIG_LONG_LONG:
      *((long long*)record->dest) = strtoll(env_val, &err, 10);
      if (*err) {
        fprintf(stderr, "Error: Invalid long long value for %s.\n", record->name);
        exit(1);
      }
      break;

    case CONFIG_DBL:
      *((double*)record->dest) = strtod(env_val, &err);
      if (*err) {
        fprintf(stderr, "Error: Invalid double value for %s.\n", record->name);
        exit(1);
      }
      break;

    case CONFIG_STRING:
      strcpy((char*)record->dest, env_val);
      break;

    default:
      fprintf(stderr, "Error: invalid confgiuration record.");
      assert(0);
      break;
    }
  }
}
