/*
 * Filename: internal.c
 * Author:   prasun
 * Created:  Mon Mar 17 03:38:41 2008
 * $Id: internal.c,v 1.4 2008/09/25 10:49:47 pratn Exp $
 */
#include "internal.h"
#include <assert.h>
#include <sys/time.h>
#include <time.h>
#include "Timing.h"
#include "umpi_internal.h"

/* mike's globals */
req_handler_t record_req;
FILE *output_trace;
rsd_queue op_queue;
ptr_handler_t record_ptr;
struct timeval all_start, all_stop, all_merge_start;

#define name(a) get_struct_by_opcode(a->data.op)->name

void dump_queue_with_label(rsd_queue *queue, int from_rank, int to_rank, char *label) {
  char buf[4096];
  sprintf(buf, "%s/%d-%d.%s", output_trace_dir, from_rank, to_rank, label);
  FILE *recvd = fopen(buf, "w");
  assert(recvd);

  output_rsd_queue(recvd, queue);
  fclose(recvd);
}

void dump_queue_as_graph(rsd_queue *queue, int from_rank, int iter, char *label) {
  char buf[4096];
  sprintf(buf, "%s/%d-%d-%s.dot", output_trace_dir, from_rank, iter, label);
  FILE *recvd = fopen(buf, "w");
  assert(recvd);

  fprintf (recvd, "digraph queue {\n");
  //fprintf (recvd, "concentrate=true;\n");
  for (rsd_node * node = queue->head; node; node = node->next)
  {
    if (node->next)
      fprintf (recvd, "%s_%d_%p -> %s_%d_%p;\n",name(node), node->data.seq_num, node, name(node->next), node->next->data.seq_num, node->next);
    else
      fprintf (recvd, "%s_%d_%p -> end;\n",name(node), node->data.seq_num, node);

    for (list_t t = node->graph.out_edges; t; t = t->next)
    {
      fprintf (recvd, "%s_%d_%p -> %s_%d_%p [color=green];\n",name(node), node->data.seq_num, node, 
          name (((rsd_node*)(t->v))), ((rsd_node*)(t->v))->data.seq_num, t->v);
    }
    for (list_t t = node->graph.in_edges; t; t = t->next)
    {
      fprintf (recvd, "%s_%d_%p -> %s_%d_%p [color=red];\n"
          ,name (((rsd_node*)(t->v))), ((rsd_node*)(t->v))->data.seq_num, t->v
          ,name(node), node->data.seq_num, node
          );
    }

  }
  fprintf (recvd, "} ;\n");
  fclose(recvd);
}

double output_time(FILE *file, const char *label) {
  static double last_time = -1;
  if (last_time < 0) {
    last_time = all_start.tv_sec + all_start.tv_usec / 1e6;
  }

  struct timeval now_val;
  gettimeofday(&now_val, 0);
  double now = now_val.tv_sec + now_val.tv_usec / 1e6;
  double delta = now - last_time;
  
  fprintf(file, "%s:\t%g\n", label, delta);
  fflush(file);

  last_time = now;
  return now;
}

#define NUM_TIME_PROFILES 16
double profile_times [NUM_TIME_PROFILES];
static int profile_index = 0;
void profile_timing ()
{
  profile_times[profile_index] = (double)(get_cur_time_ns ()/1000);
  profile_index++;
  assert (profile_index < NUM_TIME_PROFILES);
}

void output_profile_timing (FILE * file)
{
  int i = 0;

  for (i = 1; i < profile_index; i++) {
    double delta = profile_times[i] - profile_times[i-1];
    fprintf(file, "%g\n", delta);
  }
}

/* vim: set tw=80 ts=2 sw=2 expandtab: */
