/*
 * Filename: rsd_compress.c
 * Author:   pratn
 * Created:  Sat Apr  7 14:27:09 2007
 * $Id: rsd_compress.c,v 1.32 2008/12/26 20:44:49 gamlin Exp $
 */
#include "umpi_internal.h"
#include "internal.h"
#include "rsd_queue.h"
#include "rank_recursive.h"
#include "Histogram.h"
#include "utils.h"
#include "TimeList.h"
#include "cvector.h"
#include "Event.h"

#define DEBUG(...)
static void update_file_offset_info(rsd_node *merge_head, rsd_node
*target_head);
int merge_vectors (cvector ** merge_vector, int merge_val,
    cvector ** target_vector, int target_val)
{
  char buf[2000];
  char buf1[2000];
  memset(buf,0,2000);
  memset(buf1, 0, 2000);
  if(NULL == *merge_vector && NULL == *target_vector) {
    vector_init (merge_vector, merge_val);
    vector_undo_info (*merge_vector, Undo_pos);
    vector_push_back (*merge_vector, target_val);
  } else if (*merge_vector && NULL == *target_vector) {
    if (vector_match_first (*merge_vector, target_val)) {
      return FALSE;
    } else {
      vector_undo_info (*merge_vector, Undo_pos);
      vector_push_back (*merge_vector, target_val);
    }
  } else if (merge_vector && target_vector) {
    if (vector_match (*merge_vector, *target_vector)) {
      vector_undo_info (*merge_vector, Undo_iters);
      vector_increment_count (*merge_vector);
    } else {
      vector_undo_info (*merge_vector, Undo_pos);
      vector_concat (*merge_vector, *target_vector);
    }
  } else {
    assert (0);
  }
  return TRUE;
}

/** 
 * This routine matches all pertinent fields of opstructs in rsd_nodes.
 * It also speculatively merges parameter queues in the merge sequence.
 */
static int match_loop(rsd_node *merge_head, rsd_node *merge_tail, 
    rsd_node *target_head, rsd_node *target_tail) {
  rsd_node *merge_iter = merge_tail;
  rsd_node *target_iter = target_tail;

  do {
    // check signatures 
    if(!SIG_MATCH(target_iter->signature, merge_iter->signature)) {
      return FALSE;
    }

    // check loop information - the merge head is a special case 
    // we can merge the target into any loop layer 
    int cur_depth;
    if(merge_head == merge_iter) {
      cur_depth = get_depth(target_iter->prsd_info.head);
    } else {
      cur_depth = get_depth(merge_iter->prsd_info.head);
    }

    if(!prsd_match(&target_iter->prsd_info, &merge_iter->prsd_info, cur_depth)) {
      DEBUG ("prsd match failed %d %d\n", merge_iter->data.op,
          target_iter->data.op);
      return FALSE;
    }

    // First, check that params all match.
    int param_ret = params_match(&target_iter->data, &merge_iter->data);
    int event_ret = compare_event_params(merge_iter->event, target_iter->event);
    //printf("rsd_compress: Event %d param match return value %d param: %d\n", target_iter->data.op, event_ret, param_ret);fflush(stdout);

    if(event_ret != MATCH) {

      // dependng on what type of op this is, we need to push different params on the param queue.
      // this figures out which is which before we init the queue and push.
      switch(event_ret) {
        case MFAIL_SRC:
          if (merge_vectors (&merge_iter->data.source_vector,
                merge_iter->data.source,
                &target_iter->data.source_vector,
                target_iter->data.source)) {
          } else {
            return FALSE;
          }
	  if(event_merge_vectors(merge_iter->event, 
		target_iter->event,TYPE_SOURCE)) {
	  } else {
	    return FALSE;
	  }
          break;

        case MFAIL_DST:
          if (merge_vectors (&merge_iter->data.dest_vector,
                merge_iter->data.dest,
                &target_iter->data.dest_vector,
                target_iter->data.dest) ) {
          } else {
            return FALSE;
          }
       	  if(event_merge_vectors(merge_iter->event, 
		target_iter->event,TYPE_DEST)) {
	  } else {
	    return FALSE;
	  }
	  break;

        case MFAIL_REQ:
          if (merge_vectors (&merge_iter->data.request_vector,
                merge_iter->data.request,
                &target_iter->data.request_vector,
                target_iter->data.request)) {
          } else {
            return FALSE;
          }
       	  if(event_merge_vectors(merge_iter->event, 
		target_iter->event,TYPE_REQUEST)) {
	  } else {
	    return FALSE;
	  }
          break;

        case MFAIL_CNT:
          if (merge_vectors (&merge_iter->data.count_vector,
                merge_iter->data.count,
                &target_iter->data.count_vector,
                target_iter->data.count)) {
          } else {
            return FALSE;
          }
       	  if(event_merge_vectors(merge_iter->event, 
		target_iter->event,TYPE_COUNT)) {
	  } else {
	    return FALSE;
	  }
          break;

        case MFAIL_ROOT:
          if (merge_vectors (&merge_iter->data.root_vector,
                merge_iter->data.root,
                &target_iter->data.root_vector,
                target_iter->data.root)) {
          } else {
            return FALSE;
          }
       	  if(event_merge_vectors(merge_iter->event, 
		target_iter->event,TYPE_ROOT)) {
	  } else {
	    return FALSE;
	  }
          break;

        //Added by Karthik - Start
        #ifdef FEATURE_FILE_IO
        case MFAIL_OFF:
          update_file_offset_info(merge_iter,target_iter);
          break;

        case MFAIL_DISP:
          if (merge_vectors (&merge_iter->data.disp_vector,
                merge_iter->data.disp,
                &target_iter->data.disp_vector,
                target_iter->data.disp)) {
          } else {
            return FALSE;
          }
          break;          
        #endif
        //Added by Karthik - End

        case MFAIL_ARR:  //fall-thru
        case MFAIL_GEN:  //fall-thru
        case MFAIL_FLN:  //fall-thru
        default:
          // just return false in case of a general mismatch.
          return FALSE;
      }

    }

    merge_iter = merge_iter->prev;
    target_iter = target_iter->prev;

  } while (merge_iter && 
      target_iter && 
      merge_iter != merge_head->prev && 
      target_iter != target_head->prev);

  // if we got through the whole sequence, this is true.
  return (merge_iter == merge_head->prev && target_iter == target_head->prev);
}


/**
 * Merges timing info if we have delta tracking enabled.
 */
static void merge_timing_info(rsd_node *merge, rsd_node *target) {
#if defined ( FEATURE_TIMING_HISTO ) || defined ( FEATURE_TIMING_BITS )
  if (!merge->prev || !target->prev) return;
        for (int i = 0; i < TIMING_MAX; i++) {
          timing_list_merge (&merge->data.time_info[i].time_list,
              merge->data.time_info[i].time,
              &merge->prev->signature,
              merge->node,
              &target->data.time_info[i].time_list,
              target->data.time_info[i].time,
              &target->prev->signature,
              target->node,
              NUM_BUCKETS
              );
          event_time_merge (i, merge->event,
              &merge->prev->signature,
              merge->node,
              target->event,
              &target->prev->signature,
              target->node,
              NUM_BUCKETS
              );

        }
#endif
}


/**
 * This routine is called when a full matching sequence is found in the rsd queue.
 * It merges timing information in the target sequence into the merge sequence, then
 * deletes the target sequence from the end of the queue.
 *
 * PRE: 
 * - Merge sequence and target sequence match.
 * - The target sequence starts at queue->tail and extends to merge_tail->next.
 * - The merge sequence is assumed to start at merge_tail.
 * - Merge sequence must be the same length as target sequence.
 */
static void merge_tail_into_queue(rsd_queue *queue, rsd_node *merge_tail) {
  rsd_node *target = queue->tail;   //target sequence always starts at queue tail, extends to just after merge tail
  rsd_node *merge = merge_tail;     //merge sequence starts at merge tail
  while(target != merge_tail) {
    merge_timing_info(merge, target);
    target = target->prev;

    free_rsd_node(target->next);      
    //if(merge->pqueue) {
    /* Mark three param flags as unused. */
    //  merge->data.fields &= ~(SOURCE | DEST | REQUEST); 
    //  reset_flag(merge->pqueue);
    //}
    vector_reset_undo_info (merge->data.source_vector);
    vector_reset_undo_info (merge->data.dest_vector);
    vector_reset_undo_info (merge->data.request_vector);
    vector_reset_undo_info (merge->data.count_vector);
    vector_reset_undo_info (merge->data.root_vector);
    event_vector_reset_undo(merge->event, TYPE_SOURCE);
    event_vector_reset_undo(merge->event, TYPE_DEST);
    event_vector_reset_undo(merge->event, TYPE_REQUEST);
    event_vector_reset_undo(merge->event, TYPE_COUNT);
    event_vector_reset_undo(merge->event, TYPE_ROOT);
    merge = merge->prev;
  
  }
  
  queue->tail = merge_tail;
}


/**
 * This routine undoes speculative merges of param queues in rsd_nodes.  It must be used if
 * match_loop() is called on an rsd sequence but fails.  This will return param queues in the
 * merge sequence to their pre-match state.
 */
static void undo_param_match(rsd_node *merge_head, rsd_node *merge_tail) {
  /* undo all param merges on failed verification phase */
  rsd_node *merge_iter = merge_tail;

  while(merge_iter != merge_head->prev) {
    vector_undo_addition (&merge_iter->data.source_vector, &merge_iter->data.source);
    vector_undo_addition (&merge_iter->data.dest_vector, &merge_iter->data.dest);
    vector_undo_addition (&merge_iter->data.request_vector, &merge_iter->data.request);
    vector_undo_addition (&merge_iter->data.count_vector, &merge_iter->data.count);
    vector_undo_addition (&merge_iter->data.root_vector, &merge_iter->data.root);
    event_vector_undo_addition(merge_iter->event, TYPE_SOURCE);
    event_vector_undo_addition(merge_iter->event, TYPE_DEST);
    event_vector_undo_addition(merge_iter->event, TYPE_REQUEST);
    event_vector_undo_addition(merge_iter->event, TYPE_COUNT);
    event_vector_undo_addition(merge_iter->event, TYPE_ROOT);
    merge_iter = merge_iter->prev;
  }
}
//Added by Karthik - Start
#ifdef FEATURE_FILE_IO
static void update_file_offset_info(rsd_node *merge_head, rsd_node *target_head)
{
  int diff = 0;
  if(merge_head->data.offsetinfo.stride == 0)
  {
    diff = target_head->data.offset - merge_head->data.offset;
    merge_head->data.offsetinfo.startpos = merge_head->data.offset;
    merge_head->data.offsetinfo.stride = diff;  
  }
  (merge_head->data.offsetinfo.totalcount)++;
}
#endif
//Added by Karthik - End
/**
 * Adjusts rsd info on nodes in the merge sequence to reflect the appropriate number of loop
 * iterations.  If the PRSD sequence was not determined to be a loop already, this creates a loop
 * and marks all of its members.  If there was a loop already, this will update the loop PRSD's
 * iteration count by one.  
 * 
 * Note that this routine does not do the work of merging; it only does PRSD bookkeeping.
 */
static void update_loop_info(rsd_node *merge_head, rsd_node *merge_tail, rsd_node *target_head) {
  /* already a leader - check to see if this is another iteration
     by checking if stored tail matches the current tail */
  int target_depth = get_depth(target_head->prsd_info.head);
  if ((merge_head->rsd_info & LEADER_RSD) && (get_depth(merge_head->prsd_info.head) != target_depth)) {
    prsd_node *ptr = merge_head->prsd_info.tail;
    if(target_depth != 0) {
      for(int i = 1; i <= target_depth; i++) {

        ptr = ptr->prev;
        if(!ptr) {
          fprintf(stderr, "error: attempting to go up - start = %d, iters = %d, curr = %d (err1)\n", 
              get_depth(merge_head->prsd_info.head), target_depth, i);
          exit(0);
        }
      }
    }
    // increment appropriate-depth prsd iteration count for the loop we're about to overlay
    ptr->loop_iters++;  

  } else {
    /* new loop leader found */
    int members = 1;
    for (rsd_node *merge = merge_head->next; merge != merge_tail->next; merge = merge->next) {
      merge->rsd_info |= MEMBER_RSD; 	/* add member */
      members++;
    }
    merge_head->rsd_info |= LEADER_RSD;
    prsd_push(&merge_head->prsd_info, members, &merge_tail->signature);
  }
  /*Added by Karthik - Start
  #ifdef FEATURE_FILE_IO
  update_file_offset_info(merge_head,target_head);
  #endif
  //Added by Karthik - End*/
}

void compress_rsd(rsd_queue *queue, replay_op *operation) {
  compress_rsd_event(queue, operation, NULL);
}

static long long int total_op_cnt;

/**
 * This routine provides the actual intra-node compression.
 */
void compress_rsd_event(rsd_queue *queue, replay_op *operation, Event *event) {
  static int waitsome = 0;
  static int warned_once = 0;
  int distance;

  int head_flag;      /* Whether target head has been found. */
  int match_flag;     /* Whether a node matching target head has been found. */
  int found_flag;     /* */

  rsd_node *target_tail = NULL;
  rsd_node *target_head = NULL;
  rsd_node *merge_tail = NULL;
  rsd_node *merge_head = NULL;

  if (operation != NULL) {
    total_op_cnt++;
    if(operation->op == umpi_MPI_Finalize)
      printf("Rank %d: Total op cnt: %lld\n", get_my_rank(), total_op_cnt);
    rsd_push(queue, operation, event);
  }

  //if(event)
    //event_print(event);

#ifdef COMPRESS_INTRA_NODE
  /* Compression algorithm follows. */

  /* waitsome handler */
  if(queue->last_op_added == umpi_MPI_Waitsome && !waitsome) {
    waitsome = 1;
    compress_rsd(queue, NULL);
    waitsome = 0;
  }

  distance = 0;
  head_flag = 0;
  found_flag = 1;

  do {
    /* we must find the target head */
    target_tail = queue->tail;

    if(found_flag || !head_flag) {
      /* if the found flag is set - start a new search */
      target_head = target_tail;
      distance = 0;
    } else {
      /* otherwise cont a previous search */
      target_head = target_head->prev;
    }

    head_flag = 0;
    match_flag = 0;
    found_flag = 0;

    /* Search backwards in the queue until we find a match with the target tail.
       After this loop is done, either head_flag will be false (and nothing found)
       or the head will be just after the matching tail.
    */
    while(target_head->prev != NULL && (window == -1 || window > distance)) {
      if(!(target_head->rsd_info & MEMBER_RSD) && SIG_MATCH(target_head->prev->signature, target_tail->signature)) {
        head_flag = 1;
        break;
      }  
      target_head = target_head->prev;
      distance++;
    }

    // didn't find head: can skip the rest.
    if (!head_flag) {
      // Warn if the window was exceeded, but only warn once. 
      if(!warned_once && window != -1 && window <= distance) {
        fprintf(stderr, "warning: window exceeded - possible compression missed on sequence %d (window set at %d)\n", 
            target_tail->data.seq_num, window);
        fflush(stderr);
        warned_once = 1;
      }
      continue;
    }

    // If the head flag is true, we've got the target_tail, the target_head, and the merge_tail (just before target_head).
    // Now we can try to find a match for the target_head somewhere before the merge tail.  This will be merge_head.
    merge_head = target_head->prev;
    merge_tail = target_head->prev;
    while(merge_head != NULL) {
      if(!(merge_head->rsd_info & MEMBER_RSD) && SIG_MATCH(merge_head->signature, target_head->signature)) {
        match_flag = 1;
        break;
      }
      merge_head = merge_head->prev;
    }

    // TODO: can we check lengths of sequences here and avoid a traversal if there's clearly no match? 
    // If we didn't find the merge head, we know there's no match in the sequence. Bail. 
    if(!match_flag) 
    {
      continue;
    }

    // Make sure the sequences [target_head .. target_tail] and [merge_head .. merge_tail] match.
    // We do partial merges as we match, and these must be undone at the end if we find that the sequences
    // don't match. 
    if (!match_loop(merge_head, merge_tail, target_head, target_tail)) {
      // undo all param merges if verification failed, then skip to next step. 
      undo_param_match(merge_head, merge_tail);
      continue;
    }
    found_flag = 1;

    // this adjusts prsd info in the merge sequence to reflect the new loop we're adding 
    update_loop_info(merge_head, merge_tail, target_head);

    // if the compression was successful, go through and free all the target nodes from the end
    // of the seqence.  this will also merge nodes
    merge_tail_into_queue(queue, merge_tail);

  } while(head_flag || found_flag);

#endif /* COMPRESS_INTRA_NODE */
}

/* each count should vary only within x percent of the average */
extern int alltoallv_limit;
static int is_deviation_within_limit (int * counts, int size, int * average) {

  int sum = 0, avg = 0;

  for (int i = 0; i < size; i++)
    sum += counts[i];
  avg = (int)( sum / size );

  int limit = (alltoallv_limit*avg)/100;
  for (int i = 0; i < size; i++)
    if (abs (counts[i] - avg) > limit)
      return FALSE;

  *average = avg;
  return TRUE;
}

int can_merge_alltoallv (int * sendcounts, int * recvcounts,
                         MPI_Comm comm, int comm_size, int * average)
{
  int sendavg = 0, recvavg = 0;

  /* check if sendcounts are okay */
  int sendmatch = is_deviation_within_limit (sendcounts, comm_size, &sendavg);
  if (sendmatch == FALSE) return FALSE;

  /* check if recvcounts are okay */
  int recvmatch = is_deviation_within_limit (recvcounts, comm_size, &recvavg);
  if (recvmatch == FALSE) return FALSE;

  /* calculate average of send and recv counts and calculate inter-node avg */
  int node_avg = (sendavg + recvavg) / 2;
  int final_avg = 0;
  PMPI_Allreduce (&node_avg, &final_avg, 1, MPI_INT, MPI_SUM, comm);
  final_avg /= comm_size;

  int within_limit = 1, retval = 0;
  int limit = (alltoallv_limit*final_avg)/100;
  if (abs (final_avg - node_avg) > limit)
    within_limit = 0;

  PMPI_Allreduce (&within_limit, &retval, 1, MPI_INT, MPI_MIN, comm);

  *average = final_avg;
  if (retval == 1) {
    return TRUE;
  } else {
    return FALSE;
  }
}
/* vim: set tw=80 ts=2 sw=2 expandtab: */
