#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "code_generator_handler.h"

var_dump_t* init_var_dump() {
  var_dump_t* var = NULL;
  var = (var_dump_t*)malloc(sizeof(var_dump_t));
  var->total_count = 0;
  return var;
}

int add_var_dump(var_dump_t* var, char* detail, int is_global) {
  int success = 1;
  if(var == NULL) {
    var = init_var_dump();
  }
  if(var != NULL && strlen(detail)>0 && var->total_count < (VAR_LIST_SIZE-1)) {
    var->var_detail[var->total_count] = (char*)malloc((strlen(detail)+1)*sizeof(char));
    strcpy(var->var_detail[var->total_count], detail);
    var->is_global[var->total_count] = is_global;
    (var->total_count)++;
    success = 0;
  }
  return success;
}

void free_var_dump(var_dump_t* var) {
  int i = 0;
  if(var != NULL && var->total_count > 0) {
    for(i=0;i<var->total_count;i++) {
      free(var->var_detail[i]);
    }
    free(var);
  }
}

derived_type_dump_t* init_derived_type_dump() {
  derived_type_dump_t* type = NULL;
  type = (derived_type_dump_t*)malloc(sizeof(derived_type_dump_t));
  type->total_count = 0;
  return type;
}

int add_derived_type(derived_type_dump_t* type, int trace_index) {
  int success = 1;
  char typevar[20];
  char typedet[100];
  char* check = NULL;
  if(type == NULL) {
    type = init_derived_type_dump();
  }
  if(check = get_type_name_from_index(type, trace_index)) {
    success = 0;
    return success;
  }
  if(type != NULL && type->total_count < (TYPE_LIST_SIZE-1)) {
    sprintf(typevar, "datatype%d",type->total_count);
    sprintf(typedet, "MPI_Datatype %s;",typevar);
    type->type_detail[type->total_count] = (char*)malloc((strlen(typedet)+1)*sizeof(char));
    type->type_name[type->total_count] = (char*)malloc((strlen(typevar)+1)*sizeof(char));
    strcpy(type->type_detail[type->total_count], typedet);
    strcpy(type->type_name[type->total_count], typevar);
    type->is_derived_type[type->total_count] = 1;
    type->type_trace_index[type->total_count] = trace_index;
    (type->total_count)++;
    success = 0;
  }
  return success;
}

int add_defined_type(derived_type_dump_t* type, char* name, int trace_index) {
  int success = 1;
  char* check = NULL;
  if(type == NULL) {
    type = init_derived_type_dump();
  }
  if(check = get_type_name_from_index(type, trace_index)) {
    success = 0;
    return success;
  }
  if(type != NULL && type->total_count < (TYPE_LIST_SIZE-1)) {
    type->type_name[type->total_count] = (char*)malloc((strlen(name)+1)*sizeof(char));
    strcpy(type->type_name[type->total_count], name);
    type->is_derived_type[type->total_count] = 0;
    type->type_trace_index[type->total_count] = trace_index;
    (type->total_count)++;
    success = 0;
  }
  return success;
}

void update_usage_count(derived_type_dump_t* type, int usage_count, int trace_index) {
  int i=0;
  if(type != NULL && type->total_count > 0) {
    for(i=0;i<type->total_count;i++) {
      if(type->type_trace_index[i] == trace_index) {
        if(type->type_usage_count[i] < usage_count) {
          type->type_usage_count[i] = usage_count;
        }
        break;
      }
    }
  }
}

char* get_type_name_from_index(derived_type_dump_t* type, int trace_index) {
  char* typename = NULL;
  int i=0;
  if(type == NULL) {
    return typename;
  }
  if(type!=NULL && type->total_count>0) {
    for(i=0;i<type->total_count;i++) {
      if(trace_index == type->type_trace_index[i]) {
        typename = type->type_name[i];
        break;
      }
    }
  }
  return typename;
}

void free_derived_type_dump(derived_type_dump_t* type) {
  int i = 0;
  if(type != NULL && type->total_count > 0) {
    for(i=0;i<type->total_count;i++) {
      free(type->type_detail[i]);
      free(type->type_name[i]);
    }
    free(type);
  }
}

file_handle_dump_t* init_fh_dump() {
  file_handle_dump_t* dump = NULL;
  dump = (file_handle_dump_t*)malloc(sizeof(file_handle_dump_t));
  dump->total_count = 0;
  return dump;
}

int add_file_handle(file_handle_dump_t* dump, int trace_index) {
  int success = 1;
  char fhvar[20];
  char fhdet[100];
  char* check = NULL;
  if(dump == NULL) {
    dump = init_fh_dump();
  }
  if(check = get_file_handle_from_index(dump, trace_index)) {
    success = 0;
    return success;
  }
  if(dump != NULL && dump->total_count < (TYPE_LIST_SIZE-1)) {
    sprintf(fhvar, "fh%d",dump->total_count);
    sprintf(fhdet, "MPI_File %s;",fhvar);
    dump->fh_detail[dump->total_count] = (char*)malloc((strlen(fhdet)+1)*sizeof(char));
    dump->fh_name[dump->total_count] = (char*)malloc((strlen(fhvar)+1)*sizeof(char));
    strcpy(dump->fh_detail[dump->total_count], fhdet);
    strcpy(dump->fh_name[dump->total_count], fhvar);
    dump->fh_trace_index[dump->total_count] = trace_index;
    (dump->total_count)++;
    success = 0;
  }
  return success;
}


char* get_file_handle_from_index(file_handle_dump_t* dump, int trace_index) {
  char* fhname = NULL;
  int i=0;
  if(dump == NULL) {
    return fhname;
  }
  if(dump!=NULL && dump->total_count>0) {
    for(i=0;i<dump->total_count;i++) {
      if(trace_index == dump->fh_trace_index[i]) {
        fhname = dump->fh_name[i];
        break;
      }
    }
  }
  return fhname;
}

void free_fh_dump(file_handle_dump_t* dump) {
  int i = 0;
  if(dump != NULL && dump->total_count > 0) {
    for(i=0;i<dump->total_count;i++) {
      free(dump->fh_detail[i]);
      free(dump->fh_name[i]);
    }
    free(dump);
  }
}
/*malloc_dump_t* init_malloc_dump() {
  malloc_dump_t* dump = NULL;
  dump = (malloc_dump_t*)malloc(sizeof(malloc_dump_t));
  dump->total_count = 0;
  return dump;
}

int add_malloc_dump(malloc_dump_t* dump, char* detail, int seq_num) {
  int success = 1;
  if(dump == NULL) {
    dump = init_malloc_dump();
  }
  if(dump != NULL && strlen(detail)>0 && var->total_count < (VAR_LIST_SIZE-1)) {
    var->var_detail[var->total_count] = (char*)malloc((strlen(detail)+1)*sizeof(char));
    strcpy(var->var_detail[var->total_count], detail);
    (var->total_count)++;
    success = 0;
  }
  return success;
}

void free_malloc_dump(malloc_dump_t* dump) {
  int i = 0;
  if(var != NULL && var->total_count > 0) {
    for(i=0;i<var->total_count;i++) {
      free(var->var_detail[i]);
    }
    free(var);
  }
}*/

