#ifndef CODE_GENERATOR_HANDLER_H
#define CODE_GENERATOR_HANDLER_H

#define TYPE_LIST_SIZE 100
#define VAR_LIST_SIZE 500

/*
 * This struct contains an array of char pointer to hold new variables
 * that needs to be created and listed in code generation.
 * total_count will maintain the count of variables currently active in
 * the array.
 */
typedef struct var_dump {
  char* var_detail[VAR_LIST_SIZE];
  int is_global[VAR_LIST_SIZE];
  int total_count;
} var_dump_t;
  
var_dump_t* init_var_dump();
  
int add_var_dump(var_dump_t* var_dump, char* detail, int is_global);

void free_var_dump(var_dump_t* var_dump);

/*
 * This struct will hold all new datatypes created and new variable names
 * to be used in the skeleton code.
 */
typedef struct derived_type_dump {
  char* type_detail[TYPE_LIST_SIZE];
  char* type_name[TYPE_LIST_SIZE];
  int type_usage_count[TYPE_LIST_SIZE];
  int is_derived_type[TYPE_LIST_SIZE];
  int type_trace_index[TYPE_LIST_SIZE];
  int total_count;
} derived_type_dump_t;
  
derived_type_dump_t* init_derived_type_dump();

int add_derived_type(derived_type_dump_t* type, int trace_index);

int add_defined_type(derived_type_dump_t* type, char* name, int trace_index);

void update_usage_count(derived_type_dump_t* type, int usage_count, int trace_index);

char* get_type_name_from_index(derived_type_dump_t* type, int trace_index);

void free_derived_type_dump(derived_type_dump_t* type);

typedef struct file_handle_dump {
  char* fh_detail[TYPE_LIST_SIZE];
  char* fh_name[TYPE_LIST_SIZE];
  int fh_trace_index[TYPE_LIST_SIZE];
  int total_count;
} file_handle_dump_t;
  
file_handle_dump_t* init_fh_dump();

int add_file_handle(file_handle_dump_t* dump, int trace_index);

char* get_file_handle_from_index(file_handle_dump_t* dump, int trace_index);

void free_fh_dump(file_handle_dump_t* dump);

#endif
