#include <mpi.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <string.h>
#include <unistd.h>
#include "prsd_traverse.h"
#include "rsd_reader.h"
#include "op_exec.h"
#include "trace.h"
#include "utils.h"
#include "mpi_util.h"

//#define DEBUG 0
#ifdef LINK_WITH_RECORD
#define MPI_Pcontrol(a)
#endif

/**
 * Visitor routine for executing replay operations.  This is called
 * by the traversal routines to translate ops to actual MPI function 
 * calls.
 */
static int replay_visitor(rsd_node *node, int rank, int depth, int iter, int
    last) {
  return exec_prsd_call(node, rank, depth, iter);
}


int main(int argc, char *argv[])
{
  int rank = -1, len = -1, whole, frac;
  int retval = -1;
  int per_node_trace = FALSE;
  struct timeval t_start, t_stop;
  struct timezone tz;
  rsd_queue *prsd_queue;
  char * dir = 0;
  char format = 'x';

  MPI_Init(&argc, &argv);
  //disable profiling at init
  MPI_Pcontrol (0);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &my_size);
  my_rank = rank;

  /* determine input type */
  if(argc < 2) {
    fprintf(stderr, "usage: %s -b|-a|-p <trace dir>\n", argv[0]);
    MPI_Finalize ();
    return 0;
  }
  while (1)
  {
    retval = getopt (argc, argv, "p:X");

    if (retval == -1)
      break;
    switch (retval)
    {
      case 'p':
        format = 'p';
        dir = optarg;
        break;
      case 'X':
        per_node_trace = TRUE;
        break;
    }
  }
  if (format == -1)
  {
    fprintf(stderr, "error: must specify the input file's format w/ -p\n");
    MPI_Finalize ();
    return 0;
  }

  /* read in op trace for each node */
  char filename[4096];
  sprintf (filename, "%s/%d", dir, (per_node_trace ? rank : 0));
  prsd_queue = read_rsd_file_rank(filename, rank);
  len = rsd_get_length(prsd_queue);

  /* playback ops in queue */
  gettimeofday(&t_start, &tz);

#if DEBUG
  if (rank == DEBUG)  output_rsd_queue (stdout, prsd_queue);
  fprintf (stdout, "output rsd queue\n");
  fflush (stdout);
#else
  replay_queue(prsd_queue->head, rank, replay_visitor);
#endif
  // finish profiling
  MPI_Pcontrol (0);

  /* stop timer */
  gettimeofday(&t_stop, &tz);
  whole = t_stop.tv_sec - t_start.tv_sec;
  frac = t_stop.tv_usec - t_start.tv_usec;
  if(frac < 0) {
    frac *= -1;
    whole --;
  }
  if(rank == 0)
    printf("execution time:  %d.%d\n", whole, frac);

  MPI_Finalize();
  return 0;
}
/* EOF */
