#include <mpi.h>
#include <stdio.h>
#include <time.h>
#include <signal.h>

#include "stackwalk/h/walker.h"
#include "stackwalk/h/frame.h"
#include "stackwalk/h/swk_errors.h"

#include <vector>
using namespace std;
using namespace Dyninst;
using namespace Stackwalker;

int rank, size;

volatile unsigned long long inc = 0;
void bar2()
{
  fprintf(stderr, "Goodbye world\n");
}

void walk_stack() 
{
  static Walker *walker = NULL;
  if (!walker) {
    walker = Walker::newWalker();
  }
  if (!walker) {
    fprintf(stderr, "Could not create walker: %s", getLastErrorMsg());
    return;
  }

  vector<Frame> stackwalk;
  walker->walkStack(stackwalk);
  for (unsigned i=0; i<stackwalk.size(); i++)
  {
    printf("%d: %lx\n", rank, stackwalk[i].getRA());
  }
}

time_t start;
time_t current;

int bar()
{
  time_t last;
  unsigned i;
  current = start;
  last = current;
  while (current < start + (30 * 1)) {
    volatile double f = 1.0f;
    for (i=0; i<100000; i++) {
      f = f * 1.0001;
    }
    inc++;
    current = time(NULL);
    if (current >= last + 5) {
      fprintf(stderr, "Looping in bar\n");
      last = current;
      //      kill(getpid(), SIGUSR1);
      walk_stack();
    }
  }
  return 0;
}

int foo()
{
  start = time(NULL);
  return bar();
}

void handler(int sig)
{
  fprintf(stderr, "In handler\n");
}

int main(int argc, char *argv[]) 
{
   MPI_Init(&argc, &argv);
   MPI_Comm_rank(MPI_COMM_WORLD, &rank);
   MPI_Comm_size(MPI_COMM_WORLD, &size);

   signal(SIGUSR1, handler);
   fprintf(stderr, "Hello world %d/%d\n", rank, size );
   foo();
   fprintf(stderr, "Goodbye world %d/%d, inc = %lld\n", rank, size, inc);


   MPI_Finalize();
   return 0;
}


